@@ -28,7 +28,6 @@ import GHC.Types.Unique.Supply
 
 import GHC.Utils.Panic
 
-import Control.Monad (replicateM)
 import Data.Maybe (fromMaybe)
 
 import GHC.Stack
@@ -73,12 +72,6 @@ instance Outputable RegUsage where
 regUsageOfInstr :: Platform -> Instr -> RegUsage
 regUsageOfInstr platform instr = case instr of
   ANN _ i                  -> regUsageOfInstr platform i
-  COMMENT{}                -> usage ([], [])
-  MULTILINE_COMMENT{}      -> usage ([], [])
-  PUSH_STACK_FRAME         -> usage ([], [])
-  POP_STACK_FRAME          -> usage ([], [])
-  DELTA{}                  -> usage ([], [])
-
   -- 1. Arithmetic Instructions ------------------------------------------------
   ADD dst src1 src2        -> usage (regOp src1 ++ regOp src2, regOp dst)
   CMN l r                  -> usage (regOp l ++ regOp r, [])
@@ -143,7 +136,7 @@ regUsageOfInstr platform instr = case instr of
   FCVTZS dst src           -> usage (regOp src, regOp dst)
   FABS dst src             -> usage (regOp src, regOp dst)
 
-  _ -> panic $ "regUsageOfInstr: " ++ instrCon instr
+  _ -> panic "regUsageOfInstr"
 
   where
         -- filtering the usage is necessary, otherwise the register
@@ -173,8 +166,6 @@ regUsageOfInstr platform instr = case instr of
         interesting _        (RegVirtual _)                 = True
         interesting _        (RegReal (RealRegSingle (-1))) = False
         interesting platform (RegReal (RealRegSingle i))    = freeReg platform i
-        interesting _        (RegReal (RealRegPair{}))
-            = panic "AArch64.Instr.interesting: no reg pairs on this arch"
 
 -- Save caller save registers
 -- This is x0-x18
@@ -209,12 +200,7 @@ callerSavedRegisters
 patchRegsOfInstr :: Instr -> (Reg -> Reg) -> Instr
 patchRegsOfInstr instr env = case instr of
     -- 0. Meta Instructions
-    ANN d i             -> ANN d (patchRegsOfInstr i env)
-    COMMENT{}           -> instr
-    MULTILINE_COMMENT{} -> instr
-    PUSH_STACK_FRAME    -> instr
-    POP_STACK_FRAME     -> instr
-    DELTA{}             -> instr
+    ANN d i        -> ANN d (patchRegsOfInstr i env)
     -- 1. Arithmetic Instructions ----------------------------------------------
     ADD o1 o2 o3   -> ADD (patchOp o1) (patchOp o2) (patchOp o3)
     CMN o1 o2      -> CMN (patchOp o1) (patchOp o2)
@@ -280,7 +266,8 @@ patchRegsOfInstr instr env = case instr of
     SCVTF o1 o2    -> SCVTF (patchOp o1) (patchOp o2)
     FCVTZS o1 o2   -> FCVTZS (patchOp o1) (patchOp o2)
     FABS o1 o2     -> FABS (patchOp o1) (patchOp o2)
-    _              -> panic $ "patchRegsOfInstr: " ++ instrCon instr
+
+    _ -> pprPanic "patchRegsOfInstr" (text $ show instr)
     where
         patchOp :: Operand -> Operand
         patchOp (OpReg w r) = OpReg w (env r)
@@ -336,7 +323,7 @@ patchJumpInstr instr patchF
         B (TBlock bid) -> B (TBlock (patchF bid))
         BL (TBlock bid) ps rs -> BL (TBlock (patchF bid)) ps rs
         BCOND c (TBlock bid) -> BCOND c (TBlock (patchF bid))
-        _ -> panic $ "patchJumpInstr: " ++ instrCon instr
+        _ -> pprPanic "patchJumpInstr" (text $ show instr)
 
 -- -----------------------------------------------------------------------------
 -- Note [Spills and Reloads]
@@ -463,7 +450,7 @@ mkStackDeallocInstr platform n
 mkStackDeallocInstr _platform n = pprPanic "mkStackDeallocInstr" (int n)
 
 --
--- See note [extra spill slots] in X86/Instr.hs
+-- See Note [extra spill slots] in X86/Instr.hs
 --
 allocMoreStack
   :: Platform
@@ -475,7 +462,7 @@ allocMoreStack _ _ top@(CmmData _ _) = return (top,[])
 allocMoreStack platform slots proc@(CmmProc info lbl live (ListGraph code)) = do
     let entries = entryBlocks proc
 
-    uniqs <- replicateM (length entries) getUniqueM
+    uniqs <- getUniquesM
 
     let
       delta = ((x + stackAlign - 1) `quot` stackAlign) * stackAlign -- round up
@@ -648,69 +635,10 @@ data Instr
     -- Float ABSolute value
     | FABS Operand Operand
 
-instrCon :: Instr -> String
-instrCon i =
-    case i of
-      COMMENT{} -> "COMMENT"
-      MULTILINE_COMMENT{} -> "COMMENT"
-      ANN{} -> "ANN"
-      LOCATION{} -> "LOCATION"
-      LDATA{} -> "LDATA"
-      NEWBLOCK{} -> "NEWBLOCK"
-      DELTA{} -> "DELTA"
-      SXTB{} -> "SXTB"
-      UXTB{} -> "UXTB"
-      SXTH{} -> "SXTH"
-      UXTH{} -> "UXTH"
-      PUSH_STACK_FRAME{} -> "PUSH_STACK_FRAME"
-      POP_STACK_FRAME{} -> "POP_STACK_FRAME"
-      ADD{} -> "ADD"
-      CMN{} -> "CMN"
-      CMP{} -> "CMP"
-      MSUB{} -> "MSUB"
-      MUL{} -> "MUL"
-      NEG{} -> "NEG"
-      SDIV{} -> "SDIV"
-      SMULH{} -> "SMULH"
-      SMULL{} -> "SMULL"
-      SUB{} -> "SUB"
-      UDIV{} -> "UDIV"
-      SBFM{} -> "SBFM"
-      UBFM{} -> "UBFM"
-      SBFX{} -> "SBFX"
-      UBFX{} -> "UBFX"
-      AND{} -> "AND"
-      ANDS{} -> "ANDS"
-      ASR{} -> "ASR"
-      BIC{} -> "BIC"
-      BICS{} -> "BICS"
-      EON{} -> "EON"
-      EOR{} -> "EOR"
-      LSL{} -> "LSL"
-      LSR{} -> "LSR"
-      MOV{} -> "MOV"
-      MOVK{} -> "MOVK"
-      MVN{} -> "MVN"
-      ORN{} -> "ORN"
-      ORR{} -> "ORR"
-      ROR{} -> "ROR"
-      TST{} -> "TST"
-      STR{} -> "STR"
-      LDR{} -> "LDR"
-      STP{} -> "STP"
-      LDP{} -> "LDP"
-      CSET{} -> "CSET"
-      CBZ{} -> "CBZ"
-      CBNZ{} -> "CBNZ"
-      J{} -> "J"
-      B{} -> "B"
-      BL{} -> "BL"
-      BCOND{} -> "BCOND"
-      DMBSY{} -> "DMBSY"
-      FCVT{} -> "FCVT"
-      SCVTF{} -> "SCVTF"
-      FCVTZS{} -> "FCVTZS"
-      FABS{} -> "FABS"
+instance Show Instr where
+    show (LDR _f o1 o2) = "LDR " ++ show o1 ++ ", " ++ show o2
+    show (MOV o1 o2) = "MOV " ++ show o1 ++ ", " ++ show o2
+    show _ = "missing"
 
 data Target
     = TBlock BlockId
@@ -838,11 +766,11 @@ opRegUExt W64 r = OpRegExt W64 r EUXTX 0
 opRegUExt W32 r = OpRegExt W32 r EUXTW 0
 opRegUExt W16 r = OpRegExt W16 r EUXTH 0
 opRegUExt W8  r = OpRegExt W8  r EUXTB 0
-opRegUExt w  _r = pprPanic "opRegUExt" (ppr w)
+opRegUExt w  _r = pprPanic "opRegUExt" (text $ show w)
 
 opRegSExt :: Width -> Reg -> Operand
 opRegSExt W64 r = OpRegExt W64 r ESXTX 0
 opRegSExt W32 r = OpRegExt W32 r ESXTW 0
 opRegSExt W16 r = OpRegExt W16 r ESXTH 0
 opRegSExt W8  r = OpRegExt W8  r ESXTB 0
-opRegSExt w  _r = pprPanic "opRegSExt" (ppr w)
+opRegSExt w  _r = pprPanic "opRegSExt" (text $ show w)
