@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP           #-}
 {-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE GADTs         #-}
 {-# LANGUAGE LambdaCase    #-}
@@ -26,14 +25,14 @@ module GHC.CmmToC
    )
 where
 
-#include "HsVersions.h"
-
--- Cmm stuff
 import GHC.Prelude
 
+import GHC.Platform
+
+import GHC.CmmToAsm.CPrim
+
 import GHC.Cmm.BlockId
 import GHC.Cmm.CLabel
-import GHC.Types.ForeignCall
 import GHC.Cmm hiding (pprBBlock)
 import GHC.Cmm.Ppr () -- For Outputable instances
 import GHC.Cmm.Dataflow.Block
@@ -41,32 +40,26 @@ import GHC.Cmm.Dataflow.Collections
 import GHC.Cmm.Dataflow.Graph
 import GHC.Cmm.Utils
 import GHC.Cmm.Switch
+import GHC.Cmm.InitFini
 
--- Utils
-import GHC.CmmToAsm.CPrim
-import GHC.Driver.Session
-import GHC.Driver.Ppr
-import GHC.Data.FastString
-import GHC.Utils.Outputable
-import GHC.Utils.Panic
-import GHC.Platform
+import GHC.Types.ForeignCall
 import GHC.Types.Unique.Set
 import GHC.Types.Unique.FM
 import GHC.Types.Unique
-import GHC.Utils.Misc
 
--- The rest
+import GHC.Utils.Outputable
+import GHC.Utils.Panic
+import GHC.Utils.Misc
+import GHC.Utils.Trace
+
 import Data.ByteString (ByteString)
 import qualified Data.ByteString as BS
-import Control.Monad.ST
 import Data.Char
 import Data.List (intersperse)
 import Data.Map (Map)
-import Data.Word
 import qualified Data.Map as Map
 import Control.Monad (ap)
-import qualified Data.Array.Unsafe as U ( castSTUArray )
-import Data.Array.ST
+import GHC.Float
 
 -- --------------------------------------------------------------------------
 -- Now do some real work
@@ -108,6 +101,9 @@ pprTop platform = \case
 
   -- We only handle (a) arrays of word-sized things and (b) strings.
 
+  cmm_data | Just (initOrFini, clbls) <- isInitOrFiniArray cmm_data ->
+    pprCtorArray platform initOrFini clbls
+
   (CmmData section (CmmStaticsRaw lbl [CmmString str])) ->
     pprExternDecl platform lbl $$
     hcat [
@@ -172,6 +168,7 @@ pprAlignment words =
      text "__attribute__((aligned(" <> int (widthInBytes words) <> text ")))"
 
 -- Note [StgWord alignment]
+-- ~~~~~~~~~~~~~~~~~~~~~~~~
 -- C codegen builds static closures as StgWord C arrays (pprWordArray).
 -- Their real C type is 'StgClosure'. Macros like UNTAG_CLOSURE assume
 -- pointers to 'StgClosure' are aligned at pointer size boundary:
@@ -207,7 +204,7 @@ pprStmt :: Platform -> CmmNode e x -> SDoc
 pprStmt platform stmt =
     case stmt of
     CmmEntry{}   -> empty
-    CmmComment _ -> empty -- (hang (text "/*") 3 (ftext s)) $$ ptext (sLit "*/")
+    CmmComment _ -> empty -- (hang (text "/*") 3 (ftext s)) $$ text "*/"
                           -- XXX if the string contains "*/", we need to fix it
                           -- XXX we probably want to emit these comments when
                           -- some debugging option is on.  They can get quite
@@ -221,7 +218,7 @@ pprStmt platform stmt =
     CmmStore  dest src align
         | typeWidth rep == W64 && wordWidth platform /= W64
         -> (if isFloatType rep then text "ASSIGN_DBL"
-                               else ptext (sLit ("ASSIGN_Word64"))) <>
+                               else text "ASSIGN_Word64") <>
            parens (mkP_ <> pprExpr1 platform dest <> comma <> pprExpr platform src) <> semi
 
         | otherwise
@@ -272,12 +269,17 @@ pprStmt platform stmt =
         hresults = zip results res_hints
         hargs    = zip args arg_hints
 
+        need_cdecl
+          | MO_ResumeThread  <- op                 = True
+          | MO_SuspendThread <- op                 = True
+          | otherwise                              = False
+
         fn_call
           -- The mem primops carry an extra alignment arg.
           -- We could maybe emit an alignment directive using this info.
           -- We also need to cast mem primops to prevent conflicts with GCC
           -- builtins (see bug #5967).
-          | Just _align <- machOpMemcpyishAlign op
+          | need_cdecl
           = (text ";EFF_(" <> fn <> char ')' <> semi) $$
             pprForeignCall platform fn cconv hresults hargs
           | otherwise
@@ -340,21 +342,23 @@ pprSwitch platform e ids
   where
     (pairs, mbdef) = switchTargetsFallThrough ids
 
+    rep = typeWidth (cmmExprType platform e)
+
     -- fall through case
     caseify (ix:ixs, ident) = vcat (map do_fallthrough ixs) $$ final_branch ix
         where
         do_fallthrough ix =
-                 hsep [ text "case" , pprHexVal platform ix (wordWidth platform) <> colon ,
+                 hsep [ text "case" , pprHexVal platform ix rep <> colon ,
                         text "/* fall through */" ]
 
         final_branch ix =
-                hsep [ text "case" , pprHexVal platform ix (wordWidth platform) <> colon ,
+                hsep [ text "case" , pprHexVal platform ix rep <> colon ,
                        text "goto" , (pprBlockId ident) <> semi ]
 
     caseify (_     , _    ) = panic "pprSwitch: switch with no cases!"
 
     def | Just l <- mbdef = text "default: goto" <+> pprBlockId l <> semi
-        | otherwise       = empty
+        | otherwise       = text "default: __builtin_unreachable();"
 
 -- ---------------------------------------------------------------------
 -- Expressions.
@@ -493,7 +497,7 @@ machOpNeedsCast platform mop args
     -- resultRepOfMachOp says).
   | isComparisonMachOp mop = Just mkW_
 
-    -- See Note [Zero-extended sub-word signed results]
+    -- See Note [Zero-extending sub-word signed results]
   | signedOp mop
   , res_ty <- machOpResultType platform mop args
   , not $ isFloatType res_ty -- only integer operations, not MO_SF_Conv
@@ -634,14 +638,15 @@ staticLitsToWords platform = go . foldMap decomposeMultiWord
     -- single-word (or smaller) literals.
     decomposeMultiWord :: CmmLit -> [CmmLit]
     decomposeMultiWord (CmmFloat n W64)
-      -- This will produce a W64 integer, which will then be broken up further
-      -- on the next iteration on 32-bit platforms.
-      = [doubleToWord64 n]
+      | W32 <- wordWidth platform = decomposeMultiWord (doubleToWord64 n)
+      | otherwise = [doubleToWord64 n]
     decomposeMultiWord (CmmFloat n W32)
       = [floatToWord32 n]
     decomposeMultiWord (CmmInt n W64)
       | W32 <- wordWidth platform
-      = [CmmInt hi W32, CmmInt lo W32]
+      = case platformByteOrder platform of
+          BigEndian -> [CmmInt hi W32, CmmInt lo W32]
+          LittleEndian -> [CmmInt lo W32, CmmInt hi W32]
       where
         hi = n `shiftR` 32
         lo = n .&. 0xffffffff
@@ -926,23 +931,27 @@ pprCallishMachOp_for_C mop
         MO_F32_Fabs     -> text "fabsf"
         MO_ReadBarrier  -> text "load_load_barrier"
         MO_WriteBarrier -> text "write_barrier"
-        MO_Memcpy _     -> text "memcpy"
-        MO_Memset _     -> text "memset"
-        MO_Memmove _    -> text "memmove"
-        MO_Memcmp _     -> text "memcmp"
-        (MO_BSwap w)    -> ptext (sLit $ bSwapLabel w)
-        (MO_BRev w)     -> ptext (sLit $ bRevLabel w)
-        (MO_PopCnt w)   -> ptext (sLit $ popCntLabel w)
-        (MO_Pext w)     -> ptext (sLit $ pextLabel w)
-        (MO_Pdep w)     -> ptext (sLit $ pdepLabel w)
-        (MO_Clz w)      -> ptext (sLit $ clzLabel w)
-        (MO_Ctz w)      -> ptext (sLit $ ctzLabel w)
-        (MO_AtomicRMW w amop) -> ptext (sLit $ atomicRMWLabel w amop)
-        (MO_Cmpxchg w)  -> ptext (sLit $ cmpxchgLabel w)
-        (MO_Xchg w)     -> ptext (sLit $ xchgLabel w)
-        (MO_AtomicRead w)  -> ptext (sLit $ atomicReadLabel w)
-        (MO_AtomicWrite w) -> ptext (sLit $ atomicWriteLabel w)
-        (MO_UF_Conv w)  -> ptext (sLit $ word2FloatLabel w)
+        MO_Memcpy _     -> text "__builtin_memcpy"
+        MO_Memset _     -> text "__builtin_memset"
+        MO_Memmove _    -> text "__builtin_memmove"
+        MO_Memcmp _     -> text "__builtin_memcmp"
+
+        MO_SuspendThread -> text "suspendThread"
+        MO_ResumeThread  -> text "resumeThread"
+
+        MO_BSwap w          -> ftext (bSwapLabel w)
+        MO_BRev w           -> ftext (bRevLabel w)
+        MO_PopCnt w         -> ftext (popCntLabel w)
+        MO_Pext w           -> ftext (pextLabel w)
+        MO_Pdep w           -> ftext (pdepLabel w)
+        MO_Clz w            -> ftext (clzLabel w)
+        MO_Ctz w            -> ftext (ctzLabel w)
+        MO_AtomicRMW w amop -> ftext (atomicRMWLabel w amop)
+        MO_Cmpxchg w        -> ftext (cmpxchgLabel w)
+        MO_Xchg w           -> ftext (xchgLabel w)
+        MO_AtomicRead w     -> ftext (atomicReadLabel w)
+        MO_AtomicWrite w    -> ftext (atomicWriteLabel w)
+        MO_UF_Conv w        -> ftext (word2FloatLabel w)
 
         MO_S_Mul2     {} -> unsupported
         MO_S_QuotRem  {} -> unsupported
@@ -1003,7 +1012,7 @@ mkJMP_ i = text "JMP_" <> parens i
 mkFN_  i = text "FN_"  <> parens i -- externally visible function
 mkIF_  i = text "IF_"  <> parens i -- locally visible
 
--- from includes/Stg.h
+-- from rts/include/Stg.h
 --
 mkC_,mkW_,mkP_ :: SDoc
 
@@ -1107,7 +1116,7 @@ pprReg r = case r of
 
 pprAsPtrReg :: CmmReg -> SDoc
 pprAsPtrReg (CmmGlobal (VanillaReg n gcp))
-  = WARN( gcp /= VGcPtr, ppr n ) char 'R' <> int n <> text ".p"
+  = warnPprTrace (gcp /= VGcPtr) "pprAsPtrReg" (ppr n) $ char 'R' <> int n <> text ".p"
 pprAsPtrReg other_reg = pprReg other_reg
 
 pprGlobalReg :: GlobalReg -> SDoc
@@ -1310,8 +1319,6 @@ cLoad platform expr rep
           bewareLoadStoreAlignment ArchMipsel   = True
           bewareLoadStoreAlignment (ArchARM {}) = True
           bewareLoadStoreAlignment ArchAArch64  = True
-          bewareLoadStoreAlignment ArchSPARC    = True
-          bewareLoadStoreAlignment ArchSPARC64  = True
           -- Pessimistically assume that they will also cause problems
           -- on unknown arches
           bewareLoadStoreAlignment ArchUnknown  = True
@@ -1404,28 +1411,10 @@ pprStringInCStyle s = doubleQuotes (text (concatMap charToC (BS.unpack s)))
 -- can safely initialise to static locations.
 
 floatToWord32 :: Rational -> CmmLit
-floatToWord32 r
-  = runST $ do
-        arr <- newArray_ ((0::Int),0)
-        writeArray arr 0 (fromRational r)
-        arr' <- castFloatToWord32Array arr
-        w32 <- readArray arr' 0
-        return (CmmInt (toInteger w32) W32)
-  where
-    castFloatToWord32Array :: STUArray s Int Float -> ST s (STUArray s Int Word32)
-    castFloatToWord32Array = U.castSTUArray
+floatToWord32 r = CmmInt (toInteger (castFloatToWord32 (fromRational r))) W32
 
 doubleToWord64 :: Rational -> CmmLit
-doubleToWord64 r
-  = runST $ do
-        arr <- newArray_ ((0::Int),1)
-        writeArray arr 0 (fromRational r)
-        arr' <- castDoubleToWord64Array arr
-        w64 <- readArray arr' 0
-        return $ CmmInt (toInteger w64) W64
-  where
-    castDoubleToWord64Array :: STUArray s Int Double -> ST s (STUArray s Int Word64)
-    castDoubleToWord64Array = U.castSTUArray
+doubleToWord64 r = CmmInt (toInteger (castDoubleToWord64 (fromRational r))) W64
 
 
 -- ---------------------------------------------------------------------------
@@ -1502,3 +1491,19 @@ pprHexVal platform w rep = parens ctype <> rawlit
              (q,r) = w' `quotRem` 16
              dig | r < 10    = char (chr (fromInteger r + ord '0'))
                  | otherwise = char (chr (fromInteger r - 10 + ord 'a'))
+
+-- | Construct a constructor/finalizer function. Instead of emitting a
+-- initializer/finalizer array we rather just emit a single function, annotated
+-- with the appropriate C attribute, which then calls each of the initializers.
+pprCtorArray :: Platform -> InitOrFini -> [CLabel] -> SDoc
+pprCtorArray platform initOrFini lbls =
+       decls
+    <> text "static __attribute__((" <> attribute <> text "))"
+    <> text "void _hs_" <> attribute <> text "()"
+    <> braces body
+  where
+    body = vcat [ pprCLabel platform CStyle lbl <> text " ();" | lbl <- lbls ]
+    decls = vcat [ text "void" <+> pprCLabel platform CStyle lbl <> text " (void);" | lbl <- lbls ]
+    attribute = case initOrFini of
+                  IsInitArray -> text "constructor"
+                  IsFiniArray -> text "destructor"
