@@ -4,14 +4,13 @@
 \section[CoreRules]{Rewrite rules}
 -}
 
-{-# LANGUAGE CPP #-}
 
 -- | Functions for collecting together and applying rewrite rules to a module.
 -- The 'CoreRule' datatype itself is declared elsewhere.
 module GHC.Core.Rules (
         -- ** Constructing
         emptyRuleBase, mkRuleBase, extendRuleBaseList,
-        unionRuleBase, pprRuleBase,
+        pprRuleBase, extendRuleEnv,
 
         -- ** Checking rule applications
         ruleCheckProgram,
@@ -26,29 +25,35 @@ module GHC.Core.Rules (
         lookupRule, mkRule, roughTopNames, initRuleOpts
     ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
-import GHC.Core         -- All of it
+import GHC.Driver.Session      ( DynFlags, gopt, targetPlatform, homeUnitId_ )
+import GHC.Driver.Flags
+
 import GHC.Unit.Types    ( primUnitId, bignumUnitId )
 import GHC.Unit.Module   ( Module )
 import GHC.Unit.Module.Env
+
+import GHC.Core         -- All of it
 import GHC.Core.Subst
 import GHC.Core.SimpleOpt ( exprIsLambda_maybe )
 import GHC.Core.FVs       ( exprFreeVars, exprsFreeVars, bindFreeVars
                           , rulesFreeVarsDSet, exprsOrphNames, exprFreeVarsList )
-import GHC.Core.Utils     ( exprType, eqExpr, mkTick, mkTicks
+import GHC.Core.Utils     ( exprType, mkTick, mkTicks
                           , stripTicksTopT, stripTicksTopE
                           , isJoinBind, mkCastMCo )
 import GHC.Core.Ppr       ( pprRules )
+import GHC.Core.Unify as Unify ( ruleMatchTyKiX )
 import GHC.Core.Type as Type
    ( Type, TCvSubst, extendTvSubst, extendCvSubst
-   , mkEmptyTCvSubst, getTyVar_maybe, substTy )
-import GHC.Tc.Utils.TcType  ( tcSplitTyConApp_maybe )
-import GHC.Builtin.Types    ( anyTypeOfKind )
+   , mkEmptyTCvSubst, substTy, getTyVar_maybe )
 import GHC.Core.Coercion as Coercion
 import GHC.Core.Tidy     ( tidyRules )
+import GHC.Core.Map.Expr ( eqCoreExpr )
+
+import GHC.Tc.Utils.TcType  ( tcSplitTyConApp_maybe )
+import GHC.Builtin.Types    ( anyTypeOfKind )
+
 import GHC.Types.Id
 import GHC.Types.Id.Info ( RuleInfo( RuleInfo ) )
 import GHC.Types.Var
@@ -59,17 +64,18 @@ import GHC.Types.Name.Set
 import GHC.Types.Name.Env
 import GHC.Types.Unique.FM
 import GHC.Types.Tickish
-import GHC.Core.Unify as Unify ( ruleMatchTyKiX )
 import GHC.Types.Basic
-import GHC.Driver.Session      ( DynFlags, gopt, targetPlatform, homeUnitId_ )
-import GHC.Driver.Ppr
-import GHC.Driver.Flags
-import GHC.Utils.Outputable
-import GHC.Utils.Panic
+
 import GHC.Data.FastString
 import GHC.Data.Maybe
 import GHC.Data.Bag
+
 import GHC.Utils.Misc as Utils
+import GHC.Utils.Trace
+import GHC.Utils.Outputable
+import GHC.Utils.Panic
+import GHC.Utils.Constants (debugIsOn)
+
 import Data.List (sortBy, mapAccumL, isPrefixOf)
 import Data.Function    ( on )
 import Control.Monad    ( guard )
@@ -130,14 +136,18 @@ Note [Overall plumbing for rules]
   [NB: we are inconsistent here.  We should do the same for external
   packages, but we don't.  Same for type-class instances.]
 
-* So in the outer simplifier loop, we combine (b-d) into a single
+* So in the outer simplifier loop (simplifyPgmIO), we combine (b & c) into a single
   RuleBase, reading
      (b) from the ModGuts,
      (c) from the GHC.Core.Opt.Monad, and
+  just before doing rule matching we read
      (d) from its mutable variable
-  [Of course this means that we won't see new EPS rules that come in
-  during a single simplifier iteration, but that probably does not
-  matter.]
+  and combine it with the results from (b & c).
+
+  In a single simplifier run new rules can be added into the EPS so it matters
+  to keep an up-to-date view of which rules have been loaded. For examples of
+  where this went wrong and caused cryptic performance regressions seee
+  see T19790 and !6735.
 
 
 ************************************************************************
@@ -307,9 +317,9 @@ rulesOfBinds binds = concatMap (concatMap idCoreRules . bindersOf) binds
 getRules :: RuleEnv -> Id -> [CoreRule]
 -- See Note [Where rules are found]
 getRules (RuleEnv { re_base = rule_base, re_visible_orphs = orphs }) fn
-  = idCoreRules fn ++ filter (ruleIsVisible orphs) imp_rules
+  = idCoreRules fn ++ concatMap imp_rules rule_base
   where
-    imp_rules = lookupNameEnv rule_base (idName fn) `orElse` []
+    imp_rules rb = filter (ruleIsVisible orphs) (lookupNameEnv rb (idName fn) `orElse` [])
 
 ruleIsVisible :: ModuleSet -> CoreRule -> Bool
 ruleIsVisible _ BuiltinRule{} = True
@@ -355,13 +365,13 @@ extendRuleBaseList :: RuleBase -> [CoreRule] -> RuleBase
 extendRuleBaseList rule_base new_guys
   = foldl' extendRuleBase rule_base new_guys
 
-unionRuleBase :: RuleBase -> RuleBase -> RuleBase
-unionRuleBase rb1 rb2 = plusNameEnv_C (++) rb1 rb2
-
 extendRuleBase :: RuleBase -> CoreRule -> RuleBase
 extendRuleBase rule_base rule
   = extendNameEnv_Acc (:) Utils.singleton rule_base (ruleIdName rule) rule
 
+extendRuleEnv :: RuleEnv -> RuleBase -> RuleEnv
+extendRuleEnv (RuleEnv rules orphs) rb = (RuleEnv (rb:rules) orphs)
+
 pprRuleBase :: RuleBase -> SDoc
 pprRuleBase rules = pprUFM rules $ \rss ->
   vcat [ pprRules (tidyRules emptyTidyEnv rs)
@@ -381,8 +391,10 @@ pprRuleBase rules = pprUFM rules $ \rss ->
 -- successful.
 lookupRule :: RuleOpts -> InScopeEnv
            -> (Activation -> Bool)      -- When rule is active
-           -> Id -> [CoreExpr]
-           -> [CoreRule] -> Maybe (CoreRule, CoreExpr)
+           -> Id -- Function head
+           -> [CoreExpr] -- Args
+           -> [CoreRule] -- Rules
+           -> Maybe (CoreRule, CoreExpr)
 
 -- See Note [Extra args in the target]
 -- See comments on matchRule
@@ -394,7 +406,7 @@ lookupRule opts rule_env@(in_scope,_) is_active fn args rules
   where
     rough_args = map roughTopName args
 
-    -- Strip ticks from arguments, see note [Tick annotations in RULE
+    -- Strip ticks from arguments, see Note [Tick annotations in RULE
     -- matching]. We only collect ticks if a rule actually matches -
     -- this matters for performance tests.
     args' = map (stripTicksTopE tickishFloatable) args
@@ -765,10 +777,6 @@ match_exprs renv subst (e1:es1) (e2:es2)
        ; match_exprs renv subst' es1 es2 }
 match_exprs _ _ _ _ = Nothing
 
---      I now think this is probably a bad idea.
---      Should the template (map f xs) match (map g)?  I think not.
---      For a start, in general eta expansion wastes work.
---      SLPJ July 99
 
 {- Note [Casts in the target]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -882,7 +890,7 @@ match renv subst (Type ty1) (Type ty2) _mco
   = match_ty renv subst ty1 ty2
 
 ------------------------ Coercions ---------------------
--- See Note [Coercion argument] for why this isn't really right
+-- See Note [Coercion arguments] for why this isn't really right
 match renv subst (Coercion co1) (Coercion co2) MRefl
   = match_co renv subst co1 co2
   -- The MCo case corresponds to matching  co ~ (co2 |> co3)
@@ -912,7 +920,7 @@ match renv subst (Cast e1 co1) e2 mco
 ------------------------ Literals ---------------------
 match _ subst (Lit lit1) (Lit lit2) mco
   | lit1 == lit2
-  = ASSERT2(isReflMCo mco, ppr mco)
+  = assertPpr (isReflMCo mco) (ppr mco) $
     Just subst
 
 ------------------------ Variables ---------------------
@@ -1233,7 +1241,7 @@ match_tmpl_var renv@(RV { rv_lcl = rn_env, rv_fltR = flt_env })
                 -- e.g. match forall a. (\x-> a x) against (\y. y y)
 
   | Just e1' <- lookupVarEnv id_subst v1'
-  = if eqExpr (rnInScopeSet rn_env) e1' e2'
+  = if eqCoreExpr e1' e2'
     then Just subst
     else Nothing
 
