@@ -21,16 +21,18 @@ just attach noSrcSpan to everything.
 -}
 
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE PatternSynonyms #-}
-{-# LANGUAGE ViewPatterns #-}
-{-# LANGUAGE TypeApplications #-}
+{-# LANGUAGE AllowAmbiguousTypes #-}
 {-# LANGUAGE DataKinds #-}
+{-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE GADTs #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE PatternSynonyms #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TypeApplications #-}
+{-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE ViewPatterns #-}
 
 {-# OPTIONS_GHC -Wno-incomplete-record-updates #-}
 
@@ -39,12 +41,12 @@ module GHC.Hs.Utils(
   mkHsPar, mkHsApp, mkHsAppWith, mkHsApps, mkHsAppsWith,
   mkHsAppType, mkHsAppTypes, mkHsCaseAlt,
   mkSimpleMatch, unguardedGRHSs, unguardedRHS,
-  mkMatchGroup, mkMatch, mkPrefixFunRhs, mkHsLam, mkHsIf,
+  mkMatchGroup, mkLamCaseMatchGroup, mkMatch, mkPrefixFunRhs, mkHsLam, mkHsIf,
   mkHsWrap, mkLHsWrap, mkHsWrapCo, mkHsWrapCoR, mkLHsWrapCo,
   mkHsDictLet, mkHsLams,
   mkHsOpApp, mkHsDo, mkHsDoAnns, mkHsComp, mkHsCompAnns, mkHsWrapPat, mkHsWrapPatCo,
   mkLHsPar, mkHsCmdWrap, mkLHsCmdWrap,
-  mkHsCmdIf,
+  mkHsCmdIf, mkConLikeTc,
 
   nlHsTyApp, nlHsTyApps, nlHsVar, nl_HsVar, nlHsDataCon,
   nlHsLit, nlHsApp, nlHsApps, nlHsSyntaxApps,
@@ -111,8 +113,6 @@ module GHC.Hs.Utils(
   lStmtsImplicits, hsValBindsImplicits, lPatImplicits
   ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
 import GHC.Hs.Decls
@@ -153,8 +153,6 @@ import GHC.Utils.Panic
 import Data.Either
 import Data.Function
 import Data.List ( partition, deleteBy )
-import Data.Proxy
-import Data.Data (Data)
 
 {-
 ************************************************************************
@@ -170,13 +168,13 @@ just attach 'noSrcSpan' to everything.
 
 -- | @e => (e)@
 mkHsPar :: LHsExpr (GhcPass id) -> LHsExpr (GhcPass id)
-mkHsPar e = L (getLoc e) (HsPar noAnn e)
+mkHsPar e = L (getLoc e) (gHsPar e)
 
 mkSimpleMatch :: (Anno (Match (GhcPass p) (LocatedA (body (GhcPass p))))
                         ~ SrcSpanAnnA,
                   Anno (GRHS (GhcPass p) (LocatedA (body (GhcPass p))))
-                        ~ SrcSpan)
-              => HsMatchContext (NoGhcTc (GhcPass p))
+                        ~ SrcAnn NoEpAnns)
+              => HsMatchContext (GhcPass p)
               -> [LPat (GhcPass p)] -> LocatedA (body (GhcPass p))
               -> LMatch (GhcPass p) (LocatedA (body (GhcPass p)))
 mkSimpleMatch ctxt pats rhs
@@ -189,17 +187,17 @@ mkSimpleMatch ctxt pats rhs
                 (pat:_) -> combineSrcSpansA (getLoc pat) (getLoc rhs)
 
 unguardedGRHSs :: Anno (GRHS (GhcPass p) (LocatedA (body (GhcPass p))))
-                     ~ SrcSpan
+                     ~ SrcAnn NoEpAnns
                => SrcSpan -> LocatedA (body (GhcPass p)) -> EpAnn GrhsAnn
                -> GRHSs (GhcPass p) (LocatedA (body (GhcPass p)))
 unguardedGRHSs loc rhs an
   = GRHSs emptyComments (unguardedRHS an loc rhs) emptyLocalBinds
 
 unguardedRHS :: Anno (GRHS (GhcPass p) (LocatedA (body (GhcPass p))))
-                     ~ SrcSpan
+                     ~ SrcAnn NoEpAnns
              => EpAnn GrhsAnn -> SrcSpan -> LocatedA (body (GhcPass p))
              -> [LGRHS (GhcPass p) (LocatedA (body (GhcPass p)))]
-unguardedRHS an loc rhs = [L loc (GRHS an [] rhs)]
+unguardedRHS an loc rhs = [L (noAnnSrcSpan loc) (GRHS an [] rhs)]
 
 type AnnoBody p body
   = ( XMG (GhcPass p) (LocatedA (body (GhcPass p))) ~ NoExtField
@@ -215,6 +213,15 @@ mkMatchGroup origin matches = MG { mg_ext = noExtField
                                  , mg_alts = matches
                                  , mg_origin = origin }
 
+mkLamCaseMatchGroup :: AnnoBody p body
+                    => Origin
+                    -> LamCaseVariant
+                    -> LocatedL [LocatedA (Match (GhcPass p) (LocatedA (body (GhcPass p))))]
+                    -> MatchGroup (GhcPass p) (LocatedA (body (GhcPass p)))
+mkLamCaseMatchGroup origin lc_variant (L l matches)
+  = mkMatchGroup origin (L l $ map fixCtxt matches)
+  where fixCtxt (L a match) = L a match{m_ctxt = LamCaseAlt lc_variant}
+
 mkLocatedList :: Semigroup a
   => [GenLocated (SrcAnn a) e2] -> LocatedAn an [GenLocated (SrcAnn a) e2]
 mkLocatedList [] = noLocA []
@@ -267,7 +274,7 @@ mkHsLams tyvars dicts expr = mkLHsWrap (mkWpTyLams tyvars
 -- |A simple case alternative with a single pattern, no binds, no guards;
 -- pre-typechecking
 mkHsCaseAlt :: (Anno (GRHS (GhcPass p) (LocatedA (body (GhcPass p))))
-                     ~ SrcSpan,
+                     ~ SrcAnn NoEpAnns,
                  Anno (Match (GhcPass p) (LocatedA (body (GhcPass p))))
                         ~ SrcSpanAnnA)
             => LPat (GhcPass p) -> (LocatedA (body (GhcPass p)))
@@ -286,17 +293,13 @@ nlHsTyApps fun_id tys xs = foldl' nlHsApp (nlHsTyApp fun_id tys) xs
 -- | Wrap in parens if @'hsExprNeedsParens' appPrec@ says it needs them
 -- So @f x@ becomes @(f x)@, but @3@ stays as @3@.
 mkLHsPar :: IsPass id => LHsExpr (GhcPass id) -> LHsExpr (GhcPass id)
-mkLHsPar le@(L loc e)
-  | hsExprNeedsParens appPrec e = L loc (HsPar noAnn le)
-  | otherwise                   = le
+mkLHsPar = parenthesizeHsExpr appPrec
 
 mkParPat :: IsPass p => LPat (GhcPass p) -> LPat (GhcPass p)
-mkParPat lp@(L loc p)
-  | patNeedsParens appPrec p = L loc (ParPat noAnn lp)
-  | otherwise                = lp
+mkParPat = parenthesizePat appPrec
 
 nlParPat :: LPat (GhcPass name) -> LPat (GhcPass name)
-nlParPat p = noLocA (ParPat noAnn p)
+nlParPat p = noLocA (gParPat p)
 
 -------------------------------
 -- These are the bits of syntax that contain rebindable names
@@ -305,17 +308,17 @@ nlParPat p = noLocA (ParPat noAnn p)
 mkHsIntegral   :: IntegralLit -> HsOverLit GhcPs
 mkHsFractional :: FractionalLit -> HsOverLit GhcPs
 mkHsIsString   :: SourceText -> FastString -> HsOverLit GhcPs
-mkHsDo         :: HsStmtContext GhcRn -> LocatedL [ExprLStmt GhcPs] -> HsExpr GhcPs
-mkHsDoAnns     :: HsStmtContext GhcRn -> LocatedL [ExprLStmt GhcPs] -> EpAnn AnnList -> HsExpr GhcPs
-mkHsComp       :: HsStmtContext GhcRn -> [ExprLStmt GhcPs] -> LHsExpr GhcPs
+mkHsDo         :: HsDoFlavour -> LocatedL [ExprLStmt GhcPs] -> HsExpr GhcPs
+mkHsDoAnns     :: HsDoFlavour -> LocatedL [ExprLStmt GhcPs] -> EpAnn AnnList -> HsExpr GhcPs
+mkHsComp       :: HsDoFlavour -> [ExprLStmt GhcPs] -> LHsExpr GhcPs
                -> HsExpr GhcPs
-mkHsCompAnns   :: HsStmtContext GhcRn -> [ExprLStmt GhcPs] -> LHsExpr GhcPs
+mkHsCompAnns   :: HsDoFlavour -> [ExprLStmt GhcPs] -> LHsExpr GhcPs
                -> EpAnn AnnList
                -> HsExpr GhcPs
 
-mkNPat      :: Located (HsOverLit GhcPs) -> Maybe (SyntaxExpr GhcPs) -> EpAnn [AddEpAnn]
+mkNPat      :: LocatedAn NoEpAnns (HsOverLit GhcPs) -> Maybe (SyntaxExpr GhcPs) -> EpAnn [AddEpAnn]
             -> Pat GhcPs
-mkNPlusKPat :: LocatedN RdrName -> Located (HsOverLit GhcPs) -> EpAnn EpaLocation
+mkNPlusKPat :: LocatedN RdrName -> LocatedAn NoEpAnns (HsOverLit GhcPs) -> EpAnn EpaLocation
             -> Pat GhcPs
 
 -- NB: The following functions all use noSyntaxExpr: the generated expressions
@@ -351,9 +354,9 @@ mkRecStmt        :: (Anno [GenLocated
                  -> StmtLR (GhcPass idL) GhcPs bodyR
 
 
-mkHsIntegral     i  = OverLit noExtField (HsIntegral       i) noExpr
-mkHsFractional   f  = OverLit noExtField (HsFractional     f) noExpr
-mkHsIsString src s  = OverLit noExtField (HsIsString   src s) noExpr
+mkHsIntegral     i  = OverLit noExtField (HsIntegral       i)
+mkHsFractional   f  = OverLit noExtField (HsFractional     f)
+mkHsIsString src s  = OverLit noExtField (HsIsString   src s)
 
 mkHsDo     ctxt stmts      = HsDo noAnn ctxt stmts
 mkHsDoAnns ctxt stmts anns = HsDo anns  ctxt stmts
@@ -468,6 +471,8 @@ mkHsStringPrimLit fs = HsStringPrim NoSourceText (bytesFS fs)
 mkHsCharPrimLit :: Char -> HsLit (GhcPass p)
 mkHsCharPrimLit c = HsChar NoSourceText c
 
+mkConLikeTc :: ConLike -> HsExpr GhcTc
+mkConLikeTc con = XExpr (ConLikeTc con [] [])
 
 {-
 ************************************************************************
@@ -487,7 +492,7 @@ nl_HsVar n = HsVar noExtField (noLocA n)
 
 -- | NB: Only for 'LHsExpr' 'Id'.
 nlHsDataCon :: DataCon -> LHsExpr GhcTc
-nlHsDataCon con = noLocA (HsConLikeOut noExtField (RealDataCon con))
+nlHsDataCon con = noLocA (mkConLikeTc (RealDataCon con))
 
 nlHsLit :: HsLit (GhcPass p) -> LHsExpr (GhcPass p)
 nlHsLit n = noLocA (HsLit noComments n)
@@ -579,7 +584,7 @@ nlWildPat  = noLocA (WildPat noExtField )
 nlWildPatName :: LPat GhcRn
 nlWildPatName  = noLocA (WildPat noExtField )
 
-nlHsDo :: HsStmtContext GhcRn -> [LStmt GhcPs (LHsExpr GhcPs)]
+nlHsDo :: HsDoFlavour -> [LStmt GhcPs (LHsExpr GhcPs)]
        -> LHsExpr GhcPs
 nlHsDo ctxt stmts = noLocA (mkHsDo ctxt (noLocA stmts))
 
@@ -594,7 +599,7 @@ nlList   :: [LHsExpr GhcPs] -> LHsExpr GhcPs
 
 -- AZ:Is this used?
 nlHsLam match = noLocA (HsLam noExtField (mkMatchGroup Generated (noLocA [match])))
-nlHsPar e     = noLocA (HsPar noAnn e)
+nlHsPar e     = noLocA (gHsPar e)
 
 -- nlHsIf should generate if-expressions which are NOT subject to
 -- RebindableSyntax, so the first field of HsIf is False. (#12080)
@@ -607,24 +612,25 @@ nlList exprs          = noLocA (ExplicitList noAnn exprs)
 
 nlHsAppTy :: LHsType (GhcPass p) -> LHsType (GhcPass p) -> LHsType (GhcPass p)
 nlHsTyVar :: IsSrcSpanAnn p a
-          => IdP (GhcPass p)                            -> LHsType (GhcPass p)
+          => PromotionFlag -> IdP (GhcPass p)           -> LHsType (GhcPass p)
 nlHsFunTy :: LHsType (GhcPass p) -> LHsType (GhcPass p) -> LHsType (GhcPass p)
 nlHsParTy :: LHsType (GhcPass p)                        -> LHsType (GhcPass p)
 
 nlHsAppTy f t = noLocA (HsAppTy noExtField f (parenthesizeHsType appPrec t))
-nlHsTyVar x   = noLocA (HsTyVar noAnn NotPromoted (noLocA x))
-nlHsFunTy a b = noLocA (HsFunTy noAnn (HsUnrestrictedArrow NormalSyntax) (parenthesizeHsType funPrec a) b)
+nlHsTyVar p x = noLocA (HsTyVar noAnn p (noLocA x))
+nlHsFunTy a b = noLocA (HsFunTy noAnn (HsUnrestrictedArrow noHsUniTok) (parenthesizeHsType funPrec a) b)
 nlHsParTy t   = noLocA (HsParTy noAnn t)
 
 nlHsTyConApp :: IsSrcSpanAnn p a
-             => LexicalFixity -> IdP (GhcPass p)
+             => PromotionFlag
+             -> LexicalFixity -> IdP (GhcPass p)
              -> [LHsTypeArg (GhcPass p)] -> LHsType (GhcPass p)
-nlHsTyConApp fixity tycon tys
+nlHsTyConApp prom fixity tycon tys
   | Infix <- fixity
   , HsValArg ty1 : HsValArg ty2 : rest <- tys
-  = foldl' mk_app (noLocA $ HsOpTy noExtField ty1 (noLocA tycon) ty2) rest
+  = foldl' mk_app (noLocA $ HsOpTy noAnn prom ty1 (noLocA tycon) ty2) rest
   | otherwise
-  = foldl' mk_app (nlHsTyVar tycon) tys
+  = foldl' mk_app (nlHsTyVar prom tycon) tys
   where
     mk_app :: LHsType (GhcPass p) -> LHsTypeArg (GhcPass p) -> LHsType (GhcPass p)
     mk_app fun@(L _ (HsOpTy {})) arg = mk_app (noLocA $ HsParTy noAnn fun) arg
@@ -790,13 +796,9 @@ mkClassOpSigs sigs
 mkLHsWrap :: HsWrapper -> LHsExpr GhcTc -> LHsExpr GhcTc
 mkLHsWrap co_fn (L loc e) = L loc (mkHsWrap co_fn e)
 
--- | Avoid @'HsWrap' co1 ('HsWrap' co2 _)@ and @'HsWrap' co1 ('HsPar' _ _)@
--- See Note [Detecting forced eta expansion] in "GHC.HsToCore.Expr"
 mkHsWrap :: HsWrapper -> HsExpr GhcTc -> HsExpr GhcTc
-mkHsWrap co_fn e | isIdHsWrapper co_fn   = e
-mkHsWrap co_fn (XExpr (WrapExpr (HsWrap co_fn' e))) = mkHsWrap (co_fn <.> co_fn') e
-mkHsWrap co_fn (HsPar x (L l e))                = HsPar x (L l (mkHsWrap co_fn e))
-mkHsWrap co_fn e                                = XExpr (WrapExpr $ HsWrap co_fn e)
+mkHsWrap co_fn e | isIdHsWrapper co_fn = e
+mkHsWrap co_fn e                       = XExpr (WrapExpr $ HsWrap co_fn e)
 
 mkHsWrapCo :: TcCoercionN   -- A Nominal coercion  a ~N b
            -> HsExpr GhcTc -> HsExpr GhcTc
@@ -880,7 +882,7 @@ isInfixFunBind (FunBind { fun_matches = MG _ matches _ })
 isInfixFunBind _ = False
 
 -- |Return the 'SrcSpan' encompassing the contents of any enclosed binds
-spanHsLocaLBinds :: (Data (HsLocalBinds (GhcPass p))) => HsLocalBinds (GhcPass p) -> SrcSpan
+spanHsLocaLBinds :: HsLocalBinds (GhcPass p) -> SrcSpan
 spanHsLocaLBinds (EmptyLocalBinds _) = noSrcSpan
 spanHsLocaLBinds (HsValBinds _ (ValBinds _ bs sigs))
   = foldr combineSrcSpans noSrcSpan (bsSpans ++ sigsSpans)
@@ -917,7 +919,7 @@ mkPrefixFunRhs n = FunRhs { mc_fun = n
 
 ------------
 mkMatch :: forall p. IsPass p
-        => HsMatchContext (NoGhcTc (GhcPass p))
+        => HsMatchContext (GhcPass p)
         -> [LPat (GhcPass p)]
         -> LHsExpr (GhcPass p)
         -> HsLocalBinds (GhcPass p)
@@ -925,13 +927,8 @@ mkMatch :: forall p. IsPass p
 mkMatch ctxt pats expr binds
   = noLocA (Match { m_ext   = noAnn
                   , m_ctxt  = ctxt
-                  , m_pats  = map paren pats
+                  , m_pats  = map mkParPat pats
                   , m_grhss = GRHSs emptyComments (unguardedRHS noAnn noSrcSpan expr) binds })
-  where
-    paren :: LPat (GhcPass p) -> LPat (GhcPass p)
-    paren lp@(L l p)
-      | patNeedsParens appPrec p = L l (ParPat noAnn lp)
-      | otherwise                = lp
 
 {-
 ************************************************************************
@@ -992,7 +989,7 @@ BUT we have a special case when abs_sig is true;
 -- information, see Note [Strict binds checks] is GHC.HsToCore.Binds.
 isUnliftedHsBind :: HsBind GhcTc -> Bool  -- works only over typechecked binds
 isUnliftedHsBind bind
-  | AbsBinds { abs_exports = exports, abs_sig = has_sig } <- bind
+  | XHsBindsLR (AbsBinds { abs_exports = exports, abs_sig = has_sig }) <- bind
   = if has_sig
     then any (is_unlifted_id . abe_poly) exports
     else any (is_unlifted_id . abe_mono) exports
@@ -1004,10 +1001,12 @@ isUnliftedHsBind bind
   = any is_unlifted_id (collectHsBindBinders CollNoDictBinders bind)
   where
     is_unlifted_id id = isUnliftedType (idType id)
+      -- bindings always have a fixed RuntimeRep, so it's OK
+      -- to call isUnliftedType here
 
 -- | Is a binding a strict variable or pattern bind (e.g. @!x = ...@)?
 isBangedHsBind :: HsBind GhcTc -> Bool
-isBangedHsBind (AbsBinds { abs_binds = binds })
+isBangedHsBind (XHsBindsLR (AbsBinds { abs_binds = binds }))
   = anyBag (isBangedHsBind . unLoc) binds
 isBangedHsBind (FunBind {fun_matches = matches})
   | [L _ match] <- unLoc $ mg_alts matches
@@ -1037,7 +1036,7 @@ collectHsIdBinders flag = collect_hs_val_binders True flag
 
 collectHsValBinders :: CollectPass (GhcPass idL)
                     => CollectFlag (GhcPass idL)
-                    -> HsValBindsLR (GhcPass idL) (GhcPass idR)
+                    -> HsValBindsLR (GhcPass idL) idR
                     -> [IdP (GhcPass idL)]
 collectHsValBinders flag = collect_hs_val_binders False flag
 
@@ -1064,7 +1063,7 @@ collectHsBindListBinders flag = foldr (collect_bind False flag . unXRec @p) []
 collect_hs_val_binders :: CollectPass (GhcPass idL)
                        => Bool
                        -> CollectFlag (GhcPass idL)
-                       -> HsValBindsLR (GhcPass idL) (GhcPass idR)
+                       -> HsValBindsLR (GhcPass idL) idR
                        -> [IdP (GhcPass idL)]
 collect_hs_val_binders ps flag = \case
     ValBinds _ binds _              -> collect_binds ps flag binds []
@@ -1092,18 +1091,15 @@ collect_bind :: forall p idR. CollectPass p
              -> HsBindLR p idR
              -> [IdP p]
              -> [IdP p]
-collect_bind _ flag (PatBind { pat_lhs = p })           acc = collect_lpat flag p acc
-collect_bind _ _ (FunBind { fun_id = f })            acc = unXRec @p f : acc
-collect_bind _ _ (VarBind { var_id = f })            acc = f : acc
-collect_bind _ _ (AbsBinds { abs_exports = dbinds }) acc = map abe_poly dbinds ++ acc
-        -- I don't think we want the binders from the abe_binds
-
-        -- binding (hence see AbsBinds) is in zonking in GHC.Tc.Utils.Zonk
+collect_bind _ _    (FunBind { fun_id = f })         acc = unXRec @p f : acc
+collect_bind _ flag (PatBind { pat_lhs = p })        acc = collect_lpat flag p acc
+collect_bind _ _    (VarBind { var_id = f })         acc = f : acc
 collect_bind omitPatSyn _ (PatSynBind _ (PSB { psb_id = ps })) acc
   | omitPatSyn                  = acc
   | otherwise                   = unXRec @p ps : acc
 collect_bind _ _ (PatSynBind _ (XPatSynBind _)) acc = acc
-collect_bind _ _ (XHsBindsLR _) acc = acc
+collect_bind _ _ (XHsBindsLR b) acc = collectXXHsBindsLR @p @idR b acc
+
 
 collectMethodBinders :: forall idL idR. UnXRec idL => LHsBindsLR idL idR -> [LIdP idL]
 -- ^ Used exclusively for the bindings of an instance decl which are all
@@ -1124,14 +1120,14 @@ collectLStmtsBinders
 collectLStmtsBinders flag = concatMap (collectLStmtBinders flag)
 
 collectStmtsBinders
-  :: (CollectPass (GhcPass idL))
+  :: CollectPass (GhcPass idL)
   => CollectFlag (GhcPass idL)
   -> [StmtLR (GhcPass idL) (GhcPass idR) body]
   -> [IdP (GhcPass idL)]
 collectStmtsBinders flag = concatMap (collectStmtBinders flag)
 
 collectLStmtBinders
-  :: (CollectPass (GhcPass idL))
+  :: CollectPass (GhcPass idL)
   => CollectFlag (GhcPass idL)
   -> LStmtLR (GhcPass idL) (GhcPass idR) body
   -> [IdP (GhcPass idL)]
@@ -1190,7 +1186,7 @@ data CollectFlag p where
     -- | Collect evidence binders
     CollWithDictBinders :: CollectFlag GhcTc
 
-collect_lpat :: forall p. (CollectPass p)
+collect_lpat :: forall p. CollectPass p
              => CollectFlag p
              -> LPat p
              -> [IdP p]
@@ -1209,7 +1205,7 @@ collect_pat flag pat bndrs = case pat of
   BangPat _ pat         -> collect_lpat flag pat bndrs
   AsPat _ a pat         -> unXRec @p a : collect_lpat flag pat bndrs
   ViewPat _ _ pat       -> collect_lpat flag pat bndrs
-  ParPat _ pat          -> collect_lpat flag pat bndrs
+  ParPat _ _ pat _      -> collect_lpat flag pat bndrs
   ListPat _ pats        -> foldr (collect_lpat flag) bndrs pats
   TuplePat _ pats _     -> foldr (collect_lpat flag) bndrs pats
   SumPat _ pat _ _      -> collect_lpat flag pat bndrs
@@ -1217,7 +1213,7 @@ collect_pat flag pat bndrs = case pat of
   NPat {}               -> bndrs
   NPlusKPat _ n _ _ _ _ -> unXRec @p n : bndrs
   SigPat _ pat _        -> collect_lpat flag pat bndrs
-  XPat ext              -> collectXXPat (Proxy @p) flag ext bndrs
+  XPat ext              -> collectXXPat @p flag ext bndrs
   SplicePat _ (HsSpliced _ _ (HsSplicedPat pat))
                         -> collect_pat flag pat bndrs
   SplicePat _ _         -> bndrs
@@ -1244,14 +1240,29 @@ add_ev_bndr (EvBind { eb_lhs = b }) bs | isId b    = b:bs
 -- In particular, Haddock already makes use of this, with an instance for its 'DocNameI' pass so that
 -- it can reuse the code in GHC for collecting binders.
 class UnXRec p => CollectPass p where
-  collectXXPat :: Proxy p -> CollectFlag p -> XXPat p -> [IdP p] -> [IdP p]
+  collectXXPat :: CollectFlag p -> XXPat p -> [IdP p] -> [IdP p]
+  collectXXHsBindsLR :: forall pR. XXHsBindsLR p pR -> [IdP p] -> [IdP p]
 
 instance IsPass p => CollectPass (GhcPass p) where
-  collectXXPat _ flag ext =
+  collectXXPat flag ext =
     case ghcPass @p of
-      GhcTc -> let CoPat _ pat _ = ext in collect_pat flag pat
-      GhcRn -> noExtCon ext
-      GhcPs -> noExtCon ext
+      GhcPs -> dataConCantHappen ext
+      GhcRn
+        | HsPatExpanded _ pat <- ext
+        -> collect_pat flag pat
+      GhcTc -> case ext of
+        CoPat _ pat _      -> collect_pat flag pat
+        ExpansionPat _ pat -> collect_pat flag pat
+  collectXXHsBindsLR ext =
+    case ghcPass @p of
+      GhcPs -> dataConCantHappen ext
+      GhcRn -> dataConCantHappen ext
+      GhcTc -> case ext of
+        AbsBinds { abs_exports = dbinds } -> (map abe_poly dbinds ++)
+        -- I don't think we want the binders from the abe_binds
+
+        -- binding (hence see AbsBinds) is in zonking in GHC.Tc.Utils.Zonk
+
 
 {-
 Note [Dictionary binders in ConPatOut]
@@ -1342,7 +1353,7 @@ hsTyClForeignBinders tycl_decls foreign_decls
          foldMap (foldMap hsLInstDeclBinders . group_instds) tycl_decls)
   where
     getSelectorNames :: ([LocatedA Name], [LFieldOcc GhcRn]) -> [Name]
-    getSelectorNames (ns, fs) = map unLoc ns ++ map (extFieldOcc . unLoc) fs
+    getSelectorNames (ns, fs) = map unLoc ns ++ map (foExt . unLoc) fs
 
 -------------------
 hsLTyClDeclBinders :: IsPass p
@@ -1482,7 +1493,7 @@ hsConDeclsBinders cons
 
     get_flds_gadt :: Seen p -> HsConDeclGADTDetails (GhcPass p)
                   -> (Seen p, [LFieldOcc (GhcPass p)])
-    get_flds_gadt remSeen (RecConGADT flds) = get_flds remSeen flds
+    get_flds_gadt remSeen (RecConGADT flds _) = get_flds remSeen flds
     get_flds_gadt remSeen _ = (remSeen, [])
 
     get_flds :: Seen p -> LocatedL [LConDeclField (GhcPass p)]
@@ -1491,7 +1502,7 @@ hsConDeclsBinders cons
        where
           fld_names = remSeen (concatMap (cd_fld_names . unLoc) (unLoc flds))
           remSeen' = foldr (.) remSeen
-                               [deleteBy ((==) `on` unLoc . rdrNameFieldOcc . unLoc) v
+                               [deleteBy ((==) `on` unLoc . foLabel . unLoc) v
                                | v <- fld_names]
 
 {-
@@ -1584,7 +1595,7 @@ lPatImplicits = hs_lpat
     hs_pat (BangPat _ pat)      = hs_lpat pat
     hs_pat (AsPat _ _ pat)      = hs_lpat pat
     hs_pat (ViewPat _ _ pat)    = hs_lpat pat
-    hs_pat (ParPat _ pat)       = hs_lpat pat
+    hs_pat (ParPat _ _ pat _)   = hs_lpat pat
     hs_pat (ListPat _ pats)     = hs_lpats pats
     hs_pat (TuplePat _ pats _)  = hs_lpats pats
 
@@ -1600,8 +1611,8 @@ lPatImplicits = hs_lpat
       [(err_loc, collectPatsBinders CollNoDictBinders implicit_pats) | Just{} <- [rec_dotdot fs] ]
         ++ hs_lpats explicit_pats
 
-      where implicit_pats = map (hsRecFieldArg . unLoc) implicit
-            explicit_pats = map (hsRecFieldArg . unLoc) explicit
+      where implicit_pats = map (hfbRHS . unLoc) implicit
+            explicit_pats = map (hfbRHS . unLoc) explicit
 
 
             (explicit, implicit) = partitionEithers [if pat_explicit then Left fld else Right fld
