@@ -18,7 +18,7 @@ they may be affected by renaming (which isn't fully worked out yet).
 
 module GHC.Rename.Bind (
    -- Renaming top-level bindings
-   rnTopBindsLHS, rnTopBindsBoot, rnValBindsRHS,
+   rnTopBindsLHS, rnTopBindsLHSBoot, rnTopBindsBoot, rnValBindsRHS,
 
    -- Renaming local bindings
    rnLocalBindsAndThen, rnLocalValBindsLHS, rnLocalValBindsRHS,
@@ -35,19 +35,23 @@ import GHC.Prelude
 import {-# SOURCE #-} GHC.Rename.Expr( rnExpr, rnLExpr, rnStmts )
 
 import GHC.Hs
+import GHC.Tc.Errors.Types
 import GHC.Tc.Utils.Monad
 import GHC.Rename.HsType
 import GHC.Rename.Pat
 import GHC.Rename.Names
 import GHC.Rename.Env
 import GHC.Rename.Fixity
-import GHC.Rename.Utils ( HsDocContext(..), mapFvRn
-                        , checkDupRdrNames, checkDupRdrNamesN, warnUnusedLocalBinds
+import GHC.Rename.Utils ( mapFvRn
+                        , checkDupRdrNames, checkDupRdrNamesN
+                        , warnUnusedLocalBinds
+                        , warnForallIdentifier
                         , checkUnusedRecordWildcard
                         , checkDupAndShadowedNames, bindLocalNamesFV
                         , addNoNestedForallsContextsErr, checkInferredVars )
 import GHC.Driver.Session
 import GHC.Unit.Module
+import GHC.Types.Error
 import GHC.Types.FieldLabel
 import GHC.Types.Name
 import GHC.Types.Name.Env
@@ -187,13 +191,24 @@ rnTopBindsLHS :: MiniFixityEnv
 rnTopBindsLHS fix_env binds
   = rnValBindsLHS (topRecNameMaker fix_env) binds
 
+-- Ensure that a hs-boot file has no top-level bindings.
+rnTopBindsLHSBoot :: MiniFixityEnv
+                  -> HsValBinds GhcPs
+                  -> RnM (HsValBindsLR GhcRn GhcPs)
+rnTopBindsLHSBoot fix_env binds
+  = do  { topBinds <- rnTopBindsLHS fix_env binds
+        ; case topBinds of
+            ValBinds x mbinds sigs ->
+              do  { mapM_ bindInHsBootFileErr mbinds
+                  ; pure (ValBinds x emptyBag sigs) }
+            _ -> pprPanic "rnTopBindsLHSBoot" (ppr topBinds) }
+
 rnTopBindsBoot :: NameSet -> HsValBindsLR GhcRn GhcPs
                -> RnM (HsValBinds GhcRn, DefUses)
 -- A hs-boot file has no bindings.
 -- Return a single HsBindGroup with empty binds and renamed signatures
-rnTopBindsBoot bound_names (ValBinds _ mbinds sigs)
-  = do  { checkErr (isEmptyLHsBinds mbinds) (bindsInHsBootFile mbinds)
-        ; (sigs', fvs) <- renameSigs (HsBootCtxt bound_names) sigs
+rnTopBindsBoot bound_names (ValBinds _ _ sigs)
+  = do  { (sigs', fvs) <- renameSigs (HsBootCtxt bound_names) sigs
         ; return (XValBindsLR (NValBinds [] sigs'), usesOnly fvs) }
 rnTopBindsBoot _ b = pprPanic "rnTopBindsBoot" (ppr b)
 
@@ -229,9 +244,9 @@ rnIPBinds (IPBinds _ ip_binds ) = do
     return (IPBinds noExtField ip_binds', plusFVs fvs_s)
 
 rnIPBind :: IPBind GhcPs -> RnM (IPBind GhcRn, FreeVars)
-rnIPBind (IPBind _ ~(Left n) expr) = do
+rnIPBind (IPBind _ n expr) = do
     (expr',fvExpr) <- rnLExpr expr
-    return (IPBind noAnn (Left n) expr', fvExpr)
+    return (IPBind noExtField n expr', fvExpr)
 
 {-
 ************************************************************************
@@ -431,7 +446,8 @@ rnBindLHS name_maker _ bind@(FunBind { fun_id = rdr_name })
 rnBindLHS name_maker _ (PatSynBind x psb@PSB{ psb_id = rdrname })
   | isTopRecNameMaker name_maker
   = do { addLocMA checkConName rdrname
-       ; name <- lookupLocatedTopBndrRnN rdrname -- Should be in scope already
+       ; name <-
+           lookupLocatedTopConstructorRnN rdrname -- Should be in scope already
        ; return (PatSynBind x psb{ psb_ext = noAnn, psb_id = name }) }
 
   | otherwise  -- Pattern synonym, not at top level
@@ -440,10 +456,8 @@ rnBindLHS name_maker _ (PatSynBind x psb@PSB{ psb_id = rdrname })
        ; name <- applyNameMaker name_maker rdrname
        ; return (PatSynBind x psb{ psb_ext = noAnn, psb_id = name }) }
   where
-    localPatternSynonymErr :: SDoc
-    localPatternSynonymErr
-      = hang (text "Illegal pattern synonym declaration for" <+> quotes (ppr rdrname))
-           2 (text "Pattern synonym declarations are only valid at top level")
+    localPatternSynonymErr :: TcRnMessage
+    localPatternSynonymErr = TcRnIllegalPatSynDecl rdrname
 
 rnBindLHS _ _ b = pprPanic "rnBindHS" (ppr b)
 
@@ -489,8 +503,7 @@ rnBind _ bind@(PatBind { pat_lhs = pat
         -- See Note [Pattern bindings that bind no variables]
         ; whenWOptM Opt_WarnUnusedPatternBinds $
           when (null bndrs && not ok_nobind_pat) $
-          addWarn (Reason Opt_WarnUnusedPatternBinds) $
-          unusedPatBindWarn bind'
+          addTcRnDiagnostic (TcRnUnusedPatternBinds bind')
 
         ; fvs' `seq` -- See Note [Free-variable space leak]
           return (bind', bndrs, all_fvs) }
@@ -651,9 +664,10 @@ makeMiniFixityEnv decls = foldlM add_one_sig emptyFsEnv decls
            ; return env}
      }
 
-dupFixityDecl :: SrcSpan -> RdrName -> SDoc
+dupFixityDecl :: SrcSpan -> RdrName -> TcRnMessage
 dupFixityDecl loc rdr_name
-  = vcat [text "Multiple fixity declarations for" <+> quotes (ppr rdr_name),
+  = TcRnUnknownMessage $ mkPlainError noHints $
+    vcat [text "Multiple fixity declarations for" <+> quotes (ppr rdr_name),
           text "also at " <+> ppr loc]
 
 
@@ -694,7 +708,7 @@ rnPatSynBind sig_fn bind@(PSB { psb_id = L l name
                       ; return ( (pat', InfixCon name1 name2)
                                , mkFVs (map unLoc [name1, name2])) }
                RecCon vars ->
-                   do { checkDupRdrNames (map (rdrNameFieldOcc . recordPatSynField) vars)
+                   do { checkDupRdrNames (map (foLabel . recordPatSynField) vars)
                       ; fls <- lookupConstructorFields name
                       ; let fld_env = mkFsEnv [ (flLabel fl, fl) | fl <- fls ]
                       ; let rnRecordPatSynField
@@ -730,7 +744,7 @@ rnPatSynBind sig_fn bind@(PSB { psb_id = L l name
                           , psb_ext = fvs' }
               selector_names = case details' of
                                  RecCon names ->
-                                  map (extFieldOcc . recordPatSynField) names
+                                  map (foExt . recordPatSynField) names
                                  _ -> []
 
         ; fvs' `seq` -- See Note [Free-variable space leak]
@@ -741,9 +755,10 @@ rnPatSynBind sig_fn bind@(PSB { psb_id = L l name
     -- See Note [Renaming pattern synonym variables]
     lookupPatSynBndr = wrapLocMA lookupLocalOccRn
 
-    patternSynonymErr :: SDoc
+    patternSynonymErr :: TcRnMessage
     patternSynonymErr
-      = hang (text "Illegal pattern synonym declaration")
+      = TcRnUnknownMessage $ mkPlainError noHints $
+        hang (text "Illegal pattern synonym declaration")
            2 (text "Use -XPatternSynonyms to enable this extension")
 
 {-
@@ -860,15 +875,17 @@ rnMethodBinds is_cls_decl cls ktv_names binds sigs
 
        -- Rename the pragmas and signatures
        -- Annoyingly the type variables /are/ in scope for signatures, but
-       -- /are not/ in scope in SPECIALISE and SPECIALISE instance pragmas.
-       -- See Note [Type variable scoping in SPECIALISE pragmas].
-       ; let (spec_prags, other_sigs) = partition (isSpecLSig <||> isSpecInstLSig) sigs
+       -- /are not/ in scope in the SPECIALISE instance pramas; e.g.
+       --    instance Eq a => Eq (T a) where
+       --       (==) :: a -> a -> a
+       --       {-# SPECIALISE instance Eq a => Eq (T [a]) #-}
+       ; let (spec_inst_prags, other_sigs) = partition isSpecInstLSig sigs
              bound_nms = mkNameSet (collectHsBindsBinders CollNoDictBinders binds')
              sig_ctxt | is_cls_decl = ClsDeclCtxt cls
                       | otherwise   = InstDeclCtxt bound_nms
-       ; (spec_prags', spg_fvs) <- renameSigs sig_ctxt spec_prags
-       ; (other_sigs', sig_fvs) <- bindLocalNamesFV ktv_names $
-                                      renameSigs sig_ctxt other_sigs
+       ; (spec_inst_prags', sip_fvs) <- renameSigs sig_ctxt spec_inst_prags
+       ; (other_sigs',      sig_fvs) <- bindLocalNamesFV ktv_names $
+                                        renameSigs sig_ctxt other_sigs
 
        -- Rename the bindings RHSs.  Again there's an issue about whether the
        -- type variables from the class/instance head are in scope.
@@ -879,47 +896,8 @@ rnMethodBinds is_cls_decl cls ktv_names binds sigs
                                            emptyFVs binds_w_dus
                  ; return (mapBag fstOf3 binds_w_dus, bind_fvs) }
 
-       ; return ( binds'', spec_prags' ++ other_sigs'
-                , sig_fvs `plusFV` spg_fvs `plusFV` bind_fvs) }
-
-{- Note [Type variable scoping in SPECIALISE pragmas]
-~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-When renaming the methods of a class or instance declaration, we must be careful
-with the scoping of the type variables that occur in SPECIALISE and SPECIALISE instance
-pragmas: the type variables from the class/instance header DO NOT scope over these,
-unlike class/instance method type signatures.
-
-Examples:
-
-  1. SPECIALISE
-
-    class C a where
-      meth :: a
-    instance C (Maybe a) where
-      meth = Nothing
-      {-# SPECIALISE INLINE meth :: Maybe [a] #-}
-
-  2. SPECIALISE instance
-
-    instance Eq a => Eq (T a) where
-       (==) :: a -> a -> a
-       {-# SPECIALISE instance Eq a => Eq (T [a]) #-}
-
-  In both cases, the type variable `a` mentioned in the PRAGMA is NOT the same
-  as the type variable `a` from the instance header.
-  For example, the SPECIALISE instance pragma above is a shorthand for
-
-      {-# SPECIALISE instance forall a. Eq a => Eq (T [a]) #-}
-
-  which is alpha-equivalent to
-
-      {-# SPECIALISE instance forall b. Eq b => Eq (T [b]) #-}
-
-  This shows that the type variables are not bound in the header.
-
-  Getting this scoping wrong can lead to out-of-scope type variable errors from
-  Core Lint, see e.g. #22913.
--}
+       ; return ( binds'', spec_inst_prags' ++ other_sigs'
+                , sig_fvs `plusFV` sip_fvs `plusFV` bind_fvs) }
 
 rnMethodBindLHS :: Bool -> Name
                 -> LHsBindLR GhcPs GhcPs
@@ -935,7 +913,7 @@ rnMethodBindLHS _ cls (L loc bind@(FunBind { fun_id = name })) rest
 -- Report error for all other forms of bindings
 -- This is why we use a fold rather than map
 rnMethodBindLHS is_cls_decl _ (L loc bind) rest
-  = do { addErrAt (locA loc) $
+  = do { addErrAt (locA loc) $ TcRnUnknownMessage $ mkPlainError noHints $
          vcat [ what <+> text "not allowed in" <+> decl_sort
               , nest 2 (ppr bind) ]
        ; return rest }
@@ -1005,6 +983,7 @@ renameSig ctxt sig@(ClassOpSig _ is_deflt vs ty)
   = do  { defaultSigs_on <- xoptM LangExt.DefaultSignatures
         ; when (is_deflt && not defaultSigs_on) $
           addErr (defaultSigErr sig)
+        ; mapM_ warnForallIdentifier vs
         ; new_v <- mapM (lookupSigOccRnN ctxt sig) vs
         ; (new_ty, fvs) <- rnHsSigType ty_ctxt TypeLevel ty
         ; return (ClassOpSig noAnn is_deflt new_v new_ty, fvs) }
@@ -1081,8 +1060,8 @@ renameSig _ctxt sig@(CompleteMatchSig _ s (L l bf) mty)
 
        return (CompleteMatchSig noAnn s (L l new_bf) new_mty, emptyFVs)
   where
-    orphanError :: SDoc
-    orphanError =
+    orphanError :: TcRnMessage
+    orphanError = TcRnUnknownMessage $ mkPlainError noHints $
       text "Orphan COMPLETE pragmas not supported" $$
       text "A COMPLETE pragma must mention at least one data constructor" $$
       text "or pattern synonym defined in the same module."
@@ -1205,20 +1184,48 @@ type AnnoBody body
     , Anno [LocatedA (Match GhcPs (LocatedA (body GhcPs)))] ~ SrcSpanAnnL
     , Anno (Match GhcRn (LocatedA (body GhcRn))) ~ SrcSpanAnnA
     , Anno (Match GhcPs (LocatedA (body GhcPs))) ~ SrcSpanAnnA
-    , Anno (GRHS GhcRn (LocatedA (body GhcRn))) ~ SrcSpan
-    , Anno (GRHS GhcPs (LocatedA (body GhcPs))) ~ SrcSpan
+    , Anno (GRHS GhcRn (LocatedA (body GhcRn))) ~ SrcAnn NoEpAnns
+    , Anno (GRHS GhcPs (LocatedA (body GhcPs))) ~ SrcAnn NoEpAnns
     , Outputable (body GhcPs)
     )
 
+-- Note [Empty MatchGroups]
+-- ~~~~~~~~~~~~~~~~~~~~~~~~
+-- In some cases, MatchGroups are allowed to be empty. Firstly, the
+-- prerequisite is that -XEmptyCase is enabled. Then you have an empty
+-- MatchGroup resulting either from a case-expression:
+--
+--     case e of {}
+--
+-- or from a \case-expression:
+--
+--     \case {}
+--
+-- NB: \cases {} is not allowed, since it's not clear how many patterns this
+-- should match on.
+--
+-- The same applies in arrow notation commands: With -XEmptyCases, it is
+-- allowed in case- and \case-commands, but not \cases.
+--
+-- Since the lambda expressions and empty function definitions are already
+-- disallowed elsewhere, here, we only need to make sure we don't accept empty
+-- \cases expressions or commands. In that case, or if we encounter an empty
+-- MatchGroup but -XEmptyCases is disabled, we add an error.
+
 rnMatchGroup :: (Outputable (body GhcPs), AnnoBody body) => HsMatchContext GhcRn
              -> (LocatedA (body GhcPs) -> RnM (LocatedA (body GhcRn), FreeVars))
              -> MatchGroup GhcPs (LocatedA (body GhcPs))
              -> RnM (MatchGroup GhcRn (LocatedA (body GhcRn)), FreeVars)
 rnMatchGroup ctxt rnBody (MG { mg_alts = L lm ms, mg_origin = origin })
-  = do { empty_case_ok <- xoptM LangExt.EmptyCase
-       ; when (null ms && not empty_case_ok) (addErr (emptyCaseErr ctxt))
+         -- see Note [Empty MatchGroups]
+  = do { whenM ((null ms &&) <$> mustn't_be_empty) (addErr (emptyCaseErr ctxt))
        ; (new_ms, ms_fvs) <- mapFvRn (rnMatch ctxt rnBody) ms
        ; return (mkMatchGroup origin (L lm new_ms), ms_fvs) }
+  where
+    mustn't_be_empty = case ctxt of
+      LamCaseAlt LamCases -> return True
+      ArrowMatchCtxt (ArrowLamCaseAlt LamCases) -> return True
+      _ -> not <$> xoptM LangExt.EmptyCase
 
 rnMatch :: AnnoBody body
         => HsMatchContext GhcRn
@@ -1242,17 +1249,29 @@ rnMatch' ctxt rnBody (Match { m_ctxt = mf, m_pats = pats, m_grhss = grhss })
         ; return (Match { m_ext = noAnn, m_ctxt = mf', m_pats = pats'
                         , m_grhss = grhss'}, grhss_fvs ) }
 
-emptyCaseErr :: HsMatchContext GhcRn -> SDoc
-emptyCaseErr ctxt = hang (text "Empty list of alternatives in" <+> pp_ctxt ctxt)
-                       2 (text "Use EmptyCase to allow this")
+emptyCaseErr :: HsMatchContext GhcRn -> TcRnMessage
+emptyCaseErr ctxt = TcRnUnknownMessage $ mkPlainError noHints $ message ctxt
   where
     pp_ctxt :: HsMatchContext GhcRn -> SDoc
     pp_ctxt c = case c of
-      CaseAlt       -> text "case expression"
-      LambdaExpr    -> text "\\case expression"
-      ArrowMatchCtxt ArrowCaseAlt -> text "case expression"
-      ArrowMatchCtxt KappaExpr    -> text "kappa abstraction"
-      _             -> text "(unexpected)" <+> pprMatchContextNoun c
+      CaseAlt                                  -> text "case expression"
+      LamCaseAlt LamCase                       -> text "\\case expression"
+      ArrowMatchCtxt (ArrowLamCaseAlt LamCase) -> text "\\case command"
+      ArrowMatchCtxt ArrowCaseAlt              -> text "case command"
+      ArrowMatchCtxt KappaExpr                 -> text "kappa abstraction"
+      _                                        -> text "(unexpected)"
+                                                  <+> pprMatchContextNoun c
+
+    message :: HsMatchContext GhcRn -> SDoc
+    message (LamCaseAlt LamCases) = lcases_msg <+> text "expression"
+    message (ArrowMatchCtxt (ArrowLamCaseAlt LamCases)) =
+      lcases_msg <+> text "command"
+    message ctxt =
+      hang (text "Empty list of alternatives in" <+> pp_ctxt ctxt)
+           2 (text "Use EmptyCase to allow this")
+
+    lcases_msg =
+      text "Empty list of alternatives is not allowed in \\cases"
 
 {-
 ************************************************************************
@@ -1277,7 +1296,7 @@ rnGRHS :: AnnoBody body
        -> (LocatedA (body GhcPs) -> RnM (LocatedA (body GhcRn), FreeVars))
        -> LGRHS GhcPs (LocatedA (body GhcPs))
        -> RnM (LGRHS GhcRn (LocatedA (body GhcRn)), FreeVars)
-rnGRHS ctxt rnBody = wrapLocFstM (rnGRHS' ctxt rnBody)
+rnGRHS ctxt rnBody = wrapLocFstMA (rnGRHS' ctxt rnBody)
 
 rnGRHS' :: HsMatchContext GhcRn
         -> (LocatedA (body GhcPs) -> RnM (LocatedA (body GhcRn), FreeVars))
@@ -1288,8 +1307,10 @@ rnGRHS' ctxt rnBody (GRHS _ guards rhs)
         ; ((guards', rhs'), fvs) <- rnStmts (PatGuard ctxt) rnExpr guards $ \ _ ->
                                     rnBody rhs
 
-        ; unless (pattern_guards_allowed || is_standard_guard guards')
-                 (addWarn NoReason (nonStdGuardErr guards'))
+        ; unless (pattern_guards_allowed || is_standard_guard guards') $
+            let diag = TcRnUnknownMessage $
+                  mkPlainDiagnostic WarningWithoutFlag noHints (nonStdGuardErr guards')
+            in addDiagnostic diag
 
         ; return (GRHS noAnn guards' rhs', fvs) }
   where
@@ -1342,7 +1363,7 @@ rnSrcFixityDecl sig_ctxt = rn_decl
 
 dupSigDeclErr :: NonEmpty (LocatedN RdrName, Sig GhcPs) -> RnM ()
 dupSigDeclErr pairs@((L loc name, sig) :| _)
-  = addErrAt (locA loc) $
+  = addErrAt (locA loc) $ TcRnUnknownMessage $ mkPlainError noHints $
     vcat [ text "Duplicate" <+> what_it_is
            <> text "s for" <+> quotes (ppr name)
          , text "at" <+> vcat (map ppr $ sortBy SrcLoc.leftmost_smallest
@@ -1354,18 +1375,19 @@ dupSigDeclErr pairs@((L loc name, sig) :| _)
 
 misplacedSigErr :: LSig GhcRn -> RnM ()
 misplacedSigErr (L loc sig)
-  = addErrAt (locA loc) $
+  = addErrAt (locA loc) $ TcRnUnknownMessage $ mkPlainError noHints $
     sep [text "Misplaced" <+> hsSigDoc sig <> colon, ppr sig]
 
-defaultSigErr :: Sig GhcPs -> SDoc
-defaultSigErr sig = vcat [ hang (text "Unexpected default signature:")
-                              2 (ppr sig)
-                         , text "Use DefaultSignatures to enable default signatures" ]
+defaultSigErr :: Sig GhcPs -> TcRnMessage
+defaultSigErr sig = TcRnUnknownMessage $ mkPlainError noHints $
+  vcat [ hang (text "Unexpected default signature:")
+         2 (ppr sig)
+       , text "Use DefaultSignatures to enable default signatures" ]
 
-bindsInHsBootFile :: LHsBindsLR GhcRn GhcPs -> SDoc
-bindsInHsBootFile mbinds
-  = hang (text "Bindings in hs-boot files are not allowed")
-       2 (ppr mbinds)
+bindInHsBootFileErr :: LHsBindLR GhcRn GhcPs -> RnM ()
+bindInHsBootFileErr (L loc _)
+  = addErrAt (locA loc) $ TcRnUnknownMessage $ mkPlainError noHints $
+      vcat [ text "Bindings in hs-boot files are not allowed" ]
 
 nonStdGuardErr :: (Outputable body,
                    Anno (Stmt GhcRn body) ~ SrcSpanAnnA)
@@ -1374,14 +1396,9 @@ nonStdGuardErr guards
   = hang (text "accepting non-standard pattern guards (use PatternGuards to suppress this message)")
        4 (interpp'SP guards)
 
-unusedPatBindWarn :: HsBind GhcRn -> SDoc
-unusedPatBindWarn bind
-  = hang (text "This pattern-binding binds no variables:")
-       2 (ppr bind)
-
 dupMinimalSigErr :: [LSig GhcPs] -> RnM ()
 dupMinimalSigErr sigs@(L loc _ : _)
-  = addErrAt (locA loc) $
+  = addErrAt (locA loc) $ TcRnUnknownMessage $ mkPlainError noHints $
     vcat [ text "Multiple minimal complete definitions"
          , text "at" <+> vcat (map ppr $ sortBy SrcLoc.leftmost_smallest $ map getLocA sigs)
          , text "Combine alternative minimal complete definitions with `|'" ]
