@@ -1,10 +1,9 @@
-{-# LANGUAGE CPP #-}
+
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE FlexibleInstances #-}
 
-
 -----------------------------------------------------------------------------
 --
 -- Stg to C-- code generation:
@@ -35,9 +34,9 @@ module GHC.StgToCmm.Closure (
         isLFThunk, isLFReEntrant, lfUpdatable,
 
         -- * Used by other modules
-        CgLoc(..), SelfLoopInfo, CallMethod(..),
+        CgLoc(..), CallMethod(..),
         nodeMustPointToIt, isKnownFun, funTag, tagForArity,
-        CallOpts(..), getCallMethod,
+        getCallMethod,
 
         -- * ClosureInfo
         ClosureInfo,
@@ -66,10 +65,9 @@ module GHC.StgToCmm.Closure (
         cafBlackHoleInfoTable,
         indStaticInfoTable,
         staticClosureNeedsLink,
+        mkClosureInfoTableLabel
     ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 import GHC.Platform
 import GHC.Platform.Profile
@@ -80,6 +78,7 @@ import GHC.Cmm
 import GHC.Cmm.Utils
 import GHC.Cmm.Ppr.Expr() -- For Outputable instances
 import GHC.StgToCmm.Types
+import GHC.StgToCmm.Sequel
 
 import GHC.Types.CostCentre
 import GHC.Cmm.BlockId
@@ -96,10 +95,13 @@ import GHC.Types.RepType
 import GHC.Types.Basic
 import GHC.Utils.Outputable
 import GHC.Utils.Panic
+import GHC.Utils.Panic.Plain
 import GHC.Utils.Misc
 
 import Data.Coerce (coerce)
 import qualified Data.ByteString.Char8 as BS8
+import GHC.StgToCmm.Config
+import GHC.Stg.InferTags.TagSig (isTaggedSig)
 
 -----------------------------------------------------------------------------
 --                Data types and synonyms
@@ -127,8 +129,6 @@ pprCgLoc platform = \case
    CmmLoc e    -> text "cmm" <+> pdoc platform e
    LneLoc b rs -> text "lne" <+> ppr b <+> ppr rs
 
-type SelfLoopInfo = (Id, BlockId, [LocalReg])
-
 -- used by ticky profiling
 isKnownFun :: LambdaFormInfo -> Bool
 isKnownFun LFReEntrant{} = True
@@ -152,23 +152,23 @@ instance (Outputable a) => Outputable (NonVoid a) where
   ppr (NonVoid a) = ppr a
 
 nonVoidIds :: [Id] -> [NonVoid Id]
-nonVoidIds ids = [NonVoid id | id <- ids, not (isVoidTy (idType id))]
+nonVoidIds ids = [NonVoid id | id <- ids, not (isZeroBitTy (idType id))]
 
 -- | Used in places where some invariant ensures that all these Ids are
 -- non-void; e.g. constructor field binders in case expressions.
 -- See Note [Post-unarisation invariants] in "GHC.Stg.Unarise".
 assertNonVoidIds :: [Id] -> [NonVoid Id]
-assertNonVoidIds ids = ASSERT(not (any (isVoidTy . idType) ids))
+assertNonVoidIds ids = assert (not (any (isZeroBitTy . idType) ids)) $
                        coerce ids
 
 nonVoidStgArgs :: [StgArg] -> [NonVoid StgArg]
-nonVoidStgArgs args = [NonVoid arg | arg <- args, not (isVoidTy (stgArgType arg))]
+nonVoidStgArgs args = [NonVoid arg | arg <- args, not (isZeroBitTy (stgArgType arg))]
 
 -- | Used in places where some invariant ensures that all these arguments are
 -- non-void; e.g. constructor arguments.
 -- See Note [Post-unarisation invariants] in "GHC.Stg.Unarise".
 assertNonVoidStgArgs :: [StgArg] -> [NonVoid StgArg]
-assertNonVoidStgArgs args = ASSERT(not (any (isVoidTy . stgArgType) args))
+assertNonVoidStgArgs args = assert (not (any (isZeroBitTy . stgArgType) args)) $
                             coerce args
 
 
@@ -209,7 +209,7 @@ argPrimRep arg = typePrimRep1 (stgArgType arg)
 mkLFArgument :: Id -> LambdaFormInfo
 mkLFArgument id
   | isUnliftedType ty      = LFUnlifted
-  | might_be_a_function ty = LFUnknown True
+  | mightBeFunTy ty = LFUnknown True
   | otherwise              = LFUnknown False
   where
     ty = idType id
@@ -233,23 +233,11 @@ mkLFReEntrant top fvs args arg_descr
 -------------
 mkLFThunk :: Type -> TopLevelFlag -> [Id] -> UpdateFlag -> LambdaFormInfo
 mkLFThunk thunk_ty top fvs upd_flag
-  = ASSERT( not (isUpdatable upd_flag) || not (isUnliftedType thunk_ty) )
+  = assert (not (isUpdatable upd_flag) || not (isUnliftedType thunk_ty)) $
     LFThunk top (null fvs)
             (isUpdatable upd_flag)
             NonStandardThunk
-            (might_be_a_function thunk_ty)
-
---------------
-might_be_a_function :: Type -> Bool
--- Return False only if we are *sure* it's a data type
--- Look through newtypes etc as much as poss
-might_be_a_function ty
-  | [LiftedRep] <- typePrimRep ty
-  , Just tc <- tyConAppTyCon_maybe (unwrapType ty)
-  , isDataTyCon tc
-  = False
-  | otherwise
-  = True
+            (mightBeFunTy thunk_ty)
 
 -------------
 mkConLFInfo :: DataCon -> LambdaFormInfo
@@ -259,13 +247,13 @@ mkConLFInfo con = LFCon con
 mkSelectorLFInfo :: Id -> Int -> Bool -> LambdaFormInfo
 mkSelectorLFInfo id offset updatable
   = LFThunk NotTopLevel False updatable (SelectorThunk offset)
-        (might_be_a_function (idType id))
+        (mightBeFunTy (idType id))
 
 -------------
 mkApLFInfo :: Id -> UpdateFlag -> Arity -> LambdaFormInfo
 mkApLFInfo id upd_flag arity
   = LFThunk NotTopLevel (arity == 0) (isUpdatable upd_flag) (ApThunk arity)
-        (might_be_a_function (idType id))
+        (mightBeFunTy (idType id))
 
 -------------
 mkLFImported :: Id -> LambdaFormInfo
@@ -479,27 +467,39 @@ When black-holing, single-entry closures could also be entered via node
 (rather than directly) to catch double-entry. -}
 
 data CallMethod
-  = EnterIt             -- No args, not a function
+  = EnterIt             -- ^ No args, not a function
 
   | JumpToIt BlockId [LocalReg] -- A join point or a header of a local loop
 
   | ReturnIt            -- It's a value (function, unboxed value,
                         -- or constructor), so just return it.
 
-  | SlowCall                -- Unknown fun, or known fun with
+  | InferedReturnIt     -- A properly tagged value, as determined by tag inference.
+                        -- See Note [Tag Inference] and Note [Tag inference passes] in
+                        -- GHC.Stg.InferTags.
+                        -- It behaves /precisely/ like `ReturnIt`, except that when debugging is
+                        -- enabled we emit an extra assertion to check that the returned value is
+                        -- properly tagged.  We can use this as a check that tag inference is working
+                        -- correctly.
+                        -- TODO: SPJ suggested we could combine this with EnterIt, but for now I decided
+                        -- not to do so.
+
+  | SlowCall            -- Unknown fun, or known fun with
                         -- too few args.
 
   | DirectEntry         -- Jump directly, with args in regs
         CLabel          --   The code label
         RepArity        --   Its arity
 
-data CallOpts = CallOpts
-   { co_profile       :: !Profile   -- ^ Platform profile
-   , co_loopification :: !Bool      -- ^ Loopification enabled (cf @-floopification@)
-   , co_ticky         :: !Bool      -- ^ Ticky profiling enabled (cf @-ticky@)
-   }
+instance Outputable CallMethod where
+  ppr (EnterIt) = text "Enter"
+  ppr (JumpToIt {}) = text "JumpToIt"
+  ppr (ReturnIt ) = text "ReturnIt"
+  ppr (InferedReturnIt) = text "InferedReturnIt"
+  ppr (SlowCall ) = text "SlowCall"
+  ppr (DirectEntry {}) = text "DirectEntry"
 
-getCallMethod :: CallOpts
+getCallMethod :: StgToCmmConfig
               -> Name           -- Function being applied
               -> Id             -- Function Id used to chech if it can refer to
                                 -- CAF's and whether the function is tail-calling
@@ -512,12 +512,11 @@ getCallMethod :: CallOpts
                                 -- tail calls using the same data constructor,
                                 -- JumpToIt. This saves us one case branch in
                                 -- cgIdApp
-              -> Maybe SelfLoopInfo -- can we perform a self-recursive tail call?
+              -> Maybe SelfLoopInfo -- can we perform a self-recursive tail-call
               -> CallMethod
 
-getCallMethod opts _ id _ n_args v_args _cg_loc
-              (Just (self_loop_id, block_id, args))
-  | co_loopification opts
+getCallMethod cfg _ id _  n_args v_args _cg_loc (Just (self_loop_id, block_id, args))
+  | stgToCmmLoopification cfg
   , id == self_loop_id
   , args `lengthIs` (n_args - v_args)
   -- If these patterns match then we know that:
@@ -528,31 +527,36 @@ getCallMethod opts _ id _ n_args v_args _cg_loc
   -- self-recursive tail calls] in GHC.StgToCmm.Expr for more details
   = JumpToIt block_id args
 
-getCallMethod opts name id (LFReEntrant _ arity _ _) n_args _v_args _cg_loc
-              _self_loop_info
+getCallMethod cfg name id (LFReEntrant _ arity _ _) n_args _v_args _cg_loc _self_loop_info
   | n_args == 0 -- No args at all
-  && not (profileIsProfiling (co_profile opts))
+  && not (profileIsProfiling (stgToCmmProfile cfg))
      -- See Note [Evaluating functions with profiling] in rts/Apply.cmm
-  = ASSERT( arity /= 0 ) ReturnIt
+  = assert (arity /= 0) ReturnIt
   | n_args < arity = SlowCall        -- Not enough args
-  | otherwise      = DirectEntry (enterIdLabel (profilePlatform (co_profile opts)) name (idCafInfo id)) arity
+  | otherwise      = DirectEntry (enterIdLabel (stgToCmmPlatform cfg) name (idCafInfo id)) arity
 
 getCallMethod _ _name _ LFUnlifted n_args _v_args _cg_loc _self_loop_info
-  = ASSERT( n_args == 0 ) ReturnIt
+  = assert (n_args == 0) ReturnIt
 
 getCallMethod _ _name _ (LFCon _) n_args _v_args _cg_loc _self_loop_info
-  = ASSERT( n_args == 0 ) ReturnIt
+  = assert (n_args == 0) ReturnIt
     -- n_args=0 because it'd be ill-typed to apply a saturated
     --          constructor application to anything
 
-getCallMethod opts name id (LFThunk _ _ updatable std_form_info is_fun)
+getCallMethod cfg name id (LFThunk _ _ updatable std_form_info is_fun)
               n_args _v_args _cg_loc _self_loop_info
+
+  | Just sig <- idTagSig_maybe id
+  , isTaggedSig sig -- Infered to be already evaluated by Tag Inference
+  , n_args == 0     -- See Note [Tag Inference]
+  = InferedReturnIt
+
   | is_fun      -- it *might* be a function, so we must "call" it (which is always safe)
   = SlowCall    -- We cannot just enter it [in eval/apply, the entry code
                 -- is the fast-entry code]
 
   -- Since is_fun is False, we are *definitely* looking at a data value
-  | updatable || co_ticky opts -- to catch double entry
+  | updatable || stgToCmmDoTicky cfg -- to catch double entry
       {- OLD: || opt_SMP
          I decided to remove this, because in SMP mode it doesn't matter
          if we enter the same thunk multiple times, so the optimisation
@@ -565,7 +569,7 @@ getCallMethod opts name id (LFThunk _ _ updatable std_form_info is_fun)
   | SelectorThunk{} <- std_form_info
   = EnterIt
 
-    -- We used to have ASSERT( n_args == 0 ), but actually it is
+    -- We used to have assert (n_args == 0 ), but actually it is
     -- possible for the optimiser to generate
     --   let bot :: Int = error Int "urk"
     --   in (bot `cast` unsafeCoerce Int (Int -> Int)) 3
@@ -573,19 +577,33 @@ getCallMethod opts name id (LFThunk _ _ updatable std_form_info is_fun)
     -- So the right thing to do is just to enter the thing
 
   | otherwise        -- Jump direct to code for single-entry thunks
-  = ASSERT( n_args == 0 )
-    DirectEntry (thunkEntryLabel (profilePlatform (co_profile opts)) name (idCafInfo id) std_form_info
+  = assert (n_args == 0) $
+    DirectEntry (thunkEntryLabel (stgToCmmPlatform cfg) name (idCafInfo id) std_form_info
                 updatable) 0
 
-getCallMethod _ _name _ (LFUnknown True) _n_arg _v_args _cg_locs _self_loop_info
-  = SlowCall -- might be a function
+-- Imported(Unknown) Ids
+getCallMethod cfg name id (LFUnknown might_be_a_function) n_args _v_args _cg_locs _self_loop_info
+  | n_args == 0
+  , Just sig <- idTagSig_maybe id
+  , isTaggedSig sig -- Infered to be already evaluated by Tag Inference
+  -- When profiling we must enter all potential functions to make sure we update the SCC
+  -- even if the function itself is already evaluated.
+  -- See Note [Evaluating functions with profiling] in rts/Apply.cmm
+  , not (profileIsProfiling (stgToCmmProfile cfg) && might_be_a_function)
+  = InferedReturnIt -- See Note [Tag Inference]
 
-getCallMethod _ name _ (LFUnknown False) n_args _v_args _cg_loc _self_loop_info
-  = ASSERT2( n_args == 0, ppr name <+> ppr n_args )
-    EnterIt -- Not a function
+  | might_be_a_function = SlowCall
 
-getCallMethod _ _name _ LFLetNoEscape _n_args _v_args (LneLoc blk_id lne_regs)
-              _self_loop_info
+  | otherwise =
+      assertPpr ( n_args == 0) ( ppr name <+> ppr n_args )
+      EnterIt   -- Not a function
+
+-- TODO: Redundant with above match?
+-- getCallMethod _ name _ (LFUnknown False) n_args _v_args _cg_loc _self_loop_info
+--   = assertPpr (n_args == 0) (ppr name <+> ppr n_args)
+--     EnterIt -- Not a function
+
+getCallMethod _ _name _ LFLetNoEscape _n_args _v_args (LneLoc blk_id lne_regs) _self_loop_info
   = JumpToIt blk_id lne_regs
 
 getCallMethod _ _ _ _ _ _ _ _ = panic "Unknown call method"
@@ -613,7 +631,7 @@ getCallMethod _ _ _ _ _ _ _ _ = panic "Unknown call method"
 
 data ClosureInfo
   = ClosureInfo {
-        closureName :: !Name,           -- The thing bound to this closure
+        closureName :: !Id,           -- The thing bound to this closure
            -- we don't really need this field: it's only used in generating
            -- code for ticky and profiling, and we could pass the information
            -- around separately, but it doesn't do much harm to keep it here.
@@ -650,13 +668,12 @@ mkClosureInfo :: Profile
               -> String         -- String descriptor
               -> ClosureInfo
 mkClosureInfo profile is_static id lf_info tot_wds ptr_wds val_descr
-  = ClosureInfo { closureName      = name
+  = ClosureInfo { closureName      = id
                 , closureLFInfo    = lf_info
                 , closureInfoLabel = info_lbl   -- These three fields are
                 , closureSMRep     = sm_rep     -- (almost) an info table
                 , closureProf      = prof }     -- (we don't have an SRT yet)
   where
-    name       = idName id
     sm_rep     = mkHeapRep profile is_static ptr_wds nonptr_wds (lfClosureType lf_info)
     prof       = mkProfilingInfo profile id val_descr
     nonptr_wds = tot_wds - ptr_wds
@@ -810,6 +827,7 @@ closureLocalEntryLabel platform
   | platformTablesNextToCode platform = toInfoLbl  platform . closureInfoLabel
   | otherwise                         = toEntryLbl platform . closureInfoLabel
 
+-- | Get the info table label for a *thunk*.
 mkClosureInfoTableLabel :: Platform -> Id -> LambdaFormInfo -> CLabel
 mkClosureInfoTableLabel platform id lf_info
   = case lf_info of
@@ -819,23 +837,14 @@ mkClosureInfoTableLabel platform id lf_info
         LFThunk _ _ upd_flag (ApThunk arity) _
                       -> mkApInfoTableLabel platform upd_flag arity
 
-        LFThunk{}     -> std_mk_lbl name cafs
-        LFReEntrant{} -> std_mk_lbl name cafs
+        LFThunk{}     -> mkInfoTableLabel name cafs
+        LFReEntrant{} -> mkInfoTableLabel name cafs
         _other        -> panic "closureInfoTableLabel"
 
   where
     name = idName id
 
-    std_mk_lbl | is_local  = mkLocalInfoTableLabel
-               | otherwise = mkInfoTableLabel
-
     cafs     = idCafInfo id
-    is_local = isDataConWorkId id
-       -- Make the _info pointer for the implicit datacon worker
-       -- binding local. The reason we can do this is that importing
-       -- code always either uses the _closure or _con_info. By the
-       -- invariants in "GHC.CoreToStg.Prep" anything else gets eta expanded.
-
 
 -- | thunkEntryLabel is a local help function, not exported.  It's used from
 -- getCallMethod.
