@@ -127,7 +127,7 @@ cgForeignCall (CCall (CCallSpec target cconv safety)) typ stg_args res_ty
          }
 
 {- Note [safe foreign call convention]
-
+   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 The simple thing to do for a safe foreign call would be the same as an
 unsafe one: just
 
@@ -174,7 +174,7 @@ We will now generate this:
  L2:
   ... r ...
 
-And when the safe foreign call is lowered later (see Note [lower safe
+And when the safe foreign call is lowered later (see Note [Lower safe
 foreign calls]) we get this:
 
   suspendThread()
@@ -601,9 +601,9 @@ closureField profile off = off + fixedHdrSize profile
 -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 --
 -- For certain types passed to foreign calls, we adjust the actual
--- value passed to the call.  For ByteArray#, Array#, SmallArray#,
--- and ArrayArray#, we pass the address of the array's payload, not
--- the address of the heap object. For example, consider
+-- value passed to the call.  For ByteArray#, Array# and SmallArray#,
+-- we pass the address of the array's payload, not the address of
+-- the heap object. For example, consider:
 --   foreign import "c_foo" foo :: ByteArray# -> Int# -> IO ()
 -- At a Haskell call like `foo x y`, we'll generate a C call that
 -- is more like
@@ -713,8 +713,6 @@ typeToStgFArgType :: Type -> StgFArgType
 typeToStgFArgType typ
   | tycon == arrayPrimTyCon = StgArrayType
   | tycon == mutableArrayPrimTyCon = StgArrayType
-  | tycon == arrayArrayPrimTyCon = StgArrayType
-  | tycon == mutableArrayArrayPrimTyCon = StgArrayType
   | tycon == smallArrayPrimTyCon = StgSmallArrayType
   | tycon == smallMutableArrayPrimTyCon = StgSmallArrayType
   | tycon == byteArrayPrimTyCon = StgByteArrayType
