@@ -3,14 +3,16 @@
 
 -}
 
-{-# LANGUAGE CPP #-}
+
 {-# LANGUAGE FlexibleContexts #-}
 
 -- | Module finder
 module GHC.Unit.Finder (
     FindResult(..),
     InstalledFindResult(..),
+    FinderOpts(..),
     FinderCache,
+    initFinderCache,
     flushFinderCaches,
     findImportedModule,
     findPluginModule,
@@ -22,6 +24,7 @@ module GHC.Unit.Finder (
     mkHiOnlyModLocation,
     mkHiPath,
     mkObjPath,
+    addModuleToFinder,
     addHomeModuleToFinder,
     uncacheModule,
     mkStubPaths,
@@ -29,26 +32,23 @@ module GHC.Unit.Finder (
     findObjectLinkableMaybe,
     findObjectLinkable,
 
+    -- Hash cache
+    lookupFileCache
   ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
-import GHC.Driver.Env
-import GHC.Driver.Session
-
 import GHC.Platform.Ways
 
 import GHC.Builtin.Names ( gHC_PRIM )
 
+import GHC.Unit.Env
 import GHC.Unit.Types
 import GHC.Unit.Module
 import GHC.Unit.Home
 import GHC.Unit.State
 import GHC.Unit.Finder.Types
 
-import GHC.Data.FastString
 import GHC.Data.Maybe    ( expectJust )
 import qualified GHC.Data.ShortText as ST
 
@@ -57,13 +57,19 @@ import GHC.Utils.Outputable as Outputable
 import GHC.Utils.Panic
 
 import GHC.Linker.Types
+import GHC.Types.PkgQual
 
-import Data.IORef       ( IORef, readIORef, atomicModifyIORef' )
+import GHC.Fingerprint
+import Data.IORef
 import System.Directory
 import System.FilePath
 import Control.Monad
 import Data.Time
-
+import qualified Data.Map as M
+import GHC.Driver.Env
+    ( hsc_home_unit_maybe, HscEnv(hsc_FC, hsc_dflags, hsc_unit_env) )
+import GHC.Driver.Config.Finder
+import qualified Data.Set as Set
 
 type FileExt = String   -- Filename extension
 type BaseName = String  -- Basename of file
@@ -81,62 +87,125 @@ type BaseName = String  -- Basename of file
 -- -----------------------------------------------------------------------------
 -- The finder's cache
 
--- remove all the home modules from the cache; package modules are
--- assumed to not move around during a session.
-flushFinderCaches :: HscEnv -> IO ()
-flushFinderCaches hsc_env =
-  atomicModifyIORef' fc_ref $ \fm -> (filterInstalledModuleEnv is_ext fm, ())
- where
-        fc_ref       = hsc_FC hsc_env
-        home_unit    = hsc_home_unit hsc_env
-        is_ext mod _ = not (isHomeInstalledModule home_unit mod)
 
-addToFinderCache :: IORef FinderCache -> InstalledModule -> InstalledFindResult -> IO ()
-addToFinderCache ref key val =
+initFinderCache :: IO FinderCache
+initFinderCache = FinderCache <$> newIORef emptyInstalledModuleEnv
+                              <*> newIORef M.empty
+
+-- remove all the home modules from the cache; package modules are
+-- assumed to not move around during a session; also flush the file hash
+-- cache
+flushFinderCaches :: FinderCache -> UnitEnv -> IO ()
+flushFinderCaches (FinderCache ref file_ref) ue = do
+  atomicModifyIORef' ref $ \fm -> (filterInstalledModuleEnv is_ext fm, ())
+  atomicModifyIORef' file_ref $ \_ -> (M.empty, ())
+ where
+  is_ext mod _ = not (isUnitEnvInstalledModule ue mod)
+
+addToFinderCache :: FinderCache -> InstalledModule -> InstalledFindResult -> IO ()
+addToFinderCache (FinderCache ref _) key val =
   atomicModifyIORef' ref $ \c -> (extendInstalledModuleEnv c key val, ())
 
-removeFromFinderCache :: IORef FinderCache -> InstalledModule -> IO ()
-removeFromFinderCache ref key =
+removeFromFinderCache :: FinderCache -> InstalledModule -> IO ()
+removeFromFinderCache (FinderCache ref _) key =
   atomicModifyIORef' ref $ \c -> (delInstalledModuleEnv c key, ())
 
-lookupFinderCache :: IORef FinderCache -> InstalledModule -> IO (Maybe InstalledFindResult)
-lookupFinderCache ref key = do
+lookupFinderCache :: FinderCache -> InstalledModule -> IO (Maybe InstalledFindResult)
+lookupFinderCache (FinderCache ref _) key = do
    c <- readIORef ref
    return $! lookupInstalledModuleEnv c key
 
+lookupFileCache :: FinderCache -> FilePath -> IO Fingerprint
+lookupFileCache (FinderCache _ ref) key = do
+   c <- readIORef ref
+   case M.lookup key c of
+     Nothing -> do
+       hash <- getFileHash key
+       atomicModifyIORef' ref $ \c -> (M.insert key hash c, ())
+       return hash
+     Just fp -> return fp
+
 -- -----------------------------------------------------------------------------
 -- The three external entry points
 
+
 -- | Locate a module that was imported by the user.  We have the
 -- module's name, and possibly a package name.  Without a package
 -- name, this function will use the search path and the known exposed
 -- packages to find the module, if a package is specified then only
 -- that package is searched for the module.
 
-findImportedModule :: HscEnv -> ModuleName -> Maybe FastString -> IO FindResult
-findImportedModule hsc_env mod_name mb_pkg =
+findImportedModule :: HscEnv -> ModuleName -> PkgQual -> IO FindResult
+findImportedModule hsc_env mod fs =
+  let fc        = hsc_FC hsc_env
+      mhome_unit = hsc_home_unit_maybe hsc_env
+      dflags    = hsc_dflags hsc_env
+      fopts     = initFinderOpts dflags
+  in do
+    findImportedModuleNoHsc fc fopts (hsc_unit_env hsc_env) mhome_unit mod fs
+
+findImportedModuleNoHsc
+  :: FinderCache
+  -> FinderOpts
+  -> UnitEnv
+  -> Maybe HomeUnit
+  -> ModuleName
+  -> PkgQual
+  -> IO FindResult
+findImportedModuleNoHsc fc fopts ue mhome_unit mod_name mb_pkg =
   case mb_pkg of
-        Nothing                        -> unqual_import
-        Just pkg | pkg == fsLit "this" -> home_import -- "this" is special
-                 | otherwise           -> pkg_import
+    NoPkgQual  -> unqual_import
+    ThisPkg uid | (homeUnitId <$> mhome_unit) == Just uid -> home_import
+                | Just os <- lookup uid other_fopts -> home_pkg_import (uid, os)
+                | otherwise -> pprPanic "findImportModule" (ppr mod_name $$ ppr mb_pkg $$ ppr (homeUnitId <$> mhome_unit) $$ ppr uid $$ ppr (map fst all_opts))
+    OtherPkg _ -> pkg_import
   where
-    home_import   = findHomeModule hsc_env mod_name
+    all_opts = case mhome_unit of
+                Nothing -> other_fopts
+                Just home_unit -> (homeUnitId home_unit, fopts) : other_fopts
 
-    pkg_import    = findExposedPackageModule hsc_env mod_name mb_pkg
 
-    unqual_import = home_import
+    home_import = case mhome_unit of
+                   Just home_unit -> findHomeModule fc fopts home_unit mod_name
+                   Nothing -> pure $ NoPackage (panic "findImportedModule: no home-unit")
+
+
+    home_pkg_import (uid, opts)
+      -- If the module is reexported, then look for it as if it was from the perspective
+      -- of that package which reexports it.
+      | mod_name `Set.member` finder_reexportedModules opts =
+        findImportedModuleNoHsc fc opts ue (Just $ DefiniteHomeUnit uid Nothing) mod_name NoPkgQual
+      | mod_name `Set.member` finder_hiddenModules opts =
+        return (mkHomeHidden uid)
+      | otherwise =
+        findHomePackageModule fc opts uid mod_name
+
+    any_home_import = foldr orIfNotFound home_import (map home_pkg_import other_fopts)
+
+    pkg_import    = findExposedPackageModule fc fopts units  mod_name mb_pkg
+
+    unqual_import = any_home_import
                     `orIfNotFound`
-                    findExposedPackageModule hsc_env mod_name Nothing
+                    findExposedPackageModule fc fopts units mod_name NoPkgQual
+
+    units     = case mhome_unit of
+                  Nothing -> ue_units ue
+                  Just home_unit -> homeUnitEnv_units $ ue_findHomeUnitEnv (homeUnitId home_unit) ue
+    hpt_deps :: [UnitId]
+    hpt_deps  = homeUnitDepends units
+    other_fopts  = map (\uid -> (uid, initFinderOpts (homeUnitEnv_dflags (ue_findHomeUnitEnv uid ue)))) hpt_deps
 
 -- | Locate a plugin module requested by the user, for a compiler
 -- plugin.  This consults the same set of exposed packages as
 -- 'findImportedModule', unless @-hide-all-plugin-packages@ or
 -- @-plugin-package@ are specified.
-findPluginModule :: HscEnv -> ModuleName -> IO FindResult
-findPluginModule hsc_env mod_name =
-  findHomeModule hsc_env mod_name
+findPluginModule :: FinderCache -> FinderOpts -> UnitState -> Maybe HomeUnit -> ModuleName -> IO FindResult
+findPluginModule fc fopts units (Just home_unit) mod_name =
+  findHomeModule fc fopts home_unit mod_name
   `orIfNotFound`
-  findExposedPluginPackageModule hsc_env mod_name
+  findExposedPluginPackageModule fc fopts units mod_name
+findPluginModule fc fopts units Nothing mod_name =
+  findExposedPluginPackageModule fc fopts units mod_name
 
 -- | Locate a specific 'Module'.  The purpose of this function is to
 -- create a 'ModLocation' for a given 'Module', that is to find out
@@ -144,12 +213,15 @@ findPluginModule hsc_env mod_name =
 -- reading the interface for a module mentioned by another interface,
 -- for example (a "system import").
 
-findExactModule :: HscEnv -> InstalledModule -> IO InstalledFindResult
-findExactModule hsc_env mod =
-    let home_unit = hsc_home_unit hsc_env
-    in if isHomeInstalledModule home_unit mod
-       then findInstalledHomeModule hsc_env (moduleName mod)
-       else findPackageModule hsc_env mod
+findExactModule :: FinderCache -> FinderOpts ->  UnitEnvGraph FinderOpts -> UnitState -> Maybe HomeUnit -> InstalledModule -> IO InstalledFindResult
+findExactModule fc fopts other_fopts unit_state mhome_unit mod = do
+  case mhome_unit of
+    Just home_unit
+     | isHomeInstalledModule home_unit mod
+        -> findInstalledHomeModule fc fopts (homeUnitId home_unit) (moduleName mod)
+     | Just home_fopts <- unitEnv_lookup_maybe (moduleUnit mod) other_fopts
+        -> findInstalledHomeModule fc home_fopts (moduleUnit mod) (moduleName mod)
+    _ -> findPackageModule fc unit_state fopts mod
 
 -- -----------------------------------------------------------------------------
 -- Helpers
@@ -184,31 +256,26 @@ orIfNotFound this or_this = do
 -- been done.  Otherwise, do the lookup (with the IO action) and save
 -- the result in the finder cache and the module location cache (if it
 -- was successful.)
-homeSearchCache :: HscEnv -> ModuleName -> IO InstalledFindResult -> IO InstalledFindResult
-homeSearchCache hsc_env mod_name do_this = do
-  let home_unit = hsc_home_unit hsc_env
-      mod = mkHomeInstalledModule home_unit mod_name
-  modLocationCache hsc_env mod do_this
+homeSearchCache :: FinderCache -> UnitId -> ModuleName -> IO InstalledFindResult -> IO InstalledFindResult
+homeSearchCache fc home_unit mod_name do_this = do
+  let mod = mkModule home_unit mod_name
+  modLocationCache fc mod do_this
 
-findExposedPackageModule :: HscEnv -> ModuleName -> Maybe FastString
-                         -> IO FindResult
-findExposedPackageModule hsc_env mod_name mb_pkg
-  = findLookupResult hsc_env
-  $ lookupModuleWithSuggestions
-        (hsc_units hsc_env) mod_name mb_pkg
+findExposedPackageModule :: FinderCache -> FinderOpts -> UnitState -> ModuleName -> PkgQual -> IO FindResult
+findExposedPackageModule fc fopts units mod_name mb_pkg =
+  findLookupResult fc fopts
+    $ lookupModuleWithSuggestions units mod_name mb_pkg
 
-findExposedPluginPackageModule :: HscEnv -> ModuleName
-                               -> IO FindResult
-findExposedPluginPackageModule hsc_env mod_name
-  = findLookupResult hsc_env
-  $ lookupPluginModuleWithSuggestions
-        (hsc_units hsc_env) mod_name Nothing
+findExposedPluginPackageModule :: FinderCache -> FinderOpts -> UnitState -> ModuleName -> IO FindResult
+findExposedPluginPackageModule fc fopts units mod_name =
+  findLookupResult fc fopts
+    $ lookupPluginModuleWithSuggestions units mod_name NoPkgQual
 
-findLookupResult :: HscEnv -> LookupResult -> IO FindResult
-findLookupResult hsc_env r = case r of
+findLookupResult :: FinderCache -> FinderOpts -> LookupResult -> IO FindResult
+findLookupResult fc fopts r = case r of
      LookupFound m pkg_conf -> do
        let im = fst (getModuleInstantiation m)
-       r' <- findPackageModule_ hsc_env im (fst pkg_conf)
+       r' <- findPackageModule_ fc fopts im (fst pkg_conf)
        case r' of
         -- TODO: ghc -M is unlikely to do the right thing
         -- with just the location of the thing that was
@@ -241,7 +308,7 @@ findLookupResult hsc_env r = case r of
                           , fr_suggestions = [] })
      LookupNotFound suggest -> do
        let suggest'
-             | gopt Opt_HelpfulErrors (hsc_dflags hsc_env) = suggest
+             | finder_enableSuggestions fopts = suggest
              | otherwise = []
        return (NotFound{ fr_paths = [], fr_pkg = Nothing
                        , fr_pkgs_hidden = []
@@ -249,36 +316,40 @@ findLookupResult hsc_env r = case r of
                        , fr_unusables = []
                        , fr_suggestions = suggest' })
 
-modLocationCache :: HscEnv -> InstalledModule -> IO InstalledFindResult -> IO InstalledFindResult
-modLocationCache hsc_env mod do_this = do
-  m <- lookupFinderCache (hsc_FC hsc_env) mod
+modLocationCache :: FinderCache -> InstalledModule -> IO InstalledFindResult -> IO InstalledFindResult
+modLocationCache fc mod do_this = do
+  m <- lookupFinderCache fc mod
   case m of
     Just result -> return result
     Nothing     -> do
         result <- do_this
-        addToFinderCache (hsc_FC hsc_env) mod result
+        addToFinderCache fc mod result
         return result
 
+addModuleToFinder :: FinderCache -> Module -> ModLocation -> IO ()
+addModuleToFinder fc mod loc = do
+  let imod = toUnitId <$> mod
+  addToFinderCache fc imod (InstalledFound loc imod)
+
 -- This returns a module because it's more convenient for users
-addHomeModuleToFinder :: HscEnv -> ModuleName -> ModLocation -> IO Module
-addHomeModuleToFinder hsc_env mod_name loc = do
-  let home_unit = hsc_home_unit hsc_env
-      mod = mkHomeInstalledModule home_unit mod_name
-  addToFinderCache (hsc_FC hsc_env) mod (InstalledFound loc mod)
+addHomeModuleToFinder :: FinderCache -> HomeUnit -> ModuleName -> ModLocation -> IO Module
+addHomeModuleToFinder fc home_unit mod_name loc = do
+  let mod = mkHomeInstalledModule home_unit mod_name
+  addToFinderCache fc mod (InstalledFound loc mod)
   return (mkHomeModule home_unit mod_name)
 
-uncacheModule :: HscEnv -> ModuleName -> IO ()
-uncacheModule hsc_env mod_name = do
-  let home_unit = hsc_home_unit hsc_env
-      mod = mkHomeInstalledModule home_unit mod_name
-  removeFromFinderCache (hsc_FC hsc_env) mod
+uncacheModule :: FinderCache -> HomeUnit -> ModuleName -> IO ()
+uncacheModule fc home_unit mod_name = do
+  let mod = mkHomeInstalledModule home_unit mod_name
+  removeFromFinderCache fc mod
 
 -- -----------------------------------------------------------------------------
 --      The internal workers
 
-findHomeModule :: HscEnv -> ModuleName -> IO FindResult
-findHomeModule hsc_env mod_name = do
-  r <- findInstalledHomeModule hsc_env mod_name
+findHomeModule :: FinderCache -> FinderOpts -> HomeUnit -> ModuleName -> IO FindResult
+findHomeModule fc fopts  home_unit mod_name = do
+  let uid       = homeUnitAsUnit home_unit
+  r <- findInstalledHomeModule fc fopts (homeUnitId home_unit) mod_name
   return $ case r of
     InstalledFound loc _ -> Found loc (mkHomeModule home_unit mod_name)
     InstalledNoPackage _ -> NoPackage uid -- impossible
@@ -290,9 +361,32 @@ findHomeModule hsc_env mod_name = do
         fr_unusables = [],
         fr_suggestions = []
       }
- where
-  home_unit = hsc_home_unit hsc_env
-  uid       = homeUnitAsUnit home_unit
+
+mkHomeHidden :: UnitId -> FindResult
+mkHomeHidden uid =
+  NotFound { fr_paths = []
+           , fr_pkg = Just (RealUnit (Definite uid))
+           , fr_mods_hidden = [RealUnit (Definite uid)]
+           , fr_pkgs_hidden = []
+           , fr_unusables = []
+           , fr_suggestions = []}
+
+findHomePackageModule :: FinderCache -> FinderOpts -> UnitId -> ModuleName -> IO FindResult
+findHomePackageModule fc fopts  home_unit mod_name = do
+  let uid       = RealUnit (Definite home_unit)
+  r <- findInstalledHomeModule fc fopts home_unit mod_name
+  return $ case r of
+    InstalledFound loc _ -> Found loc (mkModule uid mod_name)
+    InstalledNoPackage _ -> NoPackage uid -- impossible
+    InstalledNotFound fps _ -> NotFound {
+        fr_paths = fps,
+        fr_pkg = Just uid,
+        fr_mods_hidden = [],
+        fr_pkgs_hidden = [],
+        fr_unusables = [],
+        fr_suggestions = []
+      }
+
 
 -- | Implements the search for a module name in the home package only.  Calling
 -- this function directly is usually *not* what you want; currently, it's used
@@ -310,51 +404,64 @@ findHomeModule hsc_env mod_name = do
 --
 --  4. Some special-case code in GHCi (ToDo: Figure out why that needs to
 --  call this.)
-findInstalledHomeModule :: HscEnv -> ModuleName -> IO InstalledFindResult
-findInstalledHomeModule hsc_env mod_name =
-   homeSearchCache hsc_env mod_name $
+findInstalledHomeModule :: FinderCache -> FinderOpts -> UnitId -> ModuleName -> IO InstalledFindResult
+findInstalledHomeModule fc fopts home_unit mod_name = do
+  homeSearchCache fc home_unit mod_name $
    let
-     dflags = hsc_dflags hsc_env
-     home_unit = hsc_home_unit hsc_env
-     home_path = importPaths dflags
-     hisuf = hiSuf dflags
-     mod = mkHomeInstalledModule home_unit mod_name
+     maybe_working_dir = finder_workingDirectory fopts
+     home_path = case maybe_working_dir of
+                  Nothing -> finder_importPaths fopts
+                  Just fp -> augmentImports fp (finder_importPaths fopts)
+     hi_dir_path =
+      case finder_hiDir fopts of
+        Just hiDir -> case maybe_working_dir of
+                        Nothing -> [hiDir]
+                        Just fp -> [fp </> hiDir]
+        Nothing -> home_path
+     hisuf = finder_hiSuf fopts
+     mod = mkModule home_unit mod_name
 
      source_exts =
-      [ ("hs",   mkHomeModLocationSearched dflags mod_name "hs")
-      , ("lhs",  mkHomeModLocationSearched dflags mod_name "lhs")
-      , ("hsig",  mkHomeModLocationSearched dflags mod_name "hsig")
-      , ("lhsig",  mkHomeModLocationSearched dflags mod_name "lhsig")
+      [ ("hs",    mkHomeModLocationSearched fopts mod_name "hs")
+      , ("lhs",   mkHomeModLocationSearched fopts mod_name "lhs")
+      , ("hsig",  mkHomeModLocationSearched fopts mod_name "hsig")
+      , ("lhsig", mkHomeModLocationSearched fopts mod_name "lhsig")
       ]
 
      -- we use mkHomeModHiOnlyLocation instead of mkHiOnlyModLocation so that
      -- when hiDir field is set in dflags, we know to look there (see #16500)
-     hi_exts = [ (hisuf,                mkHomeModHiOnlyLocation dflags mod_name)
-               , (addBootSuffix hisuf,  mkHomeModHiOnlyLocation dflags mod_name)
+     hi_exts = [ (hisuf,                mkHomeModHiOnlyLocation fopts mod_name)
+               , (addBootSuffix hisuf,  mkHomeModHiOnlyLocation fopts mod_name)
                ]
 
         -- In compilation manager modes, we look for source files in the home
         -- package because we can compile these automatically.  In one-shot
         -- compilation mode we look for .hi and .hi-boot files only.
-     exts | isOneShot (ghcMode dflags) = hi_exts
-          | otherwise                  = source_exts
+     (search_dirs, exts)
+          | finder_lookupHomeInterfaces fopts = (hi_dir_path, hi_exts)
+          | otherwise                         = (home_path, source_exts)
    in
 
-  -- special case for GHC.Prim; we won't find it in the filesystem.
-  -- This is important only when compiling the base package (where GHC.Prim
-  -- is a home module).
-  if mod `installedModuleEq` gHC_PRIM
-        then return (InstalledFound (error "GHC.Prim ModLocation") mod)
-        else searchPathExts home_path mod exts
+   -- special case for GHC.Prim; we won't find it in the filesystem.
+   -- This is important only when compiling the base package (where GHC.Prim
+   -- is a home module).
+   if mod `installedModuleEq` gHC_PRIM
+         then return (InstalledFound (error "GHC.Prim ModLocation") mod)
+         else searchPathExts search_dirs mod exts
 
+-- | Prepend the working directory to the search path.
+augmentImports :: FilePath -> [FilePath] -> [FilePath]
+augmentImports _work_dir [] = []
+augmentImports work_dir (fp:fps) | isAbsolute fp = fp : augmentImports work_dir fps
+                                 | otherwise     = (work_dir </> fp) : augmentImports work_dir fps
 
 -- | Search for a module in external packages only.
-findPackageModule :: HscEnv -> InstalledModule -> IO InstalledFindResult
-findPackageModule hsc_env mod = do
+findPackageModule :: FinderCache -> UnitState -> FinderOpts -> InstalledModule -> IO InstalledFindResult
+findPackageModule fc unit_state fopts mod = do
   let pkg_id = moduleUnit mod
-  case lookupUnitId (hsc_units hsc_env) pkg_id of
+  case lookupUnitId unit_state pkg_id of
      Nothing -> return (InstalledNoPackage pkg_id)
-     Just u  -> findPackageModule_ hsc_env mod u
+     Just u  -> findPackageModule_ fc fopts mod u
 
 -- | Look up the interface file associated with module @mod@.  This function
 -- requires a few invariants to be upheld: (1) the 'Module' in question must
@@ -363,39 +470,41 @@ findPackageModule hsc_env mod = do
 -- the 'UnitInfo' must be consistent with the unit id in the 'Module'.
 -- The redundancy is to avoid an extra lookup in the package state
 -- for the appropriate config.
-findPackageModule_ :: HscEnv -> InstalledModule -> UnitInfo -> IO InstalledFindResult
-findPackageModule_ hsc_env mod pkg_conf =
-  ASSERT2( moduleUnit mod == unitId pkg_conf, ppr (moduleUnit mod) <+> ppr (unitId pkg_conf) )
-  modLocationCache hsc_env mod $
+findPackageModule_ :: FinderCache -> FinderOpts -> InstalledModule -> UnitInfo -> IO InstalledFindResult
+findPackageModule_ fc fopts mod pkg_conf = do
+  massertPpr (moduleUnit mod == unitId pkg_conf)
+             (ppr (moduleUnit mod) <+> ppr (unitId pkg_conf))
+  modLocationCache fc mod $
 
-  -- special case for GHC.Prim; we won't find it in the filesystem.
-  if mod `installedModuleEq` gHC_PRIM
-        then return (InstalledFound (error "GHC.Prim ModLocation") mod)
-        else
+    -- special case for GHC.Prim; we won't find it in the filesystem.
+    if mod `installedModuleEq` gHC_PRIM
+          then return (InstalledFound (error "GHC.Prim ModLocation") mod)
+          else
 
-  let
-     dflags = hsc_dflags hsc_env
-     tag = waysBuildTag (ways dflags)
+    let
+       tag = waysBuildTag (finder_ways fopts)
 
-           -- hi-suffix for packages depends on the build tag.
-     package_hisuf | null tag  = "hi"
-                   | otherwise = tag ++ "_hi"
+             -- hi-suffix for packages depends on the build tag.
+       package_hisuf | null tag  = "hi"
+                     | otherwise = tag ++ "_hi"
 
-     mk_hi_loc = mkHiOnlyModLocation dflags package_hisuf
+       package_dynhisuf = waysBuildTag (addWay WayDyn (finder_ways fopts)) ++ "_hi"
 
-     import_dirs = map ST.unpack $ unitImportDirs pkg_conf
-      -- we never look for a .hi-boot file in an external package;
-      -- .hi-boot files only make sense for the home package.
-  in
-  case import_dirs of
-    [one] | MkDepend <- ghcMode dflags -> do
-          -- there's only one place that this .hi file can be, so
-          -- don't bother looking for it.
-          let basename = moduleNameSlashes (moduleName mod)
-          loc <- mk_hi_loc one basename
-          return (InstalledFound loc mod)
-    _otherwise ->
-          searchPathExts import_dirs mod [(package_hisuf, mk_hi_loc)]
+       mk_hi_loc = mkHiOnlyModLocation fopts package_hisuf package_dynhisuf
+
+       import_dirs = map ST.unpack $ unitImportDirs pkg_conf
+        -- we never look for a .hi-boot file in an external package;
+        -- .hi-boot files only make sense for the home package.
+    in
+    case import_dirs of
+      [one] | finder_bypassHiFileCheck fopts ->
+            -- there's only one place that this .hi file can be, so
+            -- don't bother looking for it.
+            let basename = moduleNameSlashes (moduleName mod)
+                loc = mk_hi_loc one basename
+            in return $ InstalledFound loc mod
+      _otherwise ->
+            searchPathExts import_dirs mod [(package_hisuf, mk_hi_loc)]
 
 -- -----------------------------------------------------------------------------
 -- General path searching
@@ -403,8 +512,8 @@ findPackageModule_ hsc_env mod pkg_conf =
 searchPathExts :: [FilePath]      -- paths to search
                -> InstalledModule -- module name
                -> [ (
-                     FileExt,                                -- suffix
-                     FilePath -> BaseName -> IO ModLocation  -- action
+                     FileExt,                             -- suffix
+                     FilePath -> BaseName -> ModLocation  -- action
                     )
                   ]
                -> IO InstalledFindResult
@@ -413,7 +522,7 @@ searchPathExts paths mod exts = search to_search
   where
     basename = moduleNameSlashes (moduleName mod)
 
-    to_search :: [(FilePath, IO ModLocation)]
+    to_search :: [(FilePath, ModLocation)]
     to_search = [ (file, fn path basename)
                 | path <- paths,
                   (ext,fn) <- exts,
@@ -424,16 +533,17 @@ searchPathExts paths mod exts = search to_search
 
     search [] = return (InstalledNotFound (map fst to_search) (Just (moduleUnit mod)))
 
-    search ((file, mk_result) : rest) = do
+    search ((file, loc) : rest) = do
       b <- doesFileExist file
       if b
-        then do { loc <- mk_result; return (InstalledFound loc mod) }
+        then return $ InstalledFound loc mod
         else search rest
 
-mkHomeModLocationSearched :: DynFlags -> ModuleName -> FileExt
-                          -> FilePath -> BaseName -> IO ModLocation
-mkHomeModLocationSearched dflags mod suff path basename =
-  mkHomeModLocation2 dflags mod (path </> basename) suff
+mkHomeModLocationSearched :: FinderOpts -> ModuleName -> FileExt
+                          -> FilePath -> BaseName -> ModLocation
+mkHomeModLocationSearched fopts mod suff path basename =
+  mkHomeModLocation2 fopts mod (path </> basename) suff
+
 
 -- -----------------------------------------------------------------------------
 -- Constructing a home module location
@@ -468,49 +578,59 @@ mkHomeModLocationSearched dflags mod suff path basename =
 -- ext
 --      The filename extension of the source file (usually "hs" or "lhs").
 
-mkHomeModLocation :: DynFlags -> ModuleName -> FilePath -> IO ModLocation
-mkHomeModLocation dflags mod src_filename = do
+mkHomeModLocation :: FinderOpts -> ModuleName -> FilePath -> ModLocation
+mkHomeModLocation dflags mod src_filename =
    let (basename,extension) = splitExtension src_filename
-   mkHomeModLocation2 dflags mod basename extension
+   in mkHomeModLocation2 dflags mod basename extension
 
-mkHomeModLocation2 :: DynFlags
+mkHomeModLocation2 :: FinderOpts
                    -> ModuleName
                    -> FilePath  -- Of source module, without suffix
                    -> String    -- Suffix
-                   -> IO ModLocation
-mkHomeModLocation2 dflags mod src_basename ext = do
+                   -> ModLocation
+mkHomeModLocation2 fopts mod src_basename ext =
    let mod_basename = moduleNameSlashes mod
 
-       obj_fn = mkObjPath  dflags src_basename mod_basename
-       hi_fn  = mkHiPath   dflags src_basename mod_basename
-       hie_fn = mkHiePath  dflags src_basename mod_basename
+       obj_fn = mkObjPath  fopts src_basename mod_basename
+       dyn_obj_fn = mkDynObjPath  fopts src_basename mod_basename
+       hi_fn  = mkHiPath   fopts src_basename mod_basename
+       dyn_hi_fn  = mkDynHiPath   fopts src_basename mod_basename
+       hie_fn = mkHiePath  fopts src_basename mod_basename
 
-   return (ModLocation{ ml_hs_file   = Just (src_basename <.> ext),
+   in (ModLocation{ ml_hs_file   = Just (src_basename <.> ext),
                         ml_hi_file   = hi_fn,
+                        ml_dyn_hi_file = dyn_hi_fn,
                         ml_obj_file  = obj_fn,
+                        ml_dyn_obj_file = dyn_obj_fn,
                         ml_hie_file  = hie_fn })
 
-mkHomeModHiOnlyLocation :: DynFlags
+mkHomeModHiOnlyLocation :: FinderOpts
                         -> ModuleName
                         -> FilePath
                         -> BaseName
-                        -> IO ModLocation
-mkHomeModHiOnlyLocation dflags mod path basename = do
-   loc <- mkHomeModLocation2 dflags mod (path </> basename) ""
-   return loc { ml_hs_file = Nothing }
+                        -> ModLocation
+mkHomeModHiOnlyLocation fopts mod path basename =
+   let loc = mkHomeModLocation2 fopts mod (path </> basename) ""
+   in loc { ml_hs_file = Nothing }
 
-mkHiOnlyModLocation :: DynFlags -> Suffix -> FilePath -> String
-                    -> IO ModLocation
-mkHiOnlyModLocation dflags hisuf path basename
- = do let full_basename = path </> basename
-          obj_fn = mkObjPath  dflags full_basename basename
-          hie_fn = mkHiePath  dflags full_basename basename
-      return ModLocation{    ml_hs_file   = Nothing,
+-- This function is used to make a ModLocation for a package module. Hence why
+-- we explicitly pass in the interface file suffixes.
+mkHiOnlyModLocation :: FinderOpts -> Suffix -> Suffix -> FilePath -> String
+                    -> ModLocation
+mkHiOnlyModLocation fopts hisuf dynhisuf path basename
+ = let full_basename = path </> basename
+       obj_fn = mkObjPath fopts full_basename basename
+       dyn_obj_fn = mkDynObjPath fopts full_basename basename
+       hie_fn = mkHiePath fopts full_basename basename
+   in ModLocation{    ml_hs_file   = Nothing,
                              ml_hi_file   = full_basename <.> hisuf,
                                 -- Remove the .hi-boot suffix from
                                 -- hi_file, if it had one.  We always
                                 -- want the name of the real .hi file
                                 -- in the ml_hi_file field.
+                             ml_dyn_obj_file = dyn_obj_fn,
+                             -- MP: TODO
+                             ml_dyn_hi_file  = full_basename <.> dynhisuf,
                              ml_obj_file  = obj_fn,
                              ml_hie_file  = hie_fn
                   }
@@ -518,14 +638,29 @@ mkHiOnlyModLocation dflags hisuf path basename
 -- | Constructs the filename of a .o file for a given source file.
 -- Does /not/ check whether the .o file exists
 mkObjPath
-  :: DynFlags
+  :: FinderOpts
   -> FilePath           -- the filename of the source file, minus the extension
   -> String             -- the module name with dots replaced by slashes
   -> FilePath
-mkObjPath dflags basename mod_basename = obj_basename <.> osuf
+mkObjPath fopts basename mod_basename = obj_basename <.> osuf
   where
-                odir = objectDir dflags
-                osuf = objectSuf dflags
+                odir = finder_objectDir fopts
+                osuf = finder_objectSuf fopts
+
+                obj_basename | Just dir <- odir = dir </> mod_basename
+                             | otherwise        = basename
+
+-- | Constructs the filename of a .dyn_o file for a given source file.
+-- Does /not/ check whether the .dyn_o file exists
+mkDynObjPath
+  :: FinderOpts
+  -> FilePath           -- the filename of the source file, minus the extension
+  -> String             -- the module name with dots replaced by slashes
+  -> FilePath
+mkDynObjPath fopts basename mod_basename = obj_basename <.> dynosuf
+  where
+                odir = finder_objectDir fopts
+                dynosuf = finder_dynObjectSuf fopts
 
                 obj_basename | Just dir <- odir = dir </> mod_basename
                              | otherwise        = basename
@@ -534,14 +669,29 @@ mkObjPath dflags basename mod_basename = obj_basename <.> osuf
 -- | Constructs the filename of a .hi file for a given source file.
 -- Does /not/ check whether the .hi file exists
 mkHiPath
-  :: DynFlags
+  :: FinderOpts
   -> FilePath           -- the filename of the source file, minus the extension
   -> String             -- the module name with dots replaced by slashes
   -> FilePath
-mkHiPath dflags basename mod_basename = hi_basename <.> hisuf
+mkHiPath fopts basename mod_basename = hi_basename <.> hisuf
  where
-                hidir = hiDir dflags
-                hisuf = hiSuf dflags
+                hidir = finder_hiDir fopts
+                hisuf = finder_hiSuf fopts
+
+                hi_basename | Just dir <- hidir = dir </> mod_basename
+                            | otherwise         = basename
+
+-- | Constructs the filename of a .dyn_hi file for a given source file.
+-- Does /not/ check whether the .dyn_hi file exists
+mkDynHiPath
+  :: FinderOpts
+  -> FilePath           -- the filename of the source file, minus the extension
+  -> String             -- the module name with dots replaced by slashes
+  -> FilePath
+mkDynHiPath fopts basename mod_basename = hi_basename <.> dynhisuf
+ where
+                hidir = finder_hiDir fopts
+                dynhisuf = finder_dynHiSuf fopts
 
                 hi_basename | Just dir <- hidir = dir </> mod_basename
                             | otherwise         = basename
@@ -549,14 +699,14 @@ mkHiPath dflags basename mod_basename = hi_basename <.> hisuf
 -- | Constructs the filename of a .hie file for a given source file.
 -- Does /not/ check whether the .hie file exists
 mkHiePath
-  :: DynFlags
+  :: FinderOpts
   -> FilePath           -- the filename of the source file, minus the extension
   -> String             -- the module name with dots replaced by slashes
   -> FilePath
-mkHiePath dflags basename mod_basename = hie_basename <.> hiesuf
+mkHiePath fopts basename mod_basename = hie_basename <.> hiesuf
  where
-                hiedir = hieDir dflags
-                hiesuf = hieSuf dflags
+                hiedir = finder_hieDir fopts
+                hiesuf = finder_hieSuf fopts
 
                 hie_basename | Just dir <- hiedir = dir </> mod_basename
                              | otherwise          = basename
@@ -570,14 +720,14 @@ mkHiePath dflags basename mod_basename = hie_basename <.> hiesuf
 -- from other available information, and they're only rarely needed.
 
 mkStubPaths
-  :: DynFlags
+  :: FinderOpts
   -> ModuleName
   -> ModLocation
   -> FilePath
 
-mkStubPaths dflags mod location
+mkStubPaths fopts mod location
   = let
-        stubdir = stubDir dflags
+        stubdir = finder_stubDir fopts
 
         mod_basename = moduleNameSlashes mod
         src_basename = dropExtension $ expectJust "mkStubPaths"
