@@ -89,7 +89,11 @@ instance NFData   GhcMode
 --   This instructs package configuration functions (such as 'configurePackage')
 --   to enable the @profiling@ Cabal flag when processing @rts.cabal@ and
 --   building RTS with profiling information.
-data ConfigurationInfo = Setup | Flags deriving (Eq, Generic, Show)
+data ConfigurationInfo
+  = Setup
+  | Flags
+  | Install
+  deriving (Eq, Generic, Show)
 
 instance Binary   ConfigurationInfo
 instance Hashable ConfigurationInfo
@@ -140,6 +144,8 @@ instance Binary   Win32TarballsMode
 instance Hashable Win32TarballsMode
 instance NFData   Win32TarballsMode
 
+
+
 -- | A 'Builder' is a (usually external) command invoked in a separate process
 -- via 'cmd'. Here are some examples:
 -- * 'Alex' is a lexical analyser generator that builds @Lexer.hs@ from @Lexer.x@.
@@ -193,18 +199,20 @@ instance NFData   Builder
 -- 'Stage' and GHC 'Package').
 builderProvenance :: Builder -> Maybe Context
 builderProvenance = \case
-    DeriveConstants  -> context Stage0 deriveConstants
-    GenApply         -> context Stage0 genapply
-    GenPrimopCode    -> context Stage0 genprimopcode
-    Ghc _ Stage0     -> Nothing
-    Ghc _ stage      -> context (pred stage) ghc
-    GhcPkg _ Stage0  -> Nothing
-    GhcPkg _ s       -> context (pred s) ghcPkg
+    DeriveConstants  -> context stage0Boot deriveConstants
+    GenApply         -> context stage0Boot genapply
+    GenPrimopCode    -> context stage0Boot genprimopcode
+    Ghc _ (Stage0 {})-> Nothing
+    Ghc _ stage      -> context (predStage stage) ghc
+    GhcPkg _ (Stage0 {}) -> Nothing
+    GhcPkg _ s       -> context (predStage s) ghcPkg
     Haddock _        -> context Stage1 haddock
+    Hsc2Hs _         -> context stage0Boot hsc2hs
+    Unlit            -> context stage0Boot unlit
+
+    -- Never used
     Hpc              -> context Stage1 hpcBin
-    Hp2Ps            -> context Stage0 hp2ps
-    Hsc2Hs _         -> context Stage0 hsc2hs
-    Unlit            -> context Stage0 unlit
+    Hp2Ps            -> context stage0Boot hp2ps
     _                -> Nothing
   where
     context s p = Just $ vanillaContext s p
@@ -220,24 +228,22 @@ instance H.Builder Builder where
         Autoreconf dir -> return [dir -/- "configure.ac"]
         Configure  dir -> return [dir -/- "configure"]
 
-        Ghc _ Stage0 -> do
+        Ghc _ (Stage0 {}) -> do
           -- Read the boot GHC version here to make sure we rebuild when it
           -- changes (#18001).
           _bootGhcVersion <- setting GhcVersion
-          includesDependencies Stage0
+          pure []
         Ghc _ stage -> do
             root <- buildRoot
-            touchyPath <- programPath (vanillaContext Stage0 touchy)
+            touchyPath <- programPath (vanillaContext (Stage0 InTreeLibs) touchy)
             unlitPath  <- builderPath Unlit
-            ghcgens <- includesDependencies stage
 
             -- GHC from the previous stage is used to build artifacts in the
             -- current stage. Need the previous stage's GHC deps.
-            ghcdeps <- ghcBinDeps (pred stage)
+            ghcdeps <- ghcBinDeps (predStage stage)
 
             return $ [ unlitPath ]
                   ++ ghcdeps
-                  ++ ghcgens
                   ++ [ touchyPath          | windowsHost ]
                   ++ [ root -/- mingwStamp | windowsHost ]
                      -- proxy for the entire mingw toolchain that
@@ -272,11 +278,10 @@ instance H.Builder Builder where
             return stdout
         Testsuite GetExtraDeps -> do
           path <- builderPath builder
-          withResources buildResources $ do
-              -- The testsuite driver reports the dependencies on stderr
-              -- buildArgs should include --only-report-hadrian-deps at this point
-              Stderr stderr <- cmd' [path] buildArgs
-              return stderr
+          withResources buildResources $
+              withTempFile $ \temp -> do
+                () <- cmd' [path] (buildArgs ++ ["--only-report-hadrian-deps", temp])
+                readFile' temp
         Git ListFiles -> do
           path <- builderPath builder
           withResources buildResources $ do
@@ -301,8 +306,6 @@ instance H.Builder Builder where
                 msgIn  = "[runBuilderWith] Exactly one input file expected."
                 output = fromSingleton msgOut buildOutputs
                 msgOut = "[runBuilderWith] Exactly one output file expected."
-                -- Suppress stdout depending on the Shake's verbosity setting.
-                echo = EchoStdout (verbosity >= Loud)
                 -- Capture stdout and write it to the output file.
                 captureStdout = do
                     Stdout stdout <- cmd' [path] buildArgs
@@ -310,21 +313,21 @@ instance H.Builder Builder where
             case builder of
                 Ar Pack _ -> do
                     useTempFile <- flag ArSupportsAtFile
-                    if useTempFile then runAr                path buildArgs
-                                   else runArWithoutTempFile path buildArgs
+                    if useTempFile then runAr                path buildArgs buildInputs
+                                   else runArWithoutTempFile path buildArgs buildInputs
 
-                Ar Unpack _ -> cmd' echo [Cwd output] [path] buildArgs
+                Ar Unpack _ -> cmd' [Cwd output] [path] buildArgs
 
                 Autoreconf dir -> do
                   bash <- bashPath
-                  cmd' echo [Cwd dir] [bash, path] buildArgs
+                  cmd' [Cwd dir] [bash, path] buildArgs
 
                 Configure  dir -> do
                     -- Inject /bin/bash into `libtool`, instead of /bin/sh,
                     -- otherwise Windows breaks. TODO: Figure out why.
                     bash <- bashPath
                     let env = AddEnv "CONFIG_SHELL" bash
-                    cmd' echo env [Cwd dir] ["sh", path] buildOptions buildArgs
+                    cmd' env [Cwd dir] ["sh", path] buildOptions buildArgs
 
                 GenApply -> captureStdout
 
@@ -343,21 +346,23 @@ instance H.Builder Builder where
                     cmd' (Stdin pkgDesc) [path] (buildArgs ++ ["-"])
 
                 GhcPkg Unregister _ -> do
-                    Exit _ <- cmd' echo [path] (buildArgs ++ [input])
+                    -- unregistering is allowed to fail (e.g. when a package
+                    -- isn't already present)
+                    Exit _ <- cmd' [path] (buildArgs ++ [input])
                     return ()
 
                 HsCpp    -> captureStdout
 
-                Make dir -> cmd' echo path ["-C", dir] buildArgs
+                Make dir -> cmd' path ["-C", dir] buildArgs
 
                 Makeinfo -> do
-                  cmd' echo [path] "--no-split" [ "-o", output] [input]
+                  cmd' [path] "--no-split" [ "-o", output] [input]
 
                 Xelatex   ->
                   -- xelatex produces an incredible amount of output, almost
                   -- all of which is useless. Suppress it unless user
                   -- requests a loud build.
-                  if verbosity >= Loud
+                  if verbosity >= Diagnostic
                     then cmd' [Cwd output] [path] buildArgs
                     else do (Stdouterr out, Exit code) <- cmd' [Cwd output] [path] buildArgs
                             when (code /= ExitSuccess) $ do
@@ -367,16 +372,16 @@ instance H.Builder Builder where
 
                 Makeindex -> unit $ cmd' [Cwd output] [path] (buildArgs ++ [input])
 
-                Tar _ -> cmd' buildOptions echo [path] buildArgs
+                Tar _ -> cmd' buildOptions [path] buildArgs
 
                 -- RunTest produces a very large amount of (colorised) output;
                 -- Don't attempt to capture it.
                 Testsuite RunTest -> do
-                  Exit code <- cmd echo [path] buildArgs
+                  Exit code <- cmd [path] buildArgs
                   when (code /= ExitSuccess) $ do
                     fail "tests failed"
 
-                _  -> cmd' echo [path] buildArgs
+                _  -> cmd' [path] buildArgs
 
 -- TODO: Some builders are required only on certain platforms. For example,
 -- 'Objdump' is only required on OpenBSD and AIX. Add support for platform
@@ -385,21 +390,25 @@ instance H.Builder Builder where
 isOptional :: Builder -> Bool
 isOptional = \case
     Objdump  -> True
+    -- alex and happy are not required when building source distributions
+    -- and ./configure will complain if they are not available when building in-tree
+    Happy    -> True
+    Alex     -> True
     _        -> False
 
 -- | Determine the location of a system 'Builder'.
 systemBuilderPath :: Builder -> Action FilePath
 systemBuilderPath builder = case builder of
     Alex            -> fromKey "alex"
-    Ar _ Stage0     -> fromKey "system-ar"
+    Ar _ (Stage0 {})-> fromKey "system-ar"
     Ar _ _          -> fromKey "ar"
     Autoreconf _    -> stripExe =<< fromKey "autoreconf"
-    Cc  _  Stage0   -> fromKey "system-cc"
+    Cc  _  (Stage0 {}) -> fromKey "system-cc"
     Cc  _  _        -> fromKey "cc"
     -- We can't ask configure for the path to configure!
     Configure _     -> return "configure"
-    Ghc _  Stage0   -> fromKey "system-ghc"
-    GhcPkg _ Stage0 -> fromKey "system-ghc-pkg"
+    Ghc _  (Stage0 {})   -> fromKey "system-ghc"
+    GhcPkg _ (Stage0 {}) -> fromKey "system-ghc-pkg"
     Happy           -> fromKey "happy"
     HsCpp           -> fromKey "hs-cpp"
     Ld _            -> fromKey "ld"
@@ -411,7 +420,7 @@ systemBuilderPath builder = case builder of
     -- parameters. E.g. building a cross-compiler on and for x86_64
     -- which will target ppc64 means that MergeObjects Stage0 will use
     -- x86_64 linker and MergeObject _ will use ppc64 linker.
-    MergeObjects Stage0 -> fromKey "system-merge-objects"
+    MergeObjects (Stage0 {}) -> fromKey "system-merge-objects"
     MergeObjects _  -> fromKey "merge-objects"
     Make _          -> fromKey "make"
     Makeinfo        -> fromKey "makeinfo"
@@ -427,6 +436,7 @@ systemBuilderPath builder = case builder of
     Xelatex         -> fromKey "xelatex"
     Makeindex       -> fromKey "makeindex"
     Win32Tarballs _ -> fromKey "python"
+    Cabal _ _       -> fromKey "cabal"
     _               -> error $ "No entry for " ++ show builder ++ inCfg
   where
     inCfg = " in " ++ quote configFile ++ " file."
@@ -477,7 +487,107 @@ applyPatch dir patch = do
     putBuild $ "| Apply patch " ++ file
     quietly $ cmd' [Cwd dir, FileStdin file] [path, "-p0"]
 
--- | Wrapper for 'cmd' that makes sure we include both stdout and stderr in
---   Shake's output when any of our builder commands fail.
-cmd' :: (Partial, CmdArguments args) => args :-> Action r
-cmd' = cmd [WithStderr True, WithStdout True]
+-- Note [cmd wrapper]
+-- ~~~~~~~~~~~~~~~~~~
+-- `cmd'` is a wrapper for Shake's `cmd` that allows us to customize what is
+-- output in the terminal in case of failure.
+--
+-- However `cmd` is quite a complex function because:
+--
+--  1) it relies on a CmdArguments type class to be variadic -- it can be called
+--  with any number of arguments, as long as they are valid arguments -- and to
+--  return either "Action r" or "IO r".
+--
+--  2) its behavior depends on the returned "r" type! In particular, if it has
+--  to return a value of type Exit or ExitCode, then it doesn't raise an
+--  exception if the exit code isn't 0! It also doesn't echo the command
+--  stdout/stderr if it is requested in a Stdout/Stderr/Stdouterr result. Result
+--  types are handled via CmdResult type class.
+--
+-- To wrap `cmd` while keeping its behavior, we need to replicate some of these
+-- type classes.
+--
+--  1) CmdWrap corresponds to CmdArguments except that we do our own stuff in
+--  the base case (i.e. in the instance for `Action r`).
+--
+--  2) Sadly CmdResult internals aren't exposed by Shake, so when we get a
+--  `CmdResult r => r` we can't tell anything about `r`. In particular, we can't
+--  tell if an Exit or ExitCode value is returned in `r`. So we use our own
+--  HasExit type class to provide the `hasExit` predicate that tells us if we
+--  should throw an exception as `cmd` would do in case of failure or not.
+
+
+-- | Wrapper for Shake's 'cmd'
+--
+-- See Note [cmd wrapper]
+cmd' :: (Partial, CmdWrap args) => args :-> Action r
+cmd' = cmdArgs mempty
+
+
+-- See Note [cmd wrapper]
+class HasExit a where
+  -- | Indicate if `a` is Exit or ExitCode
+  -- See Note [cmd wrapper]
+  hasExit :: a -> Bool
+
+instance HasExit ExitCode      where hasExit = const True
+instance HasExit Exit          where hasExit = const True
+instance HasExit ()            where hasExit = const False
+instance HasExit (Stdouterr a) where hasExit = const False
+instance HasExit (Stdout    a) where hasExit = const False
+
+instance (HasExit a, HasExit b) => HasExit (a,b) where
+  hasExit (a,b) = hasExit a || hasExit b
+instance (HasExit a, HasExit b, HasExit c) => HasExit (a,b,c) where
+  hasExit (a,b,c) = hasExit a || hasExit b || hasExit c
+
+class CmdWrap t where
+  cmdArgs :: Partial => CmdArgument -> t
+
+instance (IsCmdArgument a, CmdWrap r) => CmdWrap (a -> r) where
+  cmdArgs xs x = cmdArgs $ xs `mappend` toCmdArgument x
+
+instance CmdWrap CmdArgument where
+  cmdArgs = id
+
+instance (HasExit r, CmdResult r) => CmdWrap (Action r) where
+  cmdArgs (CmdArgument x) = do
+    verbosity <- getVerbosity
+
+    let real_args = mconcat
+          [ -- don't print stderr and stdout in command failure exception
+            toCmdArgument (WithStderr False)
+          , toCmdArgument (WithStdout False)
+            -- caller specified arguments come last to allow them to overload
+            -- the previous ones.
+          , CmdArgument x
+          ]
+    (Stdout out, Stderr err, cmdline :: CmdLine, Exit code, r :: r) <- cmd real_args
+
+    if hasExit r
+      -- if the caller queries the exit code of the command, we don't do
+      -- anything here. In particular we don't throw an exception.
+      -- (this is used e.g. to allow ghc-pkg to fail to unregister)
+      -- See Note [cmd wrapper]
+      then pure r
+      else do
+        -- In every case, we only print both command outputs (stdout/stderr)
+        -- onto Hadrian's stderr because Hadrian's stdout may be piped into
+        -- another process and we don't want random command output to break
+        -- this.
+        --
+        -- For example, the result of "hadrian tool:ghc/Main.hs --flavour=ghc-in-ghci"
+        -- is directly passed as arguments for ghc in "hadrian/ghci-cabal" script.
+        let dump x = liftIO (BSL.hPutStr stderr x)
+        case code of
+          ExitSuccess   -> do
+            -- Suppress stdout/stderr depending on Shake's verbosity setting
+            when (verbosity >  Silent)  (dump err)
+            when (verbosity >= Verbose) (dump out)
+            pure r
+          ExitFailure i -> do
+            putError ("Command line: " ++ fromCmdLine cmdline)
+            putError ("===> Command failed with error code: " ++ show i)
+            dump err
+            dump out
+            error "Command failed"
