@@ -145,8 +145,6 @@ module Data.ByteString.Lazy (
         groupBy,
         inits,
         tails,
-        initsNE,
-        tailsNE,
         stripPrefix,
         stripSuffix,
 
@@ -230,12 +228,10 @@ import Prelude hiding
     ,getContents,getLine,putStr,putStrLn ,zip,zipWith,unzip,notElem)
 
 import qualified Data.List              as List
-import qualified Data.List.NonEmpty     as NE
-import Data.List.NonEmpty (NonEmpty(..))
 import qualified Data.Bifunctor         as BF
 import qualified Data.ByteString        as P  (ByteString) -- type name only
 import qualified Data.ByteString        as S  -- S for strict (hmm...)
-import qualified Data.ByteString.Internal.Type as S
+import qualified Data.ByteString.Internal as S
 import qualified Data.ByteString.Unsafe as S
 import qualified Data.ByteString.Lazy.Internal.Deque as D
 import Data.ByteString.Lazy.Internal
@@ -346,30 +342,22 @@ snoc cs w = foldrChunks Chunk (singleton w) cs
 {-# INLINE snoc #-}
 
 -- | /O(1)/ Extract the first element of a ByteString, which must be non-empty.
---
--- This is a partial function, consider using 'uncons' instead.
 head :: HasCallStack => ByteString -> Word8
 head Empty       = errorEmptyList "head"
 head (Chunk c _) = S.unsafeHead c
 {-# INLINE head #-}
 
--- | /O(1)/ Extract the 'head' and 'tail' of a ByteString, returning 'Nothing'
+-- | /O(1)/ Extract the head and tail of a ByteString, returning Nothing
 -- if it is empty.
 uncons :: ByteString -> Maybe (Word8, ByteString)
 uncons Empty = Nothing
-uncons (Chunk c cs) = case S.length c of
-  -- Don't move this test inside of the Just or (,).
-  -- We don't want to allocate a thunk to put inside of the tuple!
-  -- And if "let !tl = ... in Just (..., tl)" seems more appealing,
-  -- remember that this function must remain lazy in cs.
-  1 -> Just (S.unsafeHead c, cs)
-  _ -> Just (S.unsafeHead c, Chunk (S.unsafeTail c) cs)
+uncons (Chunk c cs)
+    = Just (S.unsafeHead c,
+            if S.length c == 1 then cs else Chunk (S.unsafeTail c) cs)
 {-# INLINE uncons #-}
 
 -- | /O(1)/ Extract the elements after the head of a ByteString, which must be
 -- non-empty.
---
--- This is a partial function, consider using 'uncons' instead.
 tail :: HasCallStack => ByteString -> ByteString
 tail Empty          = errorEmptyList "tail"
 tail (Chunk c cs)
@@ -379,8 +367,6 @@ tail (Chunk c cs)
 
 -- | /O(n\/c)/ Extract the last element of a ByteString, which must be finite
 -- and non-empty.
---
--- This is a partial function, consider using 'unsnoc' instead.
 last :: HasCallStack => ByteString -> Word8
 last Empty          = errorEmptyList "last"
 last (Chunk c0 cs0) = go c0 cs0
@@ -388,9 +374,7 @@ last (Chunk c0 cs0) = go c0 cs0
         go _ (Chunk c cs) = go c cs
 -- XXX Don't inline this. Something breaks with 6.8.2 (haven't investigated yet)
 
--- | /O(n\/c)/ Returns all the elements of a 'ByteString' except the last one.
---
--- This is a partial function, consider using 'unsnoc' instead.
+-- | /O(n\/c)/ Return all the elements of a 'ByteString' except the last one.
 init :: HasCallStack => ByteString -> ByteString
 init Empty          = errorEmptyList "init"
 init (Chunk c0 cs0) = go c0 cs0
@@ -398,7 +382,7 @@ init (Chunk c0 cs0) = go c0 cs0
                    | otherwise       = Chunk (S.unsafeInit c) Empty
         go c (Chunk c' cs)           = Chunk c (go c' cs)
 
--- | /O(n\/c)/ Extract the 'init' and 'last' of a ByteString, returning 'Nothing'
+-- | /O(n\/c)/ Extract the 'init' and 'last' of a ByteString, returning Nothing
 -- if it is empty.
 --
 -- * It is no faster than using 'init' and 'last'
@@ -442,11 +426,9 @@ intersperse w (Chunk c cs) = Chunk (S.intersperse w c)
                                    (foldrChunks (Chunk . intersperse') Empty cs)
   where intersperse' :: P.ByteString -> P.ByteString
         intersperse' (S.BS fp l) =
-          S.unsafeCreateFp (2*l) $ \fp' ->
-            S.unsafeWithForeignPtr fp' $ \p' ->
-              S.unsafeWithForeignPtr fp $ \p -> do
-                poke p' w
-                S.c_intersperse (p' `plusPtr` 1) p (fromIntegral l) w
+          S.unsafeCreate (2*l) $ \p' -> S.unsafeWithForeignPtr fp $ \p -> do
+            poke p' w
+            S.c_intersperse (p' `plusPtr` 1) p (fromIntegral l) w
 
 -- | The 'transpose' function transposes the rows and columns of its
 -- 'ByteString' argument.
@@ -1141,8 +1123,6 @@ intercalate s = concat . List.intersperse s
 -- Indexing ByteStrings
 
 -- | /O(c)/ 'ByteString' index (subscript) operator, starting from 0.
---
--- This is a partial function, consider using 'indexMaybe' instead.
 index :: HasCallStack => ByteString -> Int64 -> Word8
 index _  i | i < 0  = moduleError "index" ("negative index: " ++ show i)
 index cs0 i         = index' cs0 i
@@ -1437,39 +1417,19 @@ unzip ls = (pack (List.map fst ls), pack (List.map snd ls))
 -- ---------------------------------------------------------------------
 -- Special lists
 
--- | Returns all initial segments of the given 'ByteString', shortest first.
+-- | /O(n)/ Return all initial segments of the given 'ByteString', shortest first.
 inits :: ByteString -> [ByteString]
--- see Note [Avoid NonEmpty combinators] in Data.ByteString
-inits bs = NE.toList $! initsNE bs
+inits = (Empty :) . inits'
+  where inits' Empty        = []
+        inits' (Chunk c cs) = List.map (`Chunk` Empty) (List.tail (S.inits c))
+                           ++ List.map (Chunk c) (inits' cs)
 
--- | Returns all initial segments of the given 'ByteString', shortest first.
---
--- @since 0.11.4.0
-initsNE :: ByteString -> NonEmpty ByteString
--- see Note [Avoid NonEmpty combinators] in Data.ByteString
-initsNE = (Empty :|) . inits' id
-  where
-    inits' :: (ByteString -> ByteString) -> ByteString -> [ByteString]
-    -- inits' f bs === map f (tail (inits bs))
-    inits' _ Empty = []
-    inits' f (Chunk c@(S.BS x len) cs)
-      = [f (S.BS x n `Chunk` Empty) | n <- [1..len]]
-      ++ inits' (f . Chunk c) cs
-
--- | /O(n)/ Returns all final segments of the given 'ByteString', longest first.
+-- | /O(n)/ Return all final segments of the given 'ByteString', longest first.
 tails :: ByteString -> [ByteString]
--- see Note [Avoid NonEmpty combinators] in Data.ByteString
-tails bs = NE.toList $! tailsNE bs
-
--- | /O(n)/ Returns all final segments of the given 'ByteString', longest first.
---
--- @since 0.11.4.0
-tailsNE :: ByteString -> NonEmpty ByteString
--- see Note [Avoid NonEmpty combinators] in Data.ByteString
-tailsNE bs = case uncons bs of
-  Nothing -> Empty :| []
-  Just (_, tl) -> bs :| tails tl
-
+tails Empty         = [Empty]
+tails cs@(Chunk c cs')
+  | S.length c == 1 = cs : tails cs'
+  | otherwise       = cs : tails (Chunk (S.unsafeTail c) cs')
 
 -- ---------------------------------------------------------------------
 -- Low level constructors
@@ -1608,11 +1568,9 @@ appendFile = modifyFile AppendMode
 getContents :: IO ByteString
 getContents = hGetContents stdin
 
--- | Outputs a 'ByteString' to the specified 'Handle'.
---
--- The chunks will be
--- written one at a time. Other threads might write to the 'Handle' in between,
--- and hence 'hPut' alone is not suitable for concurrent writes.
+-- | Outputs a 'ByteString' to the specified 'Handle'. The chunks will be
+-- written one at a time. Other threads might write to the 'Handle' between the
+-- writes, and hence 'hPut' alone might not be suitable for concurrent writes.
 --
 hPut :: Handle -> ByteString -> IO ()
 hPut h = foldrChunks (\c rest -> S.hPut h c >> rest) (return ())
@@ -1634,17 +1592,12 @@ hPutNonBlocking h bs@(Chunk c cs) = do
     0                     -> return bs
     _                     -> return (Chunk c' cs)
 
--- | A synonym for 'hPut', for compatibility
+-- | A synonym for @hPut@, for compatibility
 --
 hPutStr :: Handle -> ByteString -> IO ()
 hPutStr = hPut
 
--- | Write a ByteString to 'stdout'.
---
--- The chunks will be
--- written one at a time. Other threads might write to the 'stdout' in between,
--- and hence 'putStr' alone is not suitable for concurrent writes.
---
+-- | Write a ByteString to stdout
 putStr :: ByteString -> IO ()
 putStr = hPut stdout
 
