@@ -56,12 +56,10 @@ import Text.Read
 
 #endif
 
-import Prelude hiding (head, tail)
 import Control.Arrow
 import Data.Char
 import Data.Foldable
 import qualified Data.List as List
-import qualified Data.List.NonEmpty as NE
 import Data.Semigroup
 import Data.String
 import Data.Tuple
@@ -104,9 +102,8 @@ tests =
     -- but if it is ASCII, we should not generate Unicode filenames.
     enc <- getFileSystemEncoding
     pure $ case textEncodingName enc of
-      "ASCII"          -> property (prop . getASCIIString)
-      "ANSI_X3.4-1968" -> property (prop . getASCIIString)
-      _                -> property prop
+      "ASCII" -> property (prop . getASCIIString)
+      _       -> property prop
 #endif
 
   , testProperty "==" $
@@ -186,10 +183,6 @@ tests =
     \x -> map B.unpack (B.inits x) === List.inits (B.unpack x)
   , testProperty "tails" $
     \x -> map B.unpack (B.tails x) === List.tails (B.unpack x)
-  , testProperty "initsNE" $
-    \x -> NE.map B.unpack (B.initsNE x) === NE.inits (B.unpack x)
-  , testProperty "tailsNE" $
-    \x -> NE.map B.unpack (B.tailsNE x) === NE.tails (B.unpack x)
 #endif
   , testProperty "all" $
     \f x -> B.all f x === all f (B.unpack x)
@@ -434,15 +427,11 @@ tests =
     \x -> let n = 2^31 in (B.unpack *** B.unpack) (B.splitAt n x) === List.genericSplitAt n (B.unpack x)
 
   , testProperty "head" $
-    \x -> case B.unpack x of
-      []     -> property True
-      hd : _ -> B.head x === hd
+    \x -> not (B.null x) ==> B.head x === head (B.unpack x)
   , testProperty "last" $
     \x -> not (B.null x) ==> B.last x === last (B.unpack x)
   , testProperty "tail" $
-    \x -> case B.unpack x of
-      []     -> property True
-      _ : tl -> B.unpack (B.tail x) === tl
+    \x -> not (B.null x) ==> B.unpack (B.tail x) === tail (B.unpack x)
   , testProperty "tail length" $
     \x -> not (B.null x) ==> B.length x === 1 + B.length (B.tail x)
   , testProperty "init" $
