@@ -14,6 +14,7 @@
 #include "Threads.h"
 #include "RaiseAsync.h"
 #include "sm/Storage.h"
+#include "CloneStack.h"
 
 /* ----------------------------------------------------------------------------
    Send a message to another Capability
@@ -32,7 +33,8 @@ void sendMessage(Capability *from_cap, Capability *to_cap, Message *msg)
             i != &stg_MSG_BLACKHOLE_info &&
             i != &stg_MSG_TRY_WAKEUP_info &&
             i != &stg_IND_info && // can happen if a MSG_BLACKHOLE is revoked
-            i != &stg_WHITEHOLE_info) {
+            i != &stg_WHITEHOLE_info &&
+            i != &stg_MSG_CLONE_STACK_info) {
             barf("sendMessage: %p", i);
         }
     }
@@ -72,8 +74,8 @@ loop:
     if (i == &stg_MSG_TRY_WAKEUP_info)
     {
         StgTSO *tso = ((MessageWakeup *)m)->tso;
-        debugTraceCap(DEBUG_sched, cap, "message: try wakeup thread %ld",
-                      (W_)tso->id);
+        debugTraceCap(DEBUG_sched, cap, "message: try wakeup thread %"
+                      FMT_StgThreadID, tso->id);
         tryWakeupThread(cap, tso);
     }
     else if (i == &stg_MSG_THROWTO_info)
@@ -91,9 +93,6 @@ loop:
         debugTraceCap(DEBUG_sched, cap, "message: throwTo %ld -> %ld",
                       (W_)t->source->id, (W_)t->target->id);
 
-        ASSERT(t->source->why_blocked == BlockedOnMsgThrowTo);
-        ASSERT(t->source->block_info.closure == (StgClosure *)m);
-
         r = throwToMsg(cap, t);
 
         switch (r) {
@@ -133,6 +132,10 @@ loop:
 #endif
         goto loop;
     }
+    else if(i == &stg_MSG_CLONE_STACK_info){
+        MessageCloneStack *cloneStackMessage = (MessageCloneStack*) m;
+        handleCloneStackMessage(cloneStackMessage);
+    }
     else
     {
         barf("executeMessage: %p", i);
@@ -168,8 +171,8 @@ uint32_t messageBlackHole(Capability *cap, MessageBlackHole *msg)
     StgClosure *bh = UNTAG_CLOSURE(msg->bh);
     StgTSO *owner;
 
-    debugTraceCap(DEBUG_sched, cap, "message: thread %d blocking on "
-                  "blackhole %p", (W_)msg->tso->id, msg->bh);
+    debugTraceCap(DEBUG_sched, cap, "message: thread %" FMT_StgThreadID
+                  " blocking on blackhole %p", msg->tso->id, msg->bh);
 
     info = ACQUIRE_LOAD(&bh->header.info);
 
@@ -265,8 +268,8 @@ loop:
         }
         recordClosureMutated(cap,bh); // bh was mutated
 
-        debugTraceCap(DEBUG_sched, cap, "thread %d blocked on thread %d",
-                      (W_)msg->tso->id, (W_)owner->id);
+        debugTraceCap(DEBUG_sched, cap, "thread %" FMT_StgThreadID " blocked on"
+                      " thread %" FMT_StgThreadID, msg->tso->id, owner->id);
 
         return 1; // blocked
     }
@@ -309,9 +312,9 @@ loop:
         }
 
         debugTraceCap(DEBUG_sched, cap,
-                      "thread %d blocked on existing BLOCKING_QUEUE "
-                      "owned by thread %d",
-                      (W_)msg->tso->id, (W_)owner->id);
+                      "thread %" FMT_StgThreadID " blocked on existing "
+                      "BLOCKING_QUEUE owned by thread %" FMT_StgThreadID,
+                      msg->tso->id, owner->id);
 
         // See above, #3838
         if (owner->why_blocked == NotBlocked && owner->id != msg->tso->id) {
