@@ -8,6 +8,10 @@
 
 #pragma once
 
+#include "Capability.h"
+#include "Updates.h" // for DEBUG_FILL_SLOP
+#include "SMPClosureOps.h"
+
 #include "BeginPrivate.h"
 
 uint32_t messageBlackHole(Capability *cap, MessageBlackHole *msg);
@@ -18,21 +22,14 @@ void executeMessage (Capability *cap, Message *m);
 void sendMessage    (Capability *from_cap, Capability *to_cap, Message *msg);
 #endif
 
-#include "Capability.h"
-#include "Updates.h" // for DEBUG_FILL_SLOP
-#include "SMPClosureOps.h"
-
 INLINE_HEADER void
 doneWithMsgThrowTo (Capability *cap, MessageThrowTo *m)
 {
     // The message better be locked (unless we are running single-threaded,
     // where we are a bit more lenient (#19075).
-    ASSERT(getNumCapabilities() == 1 || m->header.info == &stg_WHITEHOLE_info);
+    ASSERT(n_capabilities == 1 || m->header.info == &stg_WHITEHOLE_info);
     IF_NONMOVING_WRITE_BARRIER_ENABLED {
-      updateRemembSetPushClosure(cap, (StgClosure *) m->link);
-      updateRemembSetPushClosure(cap, (StgClosure *) m->source);
-      updateRemembSetPushClosure(cap, (StgClosure *) m->target);
-      updateRemembSetPushClosure(cap, (StgClosure *) m->exception);
+      updateRemembSetPushMessageThrowTo(cap, m);
     }
     OVERWRITING_CLOSURE((StgClosure*)m);
     unlockClosure((StgClosure*)m, &stg_MSG_NULL_info);
