@@ -329,13 +329,7 @@ void traceConcUpdRemSetFlush(Capability *cap);
 void traceNonmovingHeapCensus(uint32_t log_blk_size,
                               const struct NonmovingAllocCensus *census);
 
-void traceIPE(StgInfoTable *info,
-               const char *table_name,
-               const char *closure_desc,
-               const char *ty_desc,
-               const char *label,
-               const char *module,
-               const char *srcloc );
+void traceIPE(const InfoProvEnt *ipe);
 void flushTrace(void);
 
 #else /* !TRACING */
@@ -372,7 +366,7 @@ void flushTrace(void);
 #define traceTaskDelete_(taskID) /* nothing */
 #define traceHeapProfBegin(profile_id) /* nothing */
 #define traceHeapProfCostCentre(ccID, label, module, srcloc, is_caf) /* nothing */
-#define traceIPE(info, table_name, closure_desc, ty_desc, label, module, srcloc) /* nothing */
+#define traceIPE(ipe) /* nothing */
 #define traceHeapProfSampleBegin(era) /* nothing */
 #define traceHeapBioProfSampleBegin(era, time) /* nothing */
 #define traceHeapProfSampleEnd(era) /* nothing */
