@@ -1410,34 +1410,28 @@ void postTickyCounterSamples(StgEntCounter *counters)
     RELEASE_LOCK(&eventBufMutex);
 }
 #endif /* TICKY_TICKY */
-void postIPE(StgWord64 info,
-             const char *table_name,
-             const char *closure_desc,
-             const char *ty_desc,
-             const char *label,
-             const char *module,
-             const char *srcloc)
+void postIPE(const InfoProvEnt *ipe)
 {
     ACQUIRE_LOCK(&eventBufMutex);
-    StgWord table_name_len = strlen(table_name);
-    StgWord closure_desc_len = strlen(closure_desc);
-    StgWord ty_desc_len = strlen(ty_desc);
-    StgWord label_len = strlen(label);
-    StgWord module_len = strlen(module);
-    StgWord srcloc_len = strlen(srcloc);
+    StgWord table_name_len = strlen(ipe->prov.table_name);
+    StgWord closure_desc_len = strlen(ipe->prov.closure_desc);
+    StgWord ty_desc_len = strlen(ipe->prov.ty_desc);
+    StgWord label_len = strlen(ipe->prov.label);
+    StgWord module_len = strlen(ipe->prov.module);
+    StgWord srcloc_len = strlen(ipe->prov.srcloc);
     // 8 for the info word
     // 6 for the number of strings in the payload as postString adds 1 to the length
     StgWord len = 8+table_name_len+closure_desc_len+ty_desc_len+label_len+module_len+srcloc_len+6;
     ensureRoomForVariableEvent(&eventBuf, len);
     postEventHeader(&eventBuf, EVENT_IPE);
     postPayloadSize(&eventBuf, len);
-    postWord64(&eventBuf, info);
-    postString(&eventBuf, table_name);
-    postString(&eventBuf, closure_desc);
-    postString(&eventBuf, ty_desc);
-    postString(&eventBuf, label);
-    postString(&eventBuf, module);
-    postString(&eventBuf, srcloc);
+    postWord64(&eventBuf, (StgWord) ipe->info);
+    postString(&eventBuf, ipe->prov.table_name);
+    postString(&eventBuf, ipe->prov.closure_desc);
+    postString(&eventBuf, ipe->prov.ty_desc);
+    postString(&eventBuf, ipe->prov.label);
+    postString(&eventBuf, ipe->prov.module);
+    postString(&eventBuf, ipe->prov.srcloc);
     RELEASE_LOCK(&eventBufMutex);
 }
 
