@@ -11,7 +11,7 @@
  *
  * ---------------------------------------------------------------------------*/
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 
 #include "Evac.h"
@@ -43,7 +43,6 @@
 
 /* Note [Selector optimisation depth limit]
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- *
  * MAX_THUNK_SELECTOR_DEPTH is used to avoid long recursion of
  * eval_thunk_selector due to nested selector thunks. Note that this *only*
  * counts nested selector thunks, e.g. `fst (fst (... (fst x)))`. The collector
@@ -177,7 +176,6 @@ alloc_for_copy (uint32_t size, uint32_t gen_no)
 /*
  * Note [Non-moving GC: Marking evacuated objects]
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- *
  * When the non-moving collector is in use we must be careful to ensure that any
  * references to objects in the non-moving generation from younger generations
  * are pushed to the mark queue.
@@ -698,7 +696,7 @@ loop:
   if (!HEAP_ALLOCED_GC(q)) {
       if (!major_gc) return;
 
-      // Note [Object unloading] in CheckUnload.c
+      // See Note [Object unloading] in CheckUnload.c
       if (RTS_UNLIKELY(unload_mark_needed)) {
           markObjectCode(q);
       }
@@ -936,7 +934,7 @@ loop:
               return;
           }
           // Note [BLACKHOLE pointing to IND]
-          //
+          // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
           // BLOCKING_QUEUE can be overwritten by IND (see
           // wakeBlockingQueue()). However, when this happens we must
           // be updating the BLACKHOLE, so the BLACKHOLE's indirectee
@@ -1251,13 +1249,8 @@ selector_chain:
         // save any space in any case, and updating with an indirection is
         // trickier in a non-collected gen: we would have to update the
         // mutable list.
-        uint16_t flags = RELAXED_LOAD(&bd->flags);
-        if (flags & (BF_EVACUATED | BF_NONMOVING)) {
+        if (RELAXED_LOAD(&bd->flags) & (BF_EVACUATED | BF_NONMOVING)) {
             unchain_thunk_selectors(prev_thunk_selector, (StgClosure *)p);
-            if (flags & BF_NONMOVING) {
-                // See Note [Non-moving GC: Marking evacuated objects].
-                markQueuePushClosureGC(&gct->cap->upd_rem_set.queue, (StgClosure*) p);
-            }
             *q = (StgClosure *)p;
             // shortcut, behave as for:  if (evac) evacuate(q);
             if (evac && bd->gen_no < gct->evac_gen_no) {
@@ -1272,7 +1265,7 @@ selector_chain:
         // (scavenge_mark_stack doesn't deal with IND).  BEWARE!  This
         // bit is very tricky to get right.  If you make changes
         // around here, test by compiling stage 3 with +RTS -c -RTS.
-        if (flags & BF_MARKED) {
+        if (bd->flags & BF_MARKED) {
             // must call evacuate() to mark this closure if evac==true
             *q = (StgClosure *)p;
             if (evac) evacuate(q);
@@ -1312,12 +1305,6 @@ selector_chain:
             //   - undo the chain we've built to point to p.
             SET_INFO((StgClosure *)p, (const StgInfoTable *)info_ptr);
             RELEASE_STORE(q, (StgClosure *) p);
-            if (Bdescr((StgPtr)p)->flags & BF_NONMOVING) {
-                // See Note [Non-moving GC: Marking evacuated objects].
-                // TODO: This really shouldn't be necessary since whoever won
-                // the race should have pushed
-                markQueuePushClosureGC(&gct->cap->upd_rem_set.queue, (StgClosure*) p);
-            }
             if (evac) evacuate(q);
             unchain_thunk_selectors(prev_thunk_selector, (StgClosure *)p);
             return;
@@ -1408,11 +1395,6 @@ selector_loop:
                   case THUNK_SELECTOR:
                       // Use payload to make a list of thunk selectors, to be
                       // used in unchain_thunk_selectors
-                      //
-                      // FIXME: This seems racy; should we lock this selector to
-                      // ensure that another thread doesn't clobber this node
-                      // of the chain. This would result in some previous
-                      // selectors not being updated when we unchain.
                       RELAXED_STORE(&((StgClosure*)p)->payload[0], (StgClosure *)prev_thunk_selector);
                       prev_thunk_selector = p;
                       p = (StgSelector*)val;
@@ -1437,12 +1419,6 @@ selector_loop:
               // eval_thunk_selector(), because we know val is not
               // a THUNK_SELECTOR.
               if (evac) evacuate(q);
-
-              if (isNonmovingClosure(*q)) {
-                  // See Note [Non-moving GC: Marking evacuated objects].
-                  markQueuePushClosureGC(&gct->cap->upd_rem_set.queue, (StgClosure*) *q);
-              }
-
               return;
           }
 
@@ -1487,10 +1463,6 @@ selector_loop:
           // recurse indefinitely, so we impose a depth bound.
           // See Note [Selector optimisation depth limit].
           if (gct->thunk_selector_depth >= MAX_THUNK_SELECTOR_DEPTH) {
-              if (isNonmovingClosure((StgClosure *) p)) {
-                  // See Note [Non-moving GC: Marking evacuated objects].
-                  markQueuePushClosureGC(&gct->cap->upd_rem_set.queue, (StgClosure*) p);
-              }
               goto bale_out;
           }
 
@@ -1537,9 +1509,5 @@ bale_out:
     if (evac) {
         copy(q,(const StgInfoTable *)info_ptr,(StgClosure *)p,THUNK_SELECTOR_sizeW(),bd->dest_no);
     }
-    if (isNonmovingClosure(*q)) {
-        // See Note [Non-moving GC: Marking evacuated objects].
-        markQueuePushClosureGC(&gct->cap->upd_rem_set.queue, *q);
-    }
     unchain_thunk_selectors(prev_thunk_selector, *q);
 }
