@@ -233,7 +233,7 @@ Mutex concurrent_coll_finished_lock;
  *  - Note [StgStack dirtiness flags and concurrent marking] (TSO.h) describes
  *    the protocol for concurrent marking of stacks.
  *
- *  - Note [Nonmoving write barrier in Perform{Take,Put}] (PrimOps.cmm) describes
+ *  - Note [Nonmoving write barrier in Perform{Put,Take}] (PrimOps.cmm) describes
  *    a tricky barrier necessary when resuming threads blocked on MVar
  *    operations.
  *
@@ -244,12 +244,6 @@ Mutex concurrent_coll_finished_lock;
  *    how we use the DIRTY flags associated with MUT_VARs and TVARs to improve
  *    barrier efficiency.
  *
- *  - Note [Weak pointer processing and the non-moving GC] (MarkWeak.c) describes
- *    how weak pointers are handled when the non-moving GC is in use.
- *
- *  - Note [Sync phase marking budget] describes how we avoid long mutator
- *    pauses during the sync phase
- *
  * [ueno 2016]:
  *   Katsuhiro Ueno and Atsushi Ohori. 2016. A fully concurrent garbage
  *   collector for functional programs on multicore processors. SIGPLAN Not. 51,
@@ -298,7 +292,6 @@ Mutex concurrent_coll_finished_lock;
  *                    ┆
  *      B ←────────────── A ←─────────────── root
  *      │             ┆     ↖─────────────── gen1 mut_list
- *      │             ┆
  *      ╰───────────────→ C
  *                    ┆
  *
@@ -339,9 +332,8 @@ Mutex concurrent_coll_finished_lock;
  * The implementation details of this are described in Note [Non-moving GC:
  * Marking evacuated objects] in Evac.c.
  *
- *
- * Note [Deadlock detection under the non-moving collector]
- * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+ * Note [Deadlock detection under nonmoving collector]
+ * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  * In GHC the garbage collector is responsible for identifying deadlocked
  * programs. Providing for this responsibility is slightly tricky in the
  * non-moving collector due to the existence of aging. In particular, the
@@ -501,44 +493,10 @@ Mutex concurrent_coll_finished_lock;
  * remembered set during the preparatory GC. This allows us to safely skip the
  * non-moving write barrier without jeopardizing the snapshot invariant.
  *
- *
- * Note [Sync phase marking budget]
- * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- * The non-moving collector is intended to provide reliably low collection
- * latencies. These latencies are primarily due to two sources:
- *
- *  a. the preparatory moving collection at the beginning of the major GC cycle
- *  b. the post-mark synchronization pause at the end
- *
- * While the cost of (a) is inherently bounded by the young generation size,
- * (b) can in principle be unbounded since the mutator may hide large swathes
- * of heap from the collector's concurrent mark phase via mutation. These will
- * only become visible to the collector during the post-mark synchronization
- * phase.
- *
- * Since we don't want to do unbounded marking work in the pause, we impose a
- * limit (specifically, sync_phase_marking_budget) on the amount of work
- * (namely, the number of marked closures) that we can do during the pause. If
- * we deplete our marking budget during the pause then we allow the mutators to
- * resume and return to concurrent marking (keeping the update remembered set
- * write barrier enabled). After we have finished marking we will again
- * attempt the post-mark synchronization.
- *
- * The choice of sync_phase_marking_budget was made empirically. On 2022
- * hardware and a "typical" test program we tend to mark ~10^7 closures per
- * second. Consequently, a sync_phase_marking_budget of 10^5 should produce
- * ~10 ms pauses, which seems like a reasonable tradeoff.
- *
- * TODO: Perhaps sync_phase_marking_budget should be controllable via a
- * command-line argument?
- *
  */
 
 memcount nonmoving_live_words = 0;
 
-// See Note [Sync phase marking budget].
-MarkBudget sync_phase_marking_budget = 200000;
-
 #if defined(THREADED_RTS)
 static void* nonmovingConcurrentMark(void *mark_queue);
 #endif
@@ -707,11 +665,10 @@ void *nonmovingAllocate(Capability *cap, StgWord sz)
     // object and not moved) which is covered by allocator 9.
     ASSERT(log_block_size < NONMOVING_ALLOCA0 + NONMOVING_ALLOCA_CNT);
 
-    unsigned int alloca_idx = log_block_size - NONMOVING_ALLOCA0;
-    struct NonmovingAllocator *alloca = &nonmovingHeap.allocators[alloca_idx];
+    struct NonmovingAllocator *alloca = nonmovingHeap.allocators[log_block_size - NONMOVING_ALLOCA0];
 
     // Allocate into current segment
-    struct NonmovingSegment *current = cap->current_segments[alloca_idx];
+    struct NonmovingSegment *current = alloca->current[cap->no];
     ASSERT(current); // current is never NULL
     void *ret = nonmovingSegmentGetBlock_(current, log_block_size, current->next_free);
     ASSERT(GET_CLOSURE_TAG(ret) == 0); // check alignment
@@ -744,12 +701,29 @@ void *nonmovingAllocate(Capability *cap, StgWord sz)
         // make it current
         new_current->link = NULL;
         SET_SEGMENT_STATE(new_current, CURRENT);
-        cap->current_segments[alloca_idx] = new_current;
+        alloca->current[cap->no] = new_current;
     }
 
     return ret;
 }
 
+/* Allocate a nonmovingAllocator */
+static struct NonmovingAllocator *alloc_nonmoving_allocator(uint32_t n_caps)
+{
+    size_t allocator_sz =
+        sizeof(struct NonmovingAllocator) +
+        sizeof(void*) * n_caps; // current segment pointer for each capability
+    struct NonmovingAllocator *alloc =
+        stgMallocBytes(allocator_sz, "nonmovingInit");
+    memset(alloc, 0, allocator_sz);
+    return alloc;
+}
+
+static void free_nonmoving_allocator(struct NonmovingAllocator *alloc)
+{
+    stgFree(alloc);
+}
+
 void nonmovingInit(void)
 {
     if (! RtsFlags.GcFlags.useNonmoving) return;
@@ -758,7 +732,10 @@ void nonmovingInit(void)
     initCondition(&concurrent_coll_finished);
     initMutex(&concurrent_coll_finished_lock);
 #endif
-    nonmovingMarkInit();
+    for (unsigned int i = 0; i < NONMOVING_ALLOCA_CNT; i++) {
+        nonmovingHeap.allocators[i] = alloc_nonmoving_allocator(n_capabilities);
+    }
+    nonmovingMarkInitUpdRemSet();
 }
 
 // Stop any nonmoving collection in preparation for RTS shutdown.
@@ -787,24 +764,44 @@ void nonmovingExit(void)
     closeCondition(&concurrent_coll_finished);
     closeMutex(&nonmoving_collection_mutex);
 #endif
+
+    for (unsigned int i = 0; i < NONMOVING_ALLOCA_CNT; i++) {
+        free_nonmoving_allocator(nonmovingHeap.allocators[i]);
+    }
 }
 
-/* Initialize a new capability. Caller must hold SM_LOCK */
-void nonmovingInitCapability(Capability *cap)
+/*
+ * Assumes that no garbage collector or mutator threads are running to safely
+ * resize the nonmoving_allocators.
+ *
+ * Must hold sm_mutex.
+ */
+void nonmovingAddCapabilities(uint32_t new_n_caps)
 {
-    // Initialize current segment array
-    struct NonmovingSegment **segs =
-        stgMallocBytes(sizeof(struct NonmovingSegment*) * NONMOVING_ALLOCA_CNT, "current segment array");
-    for (unsigned int i = 0; i < NONMOVING_ALLOCA_CNT; i++) {
-        segs[i] = nonmovingAllocSegment(cap->node);
-        nonmovingInitSegment(segs[i], NONMOVING_ALLOCA0 + i);
-        SET_SEGMENT_STATE(segs[i], CURRENT);
-    }
-    cap->current_segments = segs;
+    unsigned int old_n_caps = nonmovingHeap.n_caps;
+    struct NonmovingAllocator **allocs = nonmovingHeap.allocators;
 
-    // Initialize update remembered set
-    cap->upd_rem_set.queue.blocks = NULL;
-    nonmovingInitUpdRemSet(&cap->upd_rem_set);
+    for (unsigned int i = 0; i < NONMOVING_ALLOCA_CNT; i++) {
+        struct NonmovingAllocator *old = allocs[i];
+        allocs[i] = alloc_nonmoving_allocator(new_n_caps);
+
+        // Copy the old state
+        allocs[i]->filled = old->filled;
+        allocs[i]->active = old->active;
+        for (unsigned int j = 0; j < old_n_caps; j++) {
+            allocs[i]->current[j] = old->current[j];
+        }
+        stgFree(old);
+
+        // Initialize current segments for the new capabilities
+        for (unsigned int j = old_n_caps; j < new_n_caps; j++) {
+            allocs[i]->current[j] = nonmovingAllocSegment(capabilities[j]->node);
+            nonmovingInitSegment(allocs[i]->current[j], NONMOVING_ALLOCA0 + i);
+            SET_SEGMENT_STATE(allocs[i]->current[j], CURRENT);
+            allocs[i]->current[j]->link = NULL;
+        }
+    }
+    nonmovingHeap.n_caps = new_n_caps;
 }
 
 void nonmovingClearBitmap(struct NonmovingSegment *seg)
@@ -824,21 +821,18 @@ static void nonmovingPrepareMark(void)
     // Should have been cleared by the last sweep
     ASSERT(nonmovingHeap.sweep_list == NULL);
 
-    nonmovingHeap.n_caps = n_capabilities;
     nonmovingBumpEpoch();
     for (int alloca_idx = 0; alloca_idx < NONMOVING_ALLOCA_CNT; ++alloca_idx) {
-        struct NonmovingAllocator *alloca = &nonmovingHeap.allocators[alloca_idx];
+        struct NonmovingAllocator *alloca = nonmovingHeap.allocators[alloca_idx];
 
         // Update current segments' snapshot pointers
-        for (uint32_t cap_n = 0; cap_n < nonmovingHeap.n_caps; ++cap_n) {
-            Capability *cap = capabilities[cap_n];
-            struct NonmovingSegment *seg = cap->current_segments[alloca_idx];
+        for (uint32_t cap_n = 0; cap_n < n_capabilities; ++cap_n) {
+            struct NonmovingSegment *seg = alloca->current[cap_n];
             nonmovingSegmentInfo(seg)->next_free_snap = seg->next_free;
         }
 
         // Save the filled segments for later processing during the concurrent
         // mark phase.
-        ASSERT(alloca->saved_filled == NULL);
         alloca->saved_filled = alloca->filled;
         alloca->filled = NULL;
 
@@ -892,7 +886,38 @@ static void nonmovingPrepareMark(void)
 #endif
 }
 
-void nonmovingCollect(StgWeak **dead_weaks STG_UNUSED, StgTSO **resurrected_threads)
+// Mark weak pointers in the non-moving heap. They'll either end up in
+// dead_weak_ptr_list or stay in weak_ptr_list. Either way they need to be kept
+// during sweep. See `MarkWeak.c:markWeakPtrList` for the moving heap variant
+// of this.
+static void nonmovingMarkWeakPtrList(MarkQueue *mark_queue, StgWeak *dead_weak_ptr_list)
+{
+    for (StgWeak *w = oldest_gen->weak_ptr_list; w; w = w->link) {
+        markQueuePushClosure_(mark_queue, (StgClosure*)w);
+        // Do not mark finalizers and values here, those fields will be marked
+        // in `nonmovingMarkDeadWeaks` (for dead weaks) or
+        // `nonmovingTidyWeaks` (for live weaks)
+    }
+
+    // We need to mark dead_weak_ptr_list too. This is subtle:
+    //
+    // - By the beginning of this GC we evacuated all weaks to the non-moving
+    //   heap (in `markWeakPtrList`)
+    //
+    // - During the scavenging of the moving heap we discovered that some of
+    //   those weaks are dead and moved them to `dead_weak_ptr_list`. Note that
+    //   because of the fact above _all weaks_ are in the non-moving heap at
+    //   this point.
+    //
+    // - So, to be able to traverse `dead_weak_ptr_list` and run finalizers we
+    //   need to mark it.
+    for (StgWeak *w = dead_weak_ptr_list; w; w = w->link) {
+        markQueuePushClosure_(mark_queue, (StgClosure*)w);
+        nonmovingMarkDeadWeak(mark_queue, w);
+    }
+}
+
+void nonmovingCollect(StgWeak **dead_weaks, StgTSO **resurrected_threads)
 {
 #if defined(THREADED_RTS)
     // We can't start a new collection until the old one has finished
@@ -914,31 +939,23 @@ void nonmovingCollect(StgWeak **dead_weaks STG_UNUSED, StgTSO **resurrected_thre
     ASSERT(n_nonmoving_marked_compact_blocks == 0);
 
     MarkQueue *mark_queue = stgMallocBytes(sizeof(MarkQueue), "mark queue");
-    mark_queue->blocks = NULL;
     initMarkQueue(mark_queue);
     current_mark_queue = mark_queue;
 
     // Mark roots
     trace(TRACE_nonmoving_gc, "Marking roots for nonmoving GC");
     markCAFs((evac_fn)markQueueAddRoot, mark_queue);
-    for (unsigned int n = 0; n < getNumCapabilities(); ++n) {
+    for (unsigned int n = 0; n < n_capabilities; ++n) {
         markCapability((evac_fn)markQueueAddRoot, mark_queue,
                 capabilities[n], true/*don't mark sparks*/);
     }
     markScheduler((evac_fn)markQueueAddRoot, mark_queue);
+    nonmovingMarkWeakPtrList(mark_queue, *dead_weaks);
     markStablePtrTable((evac_fn)markQueueAddRoot, mark_queue);
 
-    // The dead weak pointer list shouldn't contain any weaks in the
-    // nonmoving heap
-#if defined(DEBUG)
-    for (StgWeak *w = *dead_weaks; w; w = w->link) {
-        ASSERT(Bdescr((StgPtr) w)->gen != oldest_gen);
-    }
-#endif
-
     // Mark threads resurrected during moving heap scavenging
     for (StgTSO *tso = *resurrected_threads; tso != END_TSO_QUEUE; tso = tso->global_link) {
-        markQueuePushClosureGC(mark_queue, (StgClosure*)tso);
+        markQueuePushClosure_(mark_queue, (StgClosure*)tso);
     }
     trace(TRACE_nonmoving_gc, "Finished marking roots for nonmoving GC");
 
@@ -961,23 +978,8 @@ void nonmovingCollect(StgWeak **dead_weaks STG_UNUSED, StgTSO **resurrected_thre
     // alive).
     ASSERT(oldest_gen->old_weak_ptr_list == NULL);
     ASSERT(nonmoving_old_weak_ptr_list == NULL);
-    {
-        // Move both oldest_gen->weak_ptr_list and nonmoving_weak_ptr_list to
-        // nonmoving_old_weak_ptr_list
-        StgWeak **weaks = &oldest_gen->weak_ptr_list;
-        uint32_t n = 0;
-        while (*weaks) {
-            weaks = &(*weaks)->link;
-            n++;
-        }
-        debugTrace(DEBUG_nonmoving_gc, "%d new nonmoving weaks", n);
-        *weaks = nonmoving_weak_ptr_list;
-        nonmoving_old_weak_ptr_list = oldest_gen->weak_ptr_list;
-        nonmoving_weak_ptr_list = NULL;
-        oldest_gen->weak_ptr_list = NULL;
-        // At this point all weaks in the nonmoving generation are on
-        // nonmoving_old_weak_ptr_list
-    }
+    nonmoving_old_weak_ptr_list = oldest_gen->weak_ptr_list;
+    oldest_gen->weak_ptr_list = NULL;
     trace(TRACE_nonmoving_gc, "Finished nonmoving GC preparation");
 
     // We are now safe to start concurrent marking
@@ -1013,25 +1015,19 @@ void nonmovingCollect(StgWeak **dead_weaks STG_UNUSED, StgTSO **resurrected_thre
 }
 
 /* Mark queue, threads, and weak pointers until no more weaks have been
- * resuscitated. If *budget is non-zero then we will mark no more than
- * Returns true if we there is no more marking work to be done, false if
- * we exceeded our marking budget.
+ * resuscitated
  */
-static bool nonmovingMarkThreadsWeaks(MarkBudget *budget, MarkQueue *mark_queue)
+static void nonmovingMarkThreadsWeaks(MarkQueue *mark_queue)
 {
     while (true) {
         // Propagate marks
-        nonmovingMark(budget, mark_queue);
-        if (*budget == 0) {
-            return false;
-        }
+        nonmovingMark(mark_queue);
 
         // Tidy threads and weaks
         nonmovingTidyThreads();
 
-        if (! nonmovingTidyWeaks(mark_queue)) {
-            return true;
-        }
+        if (! nonmovingTidyWeaks(mark_queue))
+            return;
     }
 }
 
@@ -1045,6 +1041,7 @@ static void* nonmovingConcurrentMark(void *data)
     return NULL;
 }
 
+// TODO: Not sure where to put this function.
 // Append w2 to the end of w1.
 static void appendWeakList( StgWeak **w1, StgWeak *w2 )
 {
@@ -1064,40 +1061,28 @@ static void nonmovingMark_(MarkQueue *mark_queue, StgWeak **dead_weaks, StgTSO *
     // Walk the list of filled segments that we collected during preparation,
     // updated their snapshot pointers and move them to the sweep list.
     for (int alloca_idx = 0; alloca_idx < NONMOVING_ALLOCA_CNT; ++alloca_idx) {
-        struct NonmovingSegment *filled = nonmovingHeap.allocators[alloca_idx].saved_filled;
+        struct NonmovingSegment *filled = nonmovingHeap.allocators[alloca_idx]->saved_filled;
         uint32_t n_filled = 0;
         if (filled) {
             struct NonmovingSegment *seg = filled;
             while (true) {
                 // Set snapshot
                 nonmovingSegmentInfo(seg)->next_free_snap = seg->next_free;
-                SET_SEGMENT_STATE(seg, FILLED_SWEEPING);
                 n_filled++;
-                if (seg->link) {
+                if (seg->link)
                     seg = seg->link;
-                } else {
+                else
                     break;
-                }
             }
             // add filled segments to sweep_list
+            SET_SEGMENT_STATE(seg, FILLED_SWEEPING);
             seg->link = nonmovingHeap.sweep_list;
             nonmovingHeap.sweep_list = filled;
         }
-        nonmovingHeap.allocators[alloca_idx].saved_filled = NULL;
     }
 
-    // Mark Weak#s
-    nonmovingMarkWeakPtrList(mark_queue);
-
     // Do concurrent marking; most of the heap will get marked here.
-
-#if defined(THREADED_RTS)
-concurrent_marking:
-#endif
-    {
-        MarkBudget budget = UNLIMITED_MARK_BUDGET;
-        nonmovingMarkThreadsWeaks(&budget, mark_queue);
-    }
+    nonmovingMarkThreadsWeaks(mark_queue);
 
 #if defined(THREADED_RTS)
     Task *task = newBoundTask();
@@ -1106,13 +1091,21 @@ concurrent_marking:
     if (sched_state > SCHED_RUNNING) {
         // Note that we break our invariants here and leave segments in
         // nonmovingHeap.sweep_list, don't free nonmoving_large_objects etc.
-        // However because we won't be running sweep in the final GC this
+        // However because we won't be running mark-sweep in the final GC this
         // is OK.
-        //
-        // However, we must move any weak pointers remaining on
-        // nonmoving_old_weak_ptr_list back to nonmoving_weak_ptr_list
-        // such that their C finalizers can be run by hs_exit_.
-        appendWeakList(&nonmoving_weak_ptr_list, nonmoving_old_weak_ptr_list);
+
+        // This is a RTS shutdown so we need to move our copy (snapshot) of
+        // weaks (nonmoving_old_weak_ptr_list and nonmoving_weak_ptr_list) to
+        // oldest_gen->threads to be able to run C finalizers in hs_exit_. Note
+        // that there may be more weaks added to oldest_gen->threads since we
+        // started mark, so we need to append our list to the tail of
+        // oldest_gen->threads.
+        appendWeakList(&nonmoving_old_weak_ptr_list, nonmoving_weak_ptr_list);
+        appendWeakList(&oldest_gen->weak_ptr_list, nonmoving_old_weak_ptr_list);
+        // These lists won't be used again so this is not necessary, but still
+        nonmoving_old_weak_ptr_list = NULL;
+        nonmoving_weak_ptr_list = NULL;
+
         goto finish;
     }
 
@@ -1120,17 +1113,9 @@ concurrent_marking:
     nonmovingBeginFlush(task);
 
     bool all_caps_syncd;
-    MarkBudget sync_marking_budget = sync_phase_marking_budget;
     do {
         all_caps_syncd = nonmovingWaitForFlush();
-        if (nonmovingMarkThreadsWeaks(&sync_marking_budget, mark_queue) == false) {
-            // We ran out of budget for marking. Abort sync.
-            // See Note [Sync phase marking budget].
-            traceConcSyncEnd();
-            stat_endNonmovingGcSync();
-            releaseAllCapabilities(n_capabilities, NULL, task);
-            goto concurrent_marking;
-        }
+        nonmovingMarkThreadsWeaks(mark_queue);
     } while (!all_caps_syncd);
 #endif
 
@@ -1141,7 +1126,7 @@ concurrent_marking:
     // Do last marking of weak pointers
     while (true) {
         // Propagate marks
-        nonmovingMarkUnlimitedBudget(mark_queue);
+        nonmovingMark(mark_queue);
 
         if (!nonmovingTidyWeaks(mark_queue))
             break;
@@ -1150,7 +1135,7 @@ concurrent_marking:
     nonmovingMarkDeadWeaks(mark_queue, dead_weaks);
 
     // Propagate marks
-    nonmovingMarkUnlimitedBudget(mark_queue);
+    nonmovingMark(mark_queue);
 
     // Now remove all dead objects from the mut_list to ensure that a younger
     // generation collection doesn't attempt to look at them after we've swept.
@@ -1192,14 +1177,20 @@ concurrent_marking:
         nonmoving_old_threads = END_TSO_QUEUE;
     }
 
-    // At this point point any weak that remains on nonmoving_old_weak_ptr_list
-    // has a dead key.
-    nonmoving_old_weak_ptr_list = NULL;
+    {
+        StgWeak **weaks = &oldest_gen->weak_ptr_list;
+        while (*weaks) {
+            weaks = &(*weaks)->link;
+        }
+        *weaks = nonmoving_weak_ptr_list;
+        nonmoving_weak_ptr_list = NULL;
+        nonmoving_old_weak_ptr_list = NULL;
+    }
 
     // Prune spark lists
     // See Note [Spark management under the nonmoving collector].
 #if defined(THREADED_RTS)
-    for (uint32_t n = 0; n < getNumCapabilities(); n++) {
+    for (uint32_t n = 0; n < n_capabilities; n++) {
         pruneSparkQueue(true, capabilities[n]);
     }
 #endif
@@ -1237,7 +1228,7 @@ concurrent_marking:
     traceConcSweepEnd();
 #if defined(DEBUG)
     if (RtsFlags.DebugFlags.nonmoving_gc)
-        nonmovingPrintAllocatorCensus(true);
+        nonmovingPrintAllocatorCensus();
 #endif
 #if defined(TRACING)
     if (RtsFlags.TraceFlags.nonmoving_gc)
@@ -1275,7 +1266,7 @@ void assert_in_nonmoving_heap(StgPtr p)
     bdescr *bd = Bdescr(p);
     if (bd->flags & BF_LARGE) {
         // It should be in a capability (if it's not filled yet) or in non-moving heap
-        for (uint32_t cap = 0; cap < getNumCapabilities(); ++cap) {
+        for (uint32_t cap = 0; cap < n_capabilities; ++cap) {
             if (bd == capabilities[cap]->pinned_object_block) {
                 return;
             }
@@ -1292,12 +1283,10 @@ void assert_in_nonmoving_heap(StgPtr p)
     }
 
     for (int alloca_idx = 0; alloca_idx < NONMOVING_ALLOCA_CNT; ++alloca_idx) {
-        struct NonmovingAllocator *alloca = &nonmovingHeap.allocators[alloca_idx];
-
+        struct NonmovingAllocator *alloca = nonmovingHeap.allocators[alloca_idx];
         // Search current segments
-        for (uint32_t cap_idx = 0; cap_idx < nonmovingHeap.n_caps; ++cap_idx) {
-            Capability *cap = capabilities[cap_idx];
-            struct NonmovingSegment *seg = cap->current_segments[alloca_idx];
+        for (uint32_t cap_idx = 0; cap_idx < n_capabilities; ++cap_idx) {
+            struct NonmovingSegment *seg = alloca->current[cap_idx];
             if (p >= (P_)seg && p < (((P_)seg) + NONMOVING_SEGMENT_SIZE_W)) {
                 return;
             }
@@ -1356,16 +1345,33 @@ void nonmovingPrintSegment(struct NonmovingSegment *seg)
     debugBelch("End of segment\n\n");
 }
 
+void nonmovingPrintAllocator(struct NonmovingAllocator *alloc)
+{
+    debugBelch("Allocator at %p\n", (void*)alloc);
+    debugBelch("Filled segments:\n");
+    for (struct NonmovingSegment *seg = alloc->filled; seg != NULL; seg = seg->link) {
+        debugBelch("%p ", (void*)seg);
+    }
+    debugBelch("\nActive segments:\n");
+    for (struct NonmovingSegment *seg = alloc->active; seg != NULL; seg = seg->link) {
+        debugBelch("%p ", (void*)seg);
+    }
+    debugBelch("\nCurrent segments:\n");
+    for (uint32_t i = 0; i < n_capabilities; ++i) {
+        debugBelch("%p ", alloc->current[i]);
+    }
+    debugBelch("\n");
+}
+
 void locate_object(P_ obj)
 {
     // Search allocators
     for (int alloca_idx = 0; alloca_idx < NONMOVING_ALLOCA_CNT; ++alloca_idx) {
-        struct NonmovingAllocator *alloca = &nonmovingHeap.allocators[alloca_idx];
-        for (uint32_t cap_n = 0; cap_n < nonmovingHeap.n_caps; ++cap_n) {
-            Capability *cap = capabilities[cap_n];
-            struct NonmovingSegment *seg = cap->current_segments[alloca_idx];
+        struct NonmovingAllocator *alloca = nonmovingHeap.allocators[alloca_idx];
+        for (uint32_t cap = 0; cap < n_capabilities; ++cap) {
+            struct NonmovingSegment *seg = alloca->current[cap];
             if (obj >= (P_)seg && obj < (((P_)seg) + NONMOVING_SEGMENT_SIZE_W)) {
-                debugBelch("%p is in current segment of capability %d of allocator %d at %p\n", obj, cap_n, alloca_idx, (void*)seg);
+                debugBelch("%p is in current segment of capability %d of allocator %d at %p\n", obj, cap, alloca_idx, (void*)seg);
                 return;
             }
         }
@@ -1493,7 +1499,7 @@ void nonmovingPrintSweepList()
 
 void check_in_mut_list(StgClosure *p)
 {
-    for (uint32_t cap_n = 0; cap_n < getNumCapabilities(); ++cap_n) {
+    for (uint32_t cap_n = 0; cap_n < n_capabilities; ++cap_n) {
         for (bdescr *bd = capabilities[cap_n]->mut_lists[oldest_gen->no]; bd; bd = bd->link) {
             for (StgPtr q = bd->start; q < bd->free; ++q) {
                 if (*((StgPtr**)q) == (StgPtr*)p) {
